@@ -33,12 +33,13 @@ bool pl_color_system_is_ycbcr_like(enum pl_color_system sys)
     case PL_COLOR_SYSTEM_BT_2020_C:
     case PL_COLOR_SYSTEM_BT_2100_PQ:
     case PL_COLOR_SYSTEM_BT_2100_HLG:
+    case PL_COLOR_SYSTEM_DOLBYVISION:
     case PL_COLOR_SYSTEM_YCGCO:
         return true;
     case PL_COLOR_SYSTEM_COUNT: break;
     };
 
-    abort();
+    pl_unreachable();
 }
 
 bool pl_color_system_is_linear(enum pl_color_system sys)
@@ -55,12 +56,13 @@ bool pl_color_system_is_linear(enum pl_color_system sys)
     case PL_COLOR_SYSTEM_BT_2020_C:
     case PL_COLOR_SYSTEM_BT_2100_PQ:
     case PL_COLOR_SYSTEM_BT_2100_HLG:
+    case PL_COLOR_SYSTEM_DOLBYVISION:
     case PL_COLOR_SYSTEM_XYZ:
         return false;
     case PL_COLOR_SYSTEM_COUNT: break;
     };
 
-    abort();
+    pl_unreachable();
 }
 
 enum pl_color_system pl_color_system_guess_ycbcr(int width, int height)
@@ -115,6 +117,7 @@ bool pl_color_repr_equal(const struct pl_color_repr *c1,
     return c1->sys    == c2->sys &&
            c1->levels == c2->levels &&
            c1->alpha  == c2->alpha &&
+           c1->dovi   == c2->dovi &&
            pl_bit_encoding_equal(&c1->bits, &c2->bits);
 }
 
@@ -128,19 +131,22 @@ static struct pl_bit_encoding pl_bit_encoding_merge(const struct pl_bit_encoding
     };
 }
 
-void pl_color_repr_merge(struct pl_color_repr *orig,
-                         const struct pl_color_repr *new)
+void pl_color_repr_merge(struct pl_color_repr *orig, const struct pl_color_repr *new)
 {
     *orig = (struct pl_color_repr) {
         .sys    = PL_DEF(orig->sys,    new->sys),
         .levels = PL_DEF(orig->levels, new->levels),
         .alpha  = PL_DEF(orig->alpha,  new->alpha),
+        .dovi   = PL_DEF(orig->dovi,   new->dovi),
         .bits   = pl_bit_encoding_merge(&orig->bits, &new->bits),
     };
 }
 
 enum pl_color_levels pl_color_levels_guess(const struct pl_color_repr *repr)
 {
+    if (repr->sys == PL_COLOR_SYSTEM_DOLBYVISION)
+        return PL_COLOR_LEVELS_FULL;
+
     if (repr->levels)
         return repr->levels;
 
@@ -172,7 +178,7 @@ float pl_color_repr_normalize(struct pl_color_repr *repr)
         scale *= ((1LL << tex_bits) - 1.) / ((1LL << col_bits) - 1.);
     }
 
-    bits->sample_depth = bits->color_depth;
+    bits->color_depth = bits->sample_depth;
     return scale;
 }
 
@@ -200,7 +206,7 @@ bool pl_color_primaries_is_wide_gamut(enum pl_color_primaries prim)
     case PL_COLOR_PRIM_COUNT: break;
     }
 
-    abort();
+    pl_unreachable();
 }
 
 enum pl_color_primaries pl_color_primaries_guess(int width, int height)
@@ -230,7 +236,10 @@ float pl_color_transfer_nominal_peak(enum pl_color_transfer trc)
     case PL_COLOR_TRC_SRGB:
     case PL_COLOR_TRC_LINEAR:
     case PL_COLOR_TRC_GAMMA18:
+    case PL_COLOR_TRC_GAMMA20:
     case PL_COLOR_TRC_GAMMA22:
+    case PL_COLOR_TRC_GAMMA24:
+    case PL_COLOR_TRC_GAMMA26:
     case PL_COLOR_TRC_GAMMA28:
     case PL_COLOR_TRC_PRO_PHOTO:
         return 1.0;
@@ -242,7 +251,7 @@ float pl_color_transfer_nominal_peak(enum pl_color_transfer trc)
     case PL_COLOR_TRC_COUNT: break;
     }
 
-    abort();
+    pl_unreachable();
 }
 
 bool pl_color_light_is_scene_referred(enum pl_color_light light)
@@ -258,7 +267,45 @@ bool pl_color_light_is_scene_referred(enum pl_color_light light)
     case PL_COLOR_LIGHT_COUNT: break;
     }
 
-    abort();
+    pl_unreachable();
+}
+
+const struct pl_hdr_metadata pl_hdr_metadata_empty = {0};
+const struct pl_hdr_metadata pl_hdr_metadata_hdr10 ={
+    .prim = {
+        .red   = {0.708,    0.292},
+        .green = {0.170,    0.797},
+        .blue  = {0.131,    0.046},
+        .white = {0.31271,  0.32902},
+    },
+    .min_luma = 0,
+    .max_luma = 10000,
+    .max_cll  = 10000,
+    .max_fall = 0, // unknown
+};
+
+bool pl_hdr_metadata_equal(const struct pl_hdr_metadata *a,
+                           const struct pl_hdr_metadata *b)
+{
+    return pl_raw_primaries_equal(&a->prim, &b->prim) &&
+           a->min_luma == b->min_luma &&
+           a->max_luma == b->max_luma &&
+           a->max_cll  == b->max_cll  &&
+           a->max_fall == b->max_fall;
+}
+
+void pl_hdr_metadata_merge(struct pl_hdr_metadata *orig,
+                           const struct pl_hdr_metadata *update)
+{
+    pl_raw_primaries_merge(&orig->prim, &update->prim);
+    if (!orig->min_luma)
+        orig->min_luma = update->min_luma;
+    if (!orig->max_luma)
+        orig->max_luma = update->max_luma;
+    if (!orig->max_cll)
+        orig->max_cll = update->max_cll;
+    if (!orig->max_fall)
+        orig->max_fall = update->max_fall;
 }
 
 const struct pl_color_space pl_color_space_unknown = {0};
@@ -266,39 +313,62 @@ const struct pl_color_space pl_color_space_unknown = {0};
 const struct pl_color_space pl_color_space_srgb = {
     .primaries = PL_COLOR_PRIM_BT_709,
     .transfer  = PL_COLOR_TRC_SRGB,
-    .light     = PL_COLOR_LIGHT_DISPLAY,
 };
 
 const struct pl_color_space pl_color_space_bt709 = {
     .primaries = PL_COLOR_PRIM_BT_709,
     .transfer  = PL_COLOR_TRC_BT_1886,
-    .light     = PL_COLOR_LIGHT_DISPLAY,
 };
 
 const struct pl_color_space pl_color_space_hdr10 = {
     .primaries = PL_COLOR_PRIM_BT_2020,
     .transfer  = PL_COLOR_TRC_PQ,
-    .light     = PL_COLOR_LIGHT_DISPLAY,
 };
 
 const struct pl_color_space pl_color_space_bt2020_hlg = {
     .primaries = PL_COLOR_PRIM_BT_2020,
     .transfer  = PL_COLOR_TRC_HLG,
-    .light     = PL_COLOR_LIGHT_SCENE_HLG,
 };
 
 const struct pl_color_space pl_color_space_monitor = {
     .primaries = PL_COLOR_PRIM_BT_709, // sRGB primaries
-    .transfer  = PL_COLOR_TRC_GAMMA22, // typical response
-    .light     = PL_COLOR_LIGHT_DISPLAY,
+    .transfer  = PL_COLOR_TRC_UNKNOWN, // unknown SDR response
 };
 
-bool pl_color_space_is_hdr(struct pl_color_space csp)
+bool pl_color_space_is_hdr(const struct pl_color_space *csp)
 {
-    float peak = pl_color_transfer_nominal_peak(csp.transfer);
-    peak *= PL_DEF(csp.sig_scale, 1.0);
+    return csp->hdr.max_luma > PL_COLOR_SDR_WHITE ||
+           csp->sig_scale > 1 ||
+           pl_color_transfer_is_hdr(csp->transfer);
+}
 
-    return peak > 1.0;
+bool pl_color_space_is_black_scaled(const struct pl_color_space *csp)
+{
+    switch (csp->transfer) {
+    case PL_COLOR_TRC_UNKNOWN:
+    case PL_COLOR_TRC_SRGB:
+    case PL_COLOR_TRC_LINEAR:
+    case PL_COLOR_TRC_GAMMA18:
+    case PL_COLOR_TRC_GAMMA20:
+    case PL_COLOR_TRC_GAMMA22:
+    case PL_COLOR_TRC_GAMMA24:
+    case PL_COLOR_TRC_GAMMA26:
+    case PL_COLOR_TRC_GAMMA28:
+    case PL_COLOR_TRC_PRO_PHOTO:
+    case PL_COLOR_TRC_HLG:
+        return true;
+
+    case PL_COLOR_TRC_BT_1886:
+    case PL_COLOR_TRC_PQ:
+    case PL_COLOR_TRC_V_LOG:
+    case PL_COLOR_TRC_S_LOG1:
+    case PL_COLOR_TRC_S_LOG2:
+        return false;
+
+    case PL_COLOR_TRC_COUNT: break;
+    }
+
+    pl_unreachable();
 }
 
 void pl_color_space_merge(struct pl_color_space *orig,
@@ -308,14 +378,7 @@ void pl_color_space_merge(struct pl_color_space *orig,
         orig->primaries = new->primaries;
     if (!orig->transfer)
         orig->transfer = new->transfer;
-    if (!orig->light)
-        orig->light = new->light;
-    if (!orig->sig_peak)
-        orig->sig_peak = new->sig_peak;
-    if (!orig->sig_avg)
-        orig->sig_avg = new->sig_avg;
-    if (!orig->sig_scale)
-        orig->sig_scale = new->sig_scale;
+    pl_hdr_metadata_merge(&orig->hdr, &new->hdr);
 }
 
 bool pl_color_space_equal(const struct pl_color_space *c1,
@@ -323,44 +386,83 @@ bool pl_color_space_equal(const struct pl_color_space *c1,
 {
     return c1->primaries == c2->primaries &&
            c1->transfer  == c2->transfer &&
-           c1->light     == c2->light &&
-           c1->sig_peak  == c2->sig_peak &&
-           c1->sig_avg   == c2->sig_avg &&
-           c1->sig_scale == c2->sig_scale;
+           pl_hdr_metadata_equal(&c1->hdr, &c2->hdr);
 }
 
-// Average light level for SDR signals. This is equal to a signal level of 0.5
-// under a typical presentation gamma of about 2.0.
-static const float sdr_avg = 0.25;
-
 void pl_color_space_infer(struct pl_color_space *space)
 {
     if (!space->primaries)
         space->primaries = PL_COLOR_PRIM_BT_709;
     if (!space->transfer)
-        space->transfer = PL_COLOR_TRC_GAMMA22;
-    if (!space->light) {
-        space->light = (space->transfer == PL_COLOR_TRC_HLG)
-            ? PL_COLOR_LIGHT_SCENE_HLG
-            : PL_COLOR_LIGHT_DISPLAY;
-    }
-    if (!space->sig_peak) {
-        space->sig_peak = pl_color_transfer_nominal_peak(space->transfer);
+        space->transfer = PL_COLOR_TRC_BT_1886;
 
-        // Exception: For HLG content, we want to infer a value of 1000 cd/m²
-        // (corresponding to a peak of 10.0) instead of the true nominal peak
-        // of 12.0. A peak of 1000 is considered the "reference" HLG display.
+    // Backwards-compatibility with deprecated fields
+    if (space->sig_peak) {
+        space->hdr.max_luma = space->sig_peak * PL_COLOR_SDR_WHITE;
+        space->sig_peak = 0;
+    }
+    if (space->sig_floor) {
+        space->hdr.min_luma = space->sig_floor * PL_COLOR_SDR_WHITE;
+        space->sig_floor = 0;
+    }
+
+    if (space->hdr.max_luma < space->hdr.min_luma) // sanity
+        space->hdr.max_luma = space->hdr.min_luma = 0;
+
+    if (space->hdr.max_luma < 1 || space->hdr.max_luma > 10000) {
+        space->hdr.max_luma = pl_color_transfer_nominal_peak(space->transfer)
+                              * PL_COLOR_SDR_WHITE;
+
+        // Exception: For HLG content, we want to infer a value of 1000 cd/m²,
+        // a value which is considered the "reference" HLG display.
         if (space->transfer == PL_COLOR_TRC_HLG)
-            space->sig_peak = 10.0;
+            space->hdr.max_luma = 1000;
     }
 
-    if (!space->sig_scale)
-        space->sig_scale = 1.0;
+    if (space->hdr.min_luma <= 0 || space->hdr.min_luma > 100) {
+        if (pl_color_transfer_is_hdr(space->transfer)) {
+            space->hdr.min_luma = 0.0050f; // Typical HDR black
+        } else {
+            space->hdr.min_luma = space->hdr.max_luma / 1000; // Typical SDR contrast
+        }
+    }
 
-    // In theory, for HDR signals, this is typically no longer true - but
-    // without adequate metadata there's not much else we can assume
-    if (!space->sig_avg)
-        space->sig_avg = sdr_avg / space->sig_scale;
+    if (space->sig_scale && !pl_color_transfer_is_hdr(space->transfer)) {
+        space->hdr.max_luma *= space->sig_scale;
+        space->hdr.min_luma *= space->sig_scale;
+        space->sig_scale = 0;
+    }
+
+    // Default the signal color space based on the nominal raw primaries
+    pl_raw_primaries_merge(&space->hdr.prim, pl_raw_primaries_get(space->primaries));
+}
+
+void pl_color_space_infer_ref(struct pl_color_space *space,
+                              const struct pl_color_space *refp)
+{
+    struct pl_color_space ref = *refp;
+    pl_color_space_infer(&ref);
+
+    if (!space->primaries) {
+        if (pl_color_primaries_is_wide_gamut(ref.primaries)) {
+            space->primaries = PL_COLOR_PRIM_BT_709;
+        } else {
+            space->primaries = ref.primaries;
+        }
+    }
+
+    if (!space->transfer) {
+        if (pl_color_transfer_is_hdr(ref.transfer)) {
+            space->transfer = PL_COLOR_TRC_BT_1886;
+        } else if (ref.transfer == PL_COLOR_TRC_LINEAR) {
+            space->transfer = PL_COLOR_TRC_GAMMA22;
+        } else {
+            space->transfer = ref.transfer;
+        }
+    }
+
+    // Infer the remaining fields after making the above choices
+    pl_color_space_infer(space);
 }
 
 const struct pl_color_adjustment pl_color_adjustment_neutral = {
@@ -422,6 +524,29 @@ struct pl_cie_xy pl_white_from_temp(float temp)
     };
 }
 
+bool pl_raw_primaries_equal(const struct pl_raw_primaries *a,
+                            const struct pl_raw_primaries *b)
+{
+    return pl_cie_xy_equal(&a->red,   &b->red)   &&
+           pl_cie_xy_equal(&a->green, &b->green) &&
+           pl_cie_xy_equal(&a->blue,  &b->blue)  &&
+           pl_cie_xy_equal(&a->white, &b->white);
+}
+
+void pl_raw_primaries_merge(struct pl_raw_primaries *orig,
+                            const struct pl_raw_primaries *update)
+{
+    union {
+        struct pl_raw_primaries prim;
+        float raw[8];
+    } *pa = (void *) orig,
+      *pb = (void *) update;
+
+    pl_static_assert(sizeof(*pa) == sizeof(*orig));
+    for (int i = 0; i < PL_ARRAY_SIZE(pa->raw); i++)
+        pa->raw[i] = PL_DEF(pa->raw[i], pb->raw[i]);
+}
+
 const struct pl_raw_primaries *pl_raw_primaries_get(enum pl_color_primaries prim)
 {
     /*
@@ -793,7 +918,8 @@ struct pl_matrix3x3 pl_get_cone_matrix(const struct pl_cone_params *params,
         break;
     }
 
-    default: abort();
+    default:
+        pl_unreachable();
     }
 
     // out := RGB<-LMS * distort * LMS<-RGB
@@ -923,7 +1049,7 @@ struct pl_transform3x3 pl_color_repr_decode(struct pl_color_repr *repr,
         m = (struct pl_matrix3x3) {{
             {0, 0, 1},
             {1, 0, 0},
-            {0, 1, 0}
+            {0, 1, 0},
         }};
         break;
     case PL_COLOR_SYSTEM_BT_2100_PQ: {
@@ -948,6 +1074,9 @@ struct pl_transform3x3 pl_color_repr_decode(struct pl_color_repr *repr,
         }};
         break;
     }
+    case PL_COLOR_SYSTEM_DOLBYVISION:
+        m = repr->dovi->nonlinear;
+        break;
     case PL_COLOR_SYSTEM_YCGCO:
         m = (struct pl_matrix3x3) {{
             {1,  -1,  1},
@@ -965,7 +1094,8 @@ struct pl_transform3x3 pl_color_repr_decode(struct pl_color_repr *repr,
         m = pl_get_xyz2rgb_matrix(pl_raw_primaries_get(PL_COLOR_PRIM_BT_709));
         break;
     }
-    default: abort();
+    case PL_COLOR_SYSTEM_COUNT:
+        pl_unreachable();
     }
 
     // Apply hue and saturation in the correct way depending on the colorspace.
@@ -1014,7 +1144,8 @@ struct pl_transform3x3 pl_color_repr_decode(struct pl_color_repr *repr,
         cmax = 1.0;
         cmid = 128 / 256. * scale; // *not* exactly 0.5
         break;
-    default: abort();
+    default:
+        pl_unreachable();
     }
 
     double ymul = 1.0 / (ymax - ymin);
@@ -1023,7 +1154,14 @@ struct pl_transform3x3 pl_color_repr_decode(struct pl_color_repr *repr,
     double mul[3]   = { ymul, ymul, ymul };
     double black[3] = { ymin, ymin, ymin };
 
-    if (pl_color_system_is_ycbcr_like(repr->sys)) {
+    if (repr->sys == PL_COLOR_SYSTEM_DOLBYVISION) {
+        // The RPU matrix already includes levels normalization, but in this
+        // case we also have to respect the signalled color offsets
+        for (int i = 0; i < 3; i++) {
+            mul[i] = 1.0;
+            black[i] = repr->dovi->nonlinear_offset[i] * scale;
+        }
+    } else if (pl_color_system_is_ycbcr_like(repr->sys)) {
         mul[1]   = mul[2]   = cmul;
         black[1] = black[2] = cmid;
     }
