@@ -18,38 +18,69 @@
 #pragma once
 
 #define __STDC_FORMAT_MACROS
-#include <stddef.h>
-#include <stdlib.h>
+#include <stdatomic.h>
 #include <stdbool.h>
+#include <stddef.h>
 #include <stdint.h>
+#include <stdlib.h>
 #include <inttypes.h>
 
+#if defined(__MINGW32__) && !defined(__clang__)
+#define PL_PRINTF(fmt, va) __attribute__ ((format(gnu_printf, fmt, va)))
+#elif defined(__GNUC__)
+#define PL_PRINTF(fmt, va) __attribute__ ((format(printf, fmt, va)))
+#else
+#define PL_PRINTF(fmt, va)
+#endif
+
+#ifdef __unix__
+#define PL_HAVE_UNIX
+#endif
+
+#ifdef _WIN32
+#define PL_HAVE_WIN32
+#endif
+
 #include "config_internal.h"
 #include "pl_assert.h"
-
-#include <xtalloc.h>
-#include <bstr.h>
+#include "pl_alloc.h"
+#include "pl_string.h"
 
 // Include all of the symbols that should be public in a way that marks them
 // as being externally visible. (Otherwise, all symbols are hidden by default)
 #pragma GCC visibility push(default)
 
-#include "config.h"
-#include "include/libplacebo/colorspace.h"
-#include "include/libplacebo/common.h"
-#include "include/libplacebo/context.h"
-#include "include/libplacebo/dispatch.h"
-#include "include/libplacebo/dither.h"
-#include "include/libplacebo/dummy.h"
-#include "include/libplacebo/filters.h"
-#include "include/libplacebo/gpu.h"
-#include "include/libplacebo/renderer.h"
-#include "include/libplacebo/shaders.h"
-#include "include/libplacebo/shaders/av1.h"
-#include "include/libplacebo/shaders/colorspace.h"
-#include "include/libplacebo/shaders/sampling.h"
-#include "include/libplacebo/swapchain.h"
-#include "include/libplacebo/utils/upload.h"
+#include <libplacebo/config.h>
+#undef PL_DEPRECATED
+#define PL_DEPRECATED
+
+#if PL_API_VER != BUILD_API_VER
+#error Header mismatch? <libplacebo/config.h> pulled from elsewhere!
+#endif
+
+#include <libplacebo/colorspace.h>
+#include <libplacebo/common.h>
+#include <libplacebo/log.h>
+#include <libplacebo/dispatch.h>
+#include <libplacebo/dither.h>
+#include <libplacebo/dummy.h>
+#include <libplacebo/filters.h>
+#include <libplacebo/gpu.h>
+#include <libplacebo/renderer.h>
+#include <libplacebo/shaders.h>
+#include <libplacebo/shaders/colorspace.h>
+#include <libplacebo/shaders/custom.h>
+#include <libplacebo/shaders/film_grain.h>
+#include <libplacebo/shaders/lut.h>
+#include <libplacebo/shaders/sampling.h>
+#include <libplacebo/swapchain.h>
+#include <libplacebo/tone_mapping.h>
+#include <libplacebo/utils/frame_queue.h>
+#include <libplacebo/utils/upload.h>
+
+#ifdef PL_HAVE_LCMS
+#include "include/libplacebo/shaders/icc.h"
+#endif
 
 #ifdef PL_HAVE_VULKAN
 #include "include/libplacebo/vulkan.h"
@@ -59,6 +90,10 @@
 #include "include/libplacebo/opengl.h"
 #endif
 
+#ifdef PL_HAVE_D3D11
+#include "include/libplacebo/d3d11.h"
+#endif
+
 #pragma GCC visibility pop
 
 // Align up to the nearest multiple of an arbitrary alignment, which may also
@@ -71,6 +106,9 @@
 // Returns the log base 2 of an unsigned long long
 #define PL_LOG2(x) ((unsigned) (8*sizeof (unsigned long long) - __builtin_clzll((x)) - 1))
 
+// Rounds a number up to the nearest power of two
+#define PL_ALIGN_POT(x) (0x1LLU << (PL_LOG2((x) - 1) + 1))
+
 // Returns whether or not a number is a power of two (or zero)
 #define PL_ISPOT(x) (((x) & ((x) - 1)) == 0)
 
@@ -106,10 +144,12 @@ static inline float *pl_transpose(int dim, float *out, const float *in)
 // Helper functions for some common numeric operations (careful: double-eval)
 #define PL_MAX(x, y) ((x) > (y) ? (x) : (y))
 #define PL_MIN(x, y) ((x) < (y) ? (x) : (y))
+#define PL_CLAMP(x, l, h) ((x) < (l) ? (l) : (x) > (h) ? (h) : (x))
 #define PL_CMP(a, b) ((a) < (b) ? -1 : (a) > (b) ? 1 : 0)
 #define PL_DEF(x, d) ((x) ? (x) : (d))
 #define PL_SQUARE(x) ((x) * (x))
 #define PL_CUBE(x) ((x) * (x) * (x))
+#define PL_MIX(a, b, x) ((x) * (b) + (1 - (x)) * (a))
 
 // Helpers for doing alignment calculations
 static inline size_t pl_gcd(size_t x, size_t y)
@@ -129,3 +169,23 @@ static inline size_t pl_lcm(size_t x, size_t y)
     assert(x && y);
     return x * (y / pl_gcd(x, y));
 }
+
+// Error checking macro for stuff with integer errors, aborts on failure
+#define PL_CHECK_ERR(expr)                                                      \
+    do {                                                                        \
+        int _ret = (expr);                                                      \
+        if (_ret) {                                                             \
+            fprintf(stderr, "libplacebo: internal error: %s (%s:%d)\n",         \
+                    strerror(_ret), __FILE__, __LINE__);                        \
+            abort();                                                            \
+        }                                                                       \
+    } while (0)
+
+// Refcounting helpers
+typedef _Atomic uint32_t pl_rc_t;
+#define pl_rc_init(rc)  atomic_init(rc, 1)
+#define pl_rc_ref(rc)   ((void) atomic_fetch_add_explicit(rc, 1, memory_order_acquire))
+#define pl_rc_deref(rc) (atomic_fetch_sub_explicit(rc, 1, memory_order_release) == 1)
+#define pl_rc_count(rc)  atomic_load(rc)
+
+#define pl_unreachable() (assert(!"unreachable"), __builtin_unreachable())
