@@ -37,7 +37,8 @@
 #include <math.h>
 
 #include "common.h"
-#include "context.h"
+#include "filters.h"
+#include "log.h"
 
 bool pl_filter_function_eq(const struct pl_filter_function *a,
                            const struct pl_filter_function *b)
@@ -135,16 +136,15 @@ static struct pl_filter_function *dupfilter(void *alloc,
     return f ? pl_memdup(alloc, (void *)f, sizeof(*f)) : NULL;
 }
 
-const struct pl_filter *pl_filter_generate(struct pl_context *ctx,
-                                       const struct pl_filter_params *params)
+pl_filter pl_filter_generate(pl_log log, const struct pl_filter_params *params)
 {
     pl_assert(params);
     if (params->lut_entries <= 0 || !params->config.kernel) {
-        pl_fatal(ctx, "Invalid params: missing lut_entries or config.kernel");
+        pl_fatal(log, "Invalid params: missing lut_entries or config.kernel");
         return NULL;
     }
 
-    struct pl_filter *f = pl_zalloc_ptr(ctx, f);
+    struct pl_filter *f = pl_zalloc_ptr(NULL, f);
     f->params = *params;
     f->params.config.kernel = dupfilter(f, params->config.kernel);
     f->params.config.window = dupfilter(f, params->config.window);
@@ -170,7 +170,7 @@ const struct pl_filter *pl_filter_generate(struct pl_context *ctx,
         // Pick the most appropriate row size
         f->row_size = ceil(f->radius) * 2;
         if (params->max_row_size && f->row_size > params->max_row_size) {
-            pl_info(ctx, "Required filter size %d exceeds the maximum allowed "
+            pl_info(log, "Required filter size %d exceeds the maximum allowed "
                     "size of %d. This may result in adverse effects (aliasing, "
                     "or moiré artifacts).", f->row_size, params->max_row_size);
             f->row_size = params->max_row_size;
@@ -190,7 +190,7 @@ const struct pl_filter *pl_filter_generate(struct pl_context *ctx,
     return f;
 }
 
-void pl_filter_free(const struct pl_filter **filter)
+void pl_filter_free(pl_filter *filter)
 {
     pl_free_ptr((void **) filter);
 }
@@ -229,7 +229,6 @@ static double box(const struct pl_filter_function *f, double x)
 }
 
 const struct pl_filter_function pl_filter_function_box = {
-    .resizable = true,
     .weight    = box,
     .radius    = 1.0,
 };
@@ -245,6 +244,16 @@ const struct pl_filter_function pl_filter_function_triangle = {
     .radius    = 1.0,
 };
 
+static double cosine(const struct pl_filter_function *f, double x)
+{
+    return cos(x);
+}
+
+const struct pl_filter_function pl_filter_function_cosine = {
+    .weight = cosine,
+    .radius = M_PI / 2.0,
+};
+
 static double hann(const struct pl_filter_function *f, double x)
 {
     return 0.5 + 0.5 * cos(M_PI * x);
@@ -317,6 +326,17 @@ const struct pl_filter_function pl_filter_function_blackman = {
     .params  = {0.16},
 };
 
+static double bohman(const struct pl_filter_function *f, double x)
+{
+    double pix = M_PI * x;
+    return (1.0 - x) * cos(pix) + sin(pix) / M_PI;
+}
+
+const struct pl_filter_function pl_filter_function_bohman = {
+    .weight = bohman,
+    .radius = 1.0,
+};
+
 static double gaussian(const struct pl_filter_function *f, double x)
 {
     return exp(-2.0 * x * x / f->params[0]);
@@ -330,6 +350,20 @@ const struct pl_filter_function pl_filter_function_gaussian = {
     .params    = {1.0},
 };
 
+static double quadratic(const struct pl_filter_function *f, double x)
+{
+    if (x < 0.5) {
+        return 0.75 - x * x;
+    } else {
+        return 0.5 * (x - 1.5) * (x - 1.5);
+    }
+}
+
+const struct pl_filter_function pl_filter_function_quadratic = {
+    .weight = quadratic,
+    .radius = 1.5,
+};
+
 static double sinc(const struct pl_filter_function *f, double x)
 {
     if (x < 1e-8)
@@ -497,13 +531,17 @@ const struct pl_filter_function_preset pl_filter_function_presets[] = {
     {"box",             &pl_filter_function_box},
     {"dirichlet",       &pl_filter_function_box}, // alias
     {"triangle",        &pl_filter_function_triangle},
+    {"cosine",          &pl_filter_function_cosine},
     {"hann",            &pl_filter_function_hann},
     {"hanning",         &pl_filter_function_hann}, // alias
     {"hamming",         &pl_filter_function_hamming},
     {"welch",           &pl_filter_function_welch},
     {"kaiser",          &pl_filter_function_kaiser},
     {"blackman",        &pl_filter_function_blackman},
+    {"bohman",          &pl_filter_function_bohman},
     {"gaussian",        &pl_filter_function_gaussian},
+    {"quadratic",       &pl_filter_function_quadratic},
+    {"quadric",         &pl_filter_function_quadratic}, // alias
     {"sinc",            &pl_filter_function_sinc},
     {"jinc",            &pl_filter_function_jinc},
     {"sphinx",          &pl_filter_function_sphinx},
@@ -598,6 +636,16 @@ const struct pl_filter_config pl_filter_ewa_hann = {
     .polar = true,
 };
 
+const struct pl_filter_config pl_filter_haasnsoft = {
+    .kernel = &jinc3,
+    .window = &pl_filter_function_hann,
+    // The blur is tuned to equal out orthogonal and diagonal contributions
+    // on a regular grid. This has the effect of almost completely killing
+    // aliasing.
+    .blur = 1.11,
+    .polar = true,
+};
+
 // Spline family
 const struct pl_filter_config pl_filter_bicubic = {
     .kernel = &pl_filter_function_bicubic,
@@ -636,36 +684,9 @@ const struct pl_filter_config pl_filter_ewa_robidouxsharp = {
 
 // Named filter configs
 const struct pl_filter_preset pl_filter_presets[] = {
-    // Highest priority / recommended filters
     {"none",                NULL,                   "Built-in sampling"},
-    {"bilinear",            &pl_filter_bilinear,    "Bilinear"},
-    {"nearest",             &pl_filter_nearest,     "Nearest neighbour"},
-    {"bicubic",             &pl_filter_bicubic,     "Bicubic"},
-    {"lanczos",             &pl_filter_lanczos,     "Lanczos"},
-    {"ewa_lanczos",         &pl_filter_ewa_lanczos, "Jinc (EWA Lanczos)"},
-    {"gaussian",            &pl_filter_gaussian,    "Gaussian"},
-    {"spline16",            &pl_filter_spline16,    "Spline (2 taps)"},
-    {"spline36",            &pl_filter_spline36,    "Spline (3 taps)"},
-    {"spline64",            &pl_filter_spline64,    "Spline (4 taps)"},
-    {"mitchell",            &pl_filter_mitchell,    "Mitchell-Netravali"},
-
-    // Remaining filters
-    {"sinc",                &pl_filter_sinc,        "Sinc (unwindowed)"},
-    {"ginseng",             &pl_filter_ginseng,     "Ginseng (Jinc-Sinc)"},
-    {"ewa_jinc",            &pl_filter_ewa_jinc,    "EWA Jinc (unwindowed)"},
-    {"ewa_ginseng",         &pl_filter_ewa_ginseng, "EWA Ginseng"},
-    {"ewa_hann",            &pl_filter_ewa_hann,    "EWA Hann"},
-    {"catmull_rom",         &pl_filter_catmull_rom, "Catmull-Rom"},
-    {"robidoux",            &pl_filter_robidoux,            "Robidoux"},
-    {"robidouxsharp",       &pl_filter_robidouxsharp,       "RobidouxSharp"},
-    {"ewa_robidoux",        &pl_filter_ewa_robidoux,        "EWA Robidoux"},
-    {"ewa_robidouxsharp",   &pl_filter_ewa_robidouxsharp,   "EWA RobidouxSharp"},
-
-    // Aliases
-    {"box",                 &pl_filter_nearest}, // alias
-    {"triangle",            &pl_filter_bilinear}, // alias
-    {"ewa_hanning",         &pl_filter_ewa_hann}, // alias
-    {0},
+    COMMON_FILTER_PRESETS,
+    {0}
 };
 
 const int pl_num_filter_presets = PL_ARRAY_SIZE(pl_filter_presets) - 1;
