@@ -24,7 +24,8 @@
 extern "C" {
 #endif
 
-int pl_glslang_version(void);
+#include "utils.h"
+
 bool pl_glslang_init(void);
 void pl_glslang_uninit(void);
 
@@ -38,16 +39,11 @@ struct pl_glslang_res {
     size_t size;
 };
 
-enum pl_glslang_stage {
-    PL_GLSLANG_VERTEX,
-    PL_GLSLANG_FRAGMENT,
-    PL_GLSLANG_COMPUTE,
-};
-
 // Compile GLSL into a SPIRV stream, if possible. The resulting
-// pl_glslang_res can simply be freed with talloc_free() when done.
-struct pl_glslang_res *pl_glslang_compile(const char *glsl, uint32_t api_ver,
-                                          enum pl_glslang_stage stage);
+// pl_glslang_res can simply be freed with pl_free() when done.
+struct pl_glslang_res *pl_glslang_compile(const struct pl_glsl_version *glsl,
+                                          enum glsl_shader_stage stage,
+                                          const char *shader);
 
 #ifdef __cplusplus
 }
