@@ -19,9 +19,13 @@
 #define LIBPLACEBO_COLORSPACE_H_
 
 #include <stdbool.h>
+#include <stddef.h>
 #include <stdint.h>
+
 #include <libplacebo/common.h>
 
+PL_API_BEGIN
+
 // The underlying color representation (e.g. RGB, XYZ or YCbCr)
 enum pl_color_system {
     PL_COLOR_SYSTEM_UNKNOWN = 0,
@@ -33,6 +37,7 @@ enum pl_color_system {
     PL_COLOR_SYSTEM_BT_2020_C,   // ITU-R Rec. BT.2020 (constant luminance)
     PL_COLOR_SYSTEM_BT_2100_PQ,  // ITU-R Rec. BT.2100 ICtCp PQ variant
     PL_COLOR_SYSTEM_BT_2100_HLG, // ITU-R Rec. BT.2100 ICtCp HLG variant
+    PL_COLOR_SYSTEM_DOLBYVISION, // Dolby Vision (see pl_dovi_metadata)
     PL_COLOR_SYSTEM_YCGCO,       // YCgCo (derived from RGB)
     // Other color systems:
     PL_COLOR_SYSTEM_RGB,         // Red, Green and Blue
@@ -44,7 +49,7 @@ bool pl_color_system_is_ycbcr_like(enum pl_color_system sys);
 
 // Returns true for color systems that are linear transformations of the RGB
 // equivalent, i.e. are simple matrix multiplications. For color systems with
-// this property, pl_get_decoding_matrix is sufficient for conversion to RGB.
+// this property, `pl_color_repr_decode` is sufficient for conversion to RGB.
 bool pl_color_system_is_linear(enum pl_color_system sys);
 
 // Guesses the best YCbCr-like colorspace based on a image given resolution.
@@ -74,9 +79,13 @@ enum pl_channel {
 // The numerical range of the representation (where applicable).
 enum pl_color_levels {
     PL_COLOR_LEVELS_UNKNOWN = 0,
-    PL_COLOR_LEVELS_TV,         // TV range, e.g. 16-235
-    PL_COLOR_LEVELS_PC,         // PC range, e.g. 0-255
+    PL_COLOR_LEVELS_LIMITED,    // Limited/TV range, e.g. 16-235
+    PL_COLOR_LEVELS_FULL,       // Full/PC range, e.g. 0-255
     PL_COLOR_LEVELS_COUNT,
+
+    // Compatibility aliases
+    PL_COLOR_LEVELS_TV = PL_COLOR_LEVELS_LIMITED,
+    PL_COLOR_LEVELS_PC = PL_COLOR_LEVELS_FULL,
 };
 
 // The alpha representation mode.
@@ -84,6 +93,7 @@ enum pl_alpha_mode {
     PL_ALPHA_UNKNOWN = 0,   // or no alpha channel present
     PL_ALPHA_INDEPENDENT,   // alpha channel is separate from the video
     PL_ALPHA_PREMULTIPLIED, // alpha channel is multiplied into the colors
+    PL_ALPHA_MODE_COUNT,
 };
 
 // The underlying bit-wise representation of a color sample. For example,
@@ -111,6 +121,26 @@ struct pl_bit_encoding {
 bool pl_bit_encoding_equal(const struct pl_bit_encoding *b1,
                            const struct pl_bit_encoding *b2);
 
+// Parsed metadata from the Dolby Vision RPU
+struct pl_dovi_metadata {
+    // Colorspace transformation metadata
+    float nonlinear_offset[3];      // input offset ("ycc_to_rgb_offset")
+    struct pl_matrix3x3 nonlinear;  // before PQ, also called "ycc_to_rgb"
+    struct pl_matrix3x3 linear;     // after PQ, also called "rgb_to_lms"
+
+    // Reshape data, grouped by component
+    struct pl_reshape_data {
+        uint8_t num_pivots;
+        float pivots[9]; // normalized to [0.0, 1.0] based on BL bit depth
+        uint8_t method[8]; // 0 = polynomial, 1 = MMR
+        // Note: these must be normalized (divide by coefficient_log2_denom)
+        float poly_coeffs[8][3]; // x^0, x^1, x^2, unused must be 0
+        uint8_t mmr_order[8]; // 1, 2 or 3
+        float mmr_constant[8];
+        float mmr_coeffs[8][3 /* order */][7];
+    } comp[3];
+};
+
 // Struct describing the underlying color system and representation. This
 // information is needed to convert an encoded color to a normalized RGB triple
 // in the range 0-1.
@@ -119,6 +149,14 @@ struct pl_color_repr {
     enum pl_color_levels levels;
     enum pl_alpha_mode alpha;
     struct pl_bit_encoding bits; // or {0} if unknown
+
+    // Metadata for PL_COLOR_SYSTEM_DOLBYVISION. Note that, for the sake of
+    // efficiency, this is treated purely as an opaque reference - functions
+    // like pl_color_repr_equal will merely do a pointer equality test.
+    //
+    // The only functions that actually dereference it in any way are
+    // pl_color_repr_decode,  pl_shader_decode_color and pl_render_image(_mix).
+    const struct pl_dovi_metadata *dovi;
 };
 
 // Some common color representations. It's worth pointing out that all of these
@@ -137,7 +175,7 @@ bool pl_color_repr_equal(const struct pl_color_repr *c1,
 
 // Replaces unknown values in the first struct by those of the second struct.
 void pl_color_repr_merge(struct pl_color_repr *orig,
-                         const struct pl_color_repr *new);
+                         const struct pl_color_repr *update);
 
 // This function normalizes the color representation such that
 // color_depth=sample_depth and bit_shift=0; and returns the scaling factor
@@ -159,6 +197,7 @@ enum pl_color_primaries {
     PL_COLOR_PRIM_BT_601_625,   // ITU-R Rec. BT.601 (625-line = PAL, SECAM)
     PL_COLOR_PRIM_BT_709,       // ITU-R Rec. BT.709 (HD), also sRGB
     PL_COLOR_PRIM_BT_470M,      // ITU-R Rec. BT.470 M
+    PL_COLOR_PRIM_EBU_3213,     // EBU Tech. 3213-E / JEDEC P22 phosphors
     // Wide gamut:
     PL_COLOR_PRIM_BT_2020,      // ITU-R Rec. BT.2020 (UltraHD)
     PL_COLOR_PRIM_APPLE,        // Apple RGB
@@ -169,6 +208,7 @@ enum pl_color_primaries {
     PL_COLOR_PRIM_DISPLAY_P3,   // DCI-P3 (Digital Cinema) with D65 white point
     PL_COLOR_PRIM_V_GAMUT,      // Panasonic V-Gamut (VARICAM)
     PL_COLOR_PRIM_S_GAMUT,      // Sony S-Gamut
+    PL_COLOR_PRIM_FILM_C,       // Traditional film primaries with Illuminant C
     PL_COLOR_PRIM_COUNT
 };
 
@@ -187,7 +227,10 @@ enum pl_color_transfer {
     PL_COLOR_TRC_SRGB,          // IEC 61966-2-4 sRGB (CRT emulation)
     PL_COLOR_TRC_LINEAR,        // Linear light content
     PL_COLOR_TRC_GAMMA18,       // Pure power gamma 1.8
+    PL_COLOR_TRC_GAMMA20,       // Pure power gamma 2.0
     PL_COLOR_TRC_GAMMA22,       // Pure power gamma 2.2
+    PL_COLOR_TRC_GAMMA24,       // Pure power gamma 2.4
+    PL_COLOR_TRC_GAMMA26,       // Pure power gamma 2.6
     PL_COLOR_TRC_GAMMA28,       // Pure power gamma 2.8
     PL_COLOR_TRC_PRO_PHOTO,     // ProPhoto RGB (ROMM)
     // High dynamic range:
@@ -205,9 +248,10 @@ enum pl_color_transfer {
 //
 // Note: This returns the highest encodable signal by definition of the EOTF,
 // regardless of the ultimate representation (e.g. scene or display referred).
-// For HLG in particular, this is always 12.0 - which is potentially different
-// from the signal peak after applying the OOTF to go from scene referred to
-// display referred.
+// For HLG in particular, this is always around 3.77 - which is potentially
+// different from the signal peak after applying the OOTF to go from scene
+// referred to display referred (resulting in a display-referred peak of around
+// 4.92 for a 1000 cd/m^2 HLG reference display).
 float pl_color_transfer_nominal_peak(enum pl_color_transfer trc);
 
 static inline bool pl_color_transfer_is_hdr(enum pl_color_transfer trc)
@@ -215,14 +259,83 @@ static inline bool pl_color_transfer_is_hdr(enum pl_color_transfer trc)
     return pl_color_transfer_nominal_peak(trc) > 1.0;
 }
 
-// This defines the standard reference white level (in cd/m^2) that is assumed
-// throughout standards such as those from by ITU-R, EBU, etc.
-// This is particularly relevant for HDR conversions, as this value is used
-// as a reference for conversions between absolute transfer curves (e.g. PQ)
-// and relative transfer curves (e.g. SDR, HLG).
-#define PL_COLOR_REF_WHITE 100.0
+// This defines the display-space standard reference white level (in cd/m^2)
+// that is assumed for SDR content, for use when mapping between HDR and SDR in
+// display space. See ITU-R Report BT.2408 for more information.
+#define PL_COLOR_SDR_WHITE 203.0f
 
-// The semantic interpretation of the decoded image, how is it mastered?
+// Deprecated. For compatibility with older versions of libplacebo.
+#define PL_COLOR_REF_WHITE      PL_COLOR_SDR_WHITE
+#define PL_COLOR_SDR_WHITE_HLG  3.17955
+
+// Represents a single CIE xy coordinate (e.g. CIE Yxy with Y = 1.0)
+struct pl_cie_xy {
+    float x, y;
+};
+
+// Recovers (X / Y) from a CIE xy value.
+static inline float pl_cie_X(struct pl_cie_xy xy) {
+    return xy.x / xy.y;
+}
+
+// Recovers (Z / Y) from a CIE xy value.
+static inline float pl_cie_Z(struct pl_cie_xy xy) {
+    return (1 - xy.x - xy.y) / xy.y;
+}
+
+static inline float pl_cie_xy_equal(const struct pl_cie_xy *a,
+                                    const struct pl_cie_xy *b)
+{
+    return a->x == b->x && a->y == b->y;
+}
+
+// Computes the CIE xy chromaticity coordinates of a CIE D-series illuminant
+// with the given correlated color temperature.
+//
+// `temperature` must be between 2500 K and 25000 K, inclusive.
+struct pl_cie_xy pl_white_from_temp(float temperature);
+
+// Represents the raw physical primaries corresponding to a color space.
+struct pl_raw_primaries {
+    struct pl_cie_xy red, green, blue, white;
+};
+
+// Returns whether two raw primaries are exactly identical.
+bool pl_raw_primaries_equal(const struct pl_raw_primaries *a,
+                            const struct pl_raw_primaries *b);
+
+// Replaces unknown values in the first struct by those of the second struct.
+void pl_raw_primaries_merge(struct pl_raw_primaries *orig,
+                            const struct pl_raw_primaries *update);
+
+// Returns the raw primaries for a given color space.
+const struct pl_raw_primaries *pl_raw_primaries_get(enum pl_color_primaries prim);
+
+// Represents raw HDR metadata as defined by SMPTE 2086 / CTA 861.3, which is
+// often attached to HDR sources and can be forwarded to HDR-capable displays,
+// or used to guide the libplacebo built-in tone mapping.
+struct pl_hdr_metadata {
+    // Mastering display metadata. This is used for tone-mapping.
+    struct pl_raw_primaries prim; // mastering display primaries
+    float min_luma, max_luma;     // min/max luminance (in cd/m²)
+
+    // Content light level. This is ignored by libplacebo itself.
+    float max_cll;                // max content light level (in cd/m²)
+    float max_fall;               // max frame average light level (in cd/m²)
+};
+
+extern const struct pl_hdr_metadata pl_hdr_metadata_empty; // equal to {0}
+extern const struct pl_hdr_metadata pl_hdr_metadata_hdr10; // generic HDR10 display
+
+// Returns whether two sets of HDR metadata are exactly identical.
+bool pl_hdr_metadata_equal(const struct pl_hdr_metadata *a,
+                           const struct pl_hdr_metadata *b);
+
+// Replaces unknown values in the first struct by those of the second struct.
+void pl_hdr_metadata_merge(struct pl_hdr_metadata *orig,
+                           const struct pl_hdr_metadata *update);
+
+// Deprecated. No longer used by libplacebo.
 enum pl_color_light {
     PL_COLOR_LIGHT_UNKNOWN = 0,
     PL_COLOR_LIGHT_DISPLAY,     // Display-referred, output as-is
@@ -232,7 +345,7 @@ enum pl_color_light {
     PL_COLOR_LIGHT_COUNT
 };
 
-bool pl_color_light_is_scene_referred(enum pl_color_light light);
+PL_DEPRECATED bool pl_color_light_is_scene_referred(enum pl_color_light light);
 
 // Rendering intent for colorspace transformations. These constants match the
 // ICC specification (Table 23)
@@ -249,37 +362,35 @@ enum pl_rendering_intent {
 struct pl_color_space {
     enum pl_color_primaries primaries;
     enum pl_color_transfer transfer;
-    enum pl_color_light light;
 
-    // The highest value that occurs in the signal, relative to the reference
-    // white, alternatively the brightest color value supported by a given
-    // color space. (0 = unknown)
-    float sig_peak;
+    // HDR metadata for this color space. Note that this can also be combined
+    // with SDR color transfers, in which case it's assumed that the color
+    // transfer in question is linearly "stretched" relative to these values.
+    struct pl_hdr_metadata hdr;
 
-    // The average light level that occurs in the signal, relative to the
-    // reference white, alternatively the desired average brightness level of a
-    // given color space. (0 = unknown)
-    float sig_avg;
-
-    // Additional scale factor for the signal's reference white. If this is set
-    // to a value higher than 1.0, then it's assumed that the signal's encoded
-    // reference white is assumed to be brighter than normal by this factor.
-    // This can be used to over- or under-expose content, especially HDR
-    // content. (0 = unknown)
-    //
-    // An example of where this could come in use is for using an SDR transfer
-    // function (e.g. PL_COLOR_TRC_LINEAR) to encode a HDR image or display.
-    float sig_scale;
+    // Deprecated fields
+    enum pl_color_light light PL_DEPRECATED;    // ignored
+    float sig_peak PL_DEPRECATED;               // replaced by `hdr.max_luma`
+    float sig_avg PL_DEPRECATED;                // ignored
+    float sig_floor PL_DEPRECATED;              // replaced by `hdr.min_luma`
+    float sig_scale PL_DEPRECATED;              // merged into `hdr.max/min_luma`
 };
 
+#define pl_color_space(...) (&(struct pl_color_space) { __VA_ARGS__ })
+
 // Returns whether or not a color space is considered as effectively HDR.
 // This is true when the effective signal peak is greater than the SDR
-// reference white (1.0), after application of the `sig_scale`.
-bool pl_color_space_is_hdr(struct pl_color_space csp);
+// reference white (1.0), taking into account `csp->hdr`.
+bool pl_color_space_is_hdr(const struct pl_color_space *csp);
+
+// Returns whether or not a color space is "black scaled", in which case 0.0 is
+// the true black point. This is true for SDR signals other than BT.1886, as
+// well as for HLG.
+bool pl_color_space_is_black_scaled(const struct pl_color_space *csp);
 
 // Replaces unknown values in the first struct by those of the second struct.
 void pl_color_space_merge(struct pl_color_space *orig,
-                          const struct pl_color_space *new);
+                          const struct pl_color_space *update);
 
 // Returns whether two colorspaces are exactly identical.
 bool pl_color_space_equal(const struct pl_color_space *c1,
@@ -290,8 +401,15 @@ bool pl_color_space_equal(const struct pl_color_space *c1,
 // PL_COLOR_*_UNKNOWN or 0.0.
 void pl_color_space_infer(struct pl_color_space *space);
 
+// Like `pl_color_space_infer`, but takes default values from the reference
+// color space (excluding certain special cases like HDR or wide gamut). This
+// is basically the logic used by `pl_shader_color_map` to decide the output
+// color space in a conservative way.
+void pl_color_space_infer_ref(struct pl_color_space *space,
+                              const struct pl_color_space *ref);
+
 // Some common color spaces. Note: These don't necessarily have all fields
-// filled, in particular `sig_peak` and `sig_avg` are left unset.
+// filled, in particular `hdr` is left unset.
 extern const struct pl_color_space pl_color_space_unknown;
 extern const struct pl_color_space pl_color_space_srgb;
 extern const struct pl_color_space pl_color_space_bt709;
@@ -313,6 +431,8 @@ struct pl_color_adjustment {
     float hue;
     // Gamma adjustment. 1.0 = neutral, 0.0 = solid black
     float gamma;
+    // Color temperature shift. 0.0 = 6500 K, -1.0 = 3000 K, 1.0 = 10000 K
+    float temperature;
 };
 
 // A struct pre-filled with all-neutral values.
@@ -333,31 +453,10 @@ enum pl_chroma_location {
 
 // Fills *x and *y with the offset in luma pixels corresponding to a given
 // chroma location.
+//
+// Note: PL_CHROMA_UNKNOWN defaults to PL_CHROMA_LEFT
 void pl_chroma_location_offset(enum pl_chroma_location loc, float *x, float *y);
 
-// Represents a single CIE xy coordinate (e.g. CIE Yxy with Y = 1.0)
-struct pl_cie_xy {
-    float x, y;
-};
-
-// Recovers (X / Y) from a CIE xy value.
-static inline float pl_cie_X(struct pl_cie_xy xy) {
-    return xy.x / xy.y;
-}
-
-// Recovers (Z / Y) from a CIE xy value.
-static inline float pl_cie_Z(struct pl_cie_xy xy) {
-    return (1 - xy.x - xy.y) / xy.y;
-}
-
-// Represents the raw physical primaries corresponding to a color space.
-struct pl_raw_primaries {
-    struct pl_cie_xy red, green, blue, white;
-};
-
-// Returns the raw primaries for a given color space.
-const struct pl_raw_primaries *pl_raw_primaries_get(enum pl_color_primaries prim);
-
 // Returns an RGB->XYZ conversion matrix for a given set of primaries.
 // Multiplying this into the RGB color transforms it to CIE XYZ, centered
 // around the color space's white point.
@@ -374,6 +473,15 @@ struct pl_matrix3x3 pl_get_color_mapping_matrix(const struct pl_raw_primaries *s
                                                 const struct pl_raw_primaries *dst,
                                                 enum pl_rendering_intent intent);
 
+// Return a chromatic adaptation matrix, which converts from one white point to
+// another, using the Bradford matrix. This is an RGB->RGB transformation.
+struct pl_matrix3x3 pl_get_adaptation_matrix(struct pl_cie_xy src, struct pl_cie_xy dst);
+
+// Returns true if 'b' is entirely contained in 'a'. Useful for figuring out if
+// colorimetric clipping will occur or not.
+bool pl_primaries_superset(const struct pl_raw_primaries *a,
+                           const struct pl_raw_primaries *b);
+
 // Cone types involved in human vision
 enum pl_cone {
     PL_CONE_L = 1 << 0,
@@ -395,6 +503,8 @@ struct pl_cone_params {
                         // (1.0 = Unaffected, 0.0 = Full blindness)
 };
 
+#define pl_cone_params(...) (&(struct pl_cone_params) { __VA_ARGS__ })
+
 // Built-in color blindness models
 extern const struct pl_cone_params pl_vision_normal;        // No distortion (92%)
 extern const struct pl_cone_params pl_vision_protanomaly;   // Red deficiency (0.66%)
@@ -419,15 +529,12 @@ struct pl_matrix3x3 pl_get_cone_matrix(const struct pl_cone_params *params,
                                        const struct pl_raw_primaries *prim);
 
 // Returns a color decoding matrix for a given combination of source color
-// representation and adjustment parameters. This mutates the color_repr to
-// reflect the change. If `params` is left as NULL, it defaults to
-// &pl_color_adjustment_neutral.
+// representation and adjustment parameters. This mutates `repr` to reflect the
+// change. If `params` is NULL, it defaults to &pl_color_adjustment_neutral.
 //
-// This function always performs a conversion to RGB; conversions from
-// arbitrary color representations to other arbitrary other color
-// representations are currently not supported. Not all color systems support
-// all of the color adjustment parameters. (In particular, hue/sat adjustments
-// are currently only supported for YCbCr-like color systems)
+// This function always performs a conversion to RGB. To convert to other
+// colorspaces (e.g. between YUV systems), obtain a second YUV->RGB matrix
+// and invert it using `pl_transform3x3_invert`.
 //
 // Note: For BT.2020 constant-luminance, this outputs chroma information in the
 // range [-0.5, 0.5]. Since the CL system conversion is non-linear, further
@@ -449,10 +556,10 @@ struct pl_icc_profile {
     const void *data;
     size_t len;
 
-    // If a profile is set, this signature must uniquely identify it. It could
-    // be, for example, a checksum of the profile contents. Alternatively, it
-    // could be the pointer to the ICC profile itself, as long as the user
-    // makes sure that this memory is used in an immutable way.
+    // If a profile is set, this signature must uniquely identify it, ideally
+    // using a checksum of the profile contents. The user is free to choose the
+    // method of determining this signature, but note the existence of the
+    // `pl_icc_profile_compute_signature` helper.
     uint64_t signature;
 };
 
@@ -460,4 +567,14 @@ struct pl_icc_profile {
 bool pl_icc_profile_equal(const struct pl_icc_profile *p1,
                           const struct pl_icc_profile *p2);
 
+// Sets `signature` to a hash of `profile->data`, if non-NULL. Provided as a
+// convenience function for the sake of users ingesting arbitrary ICC profiles
+// from sources where they can't reliably detect profile changes.
+//
+// Note: This is based on a very fast hash, and will compute a signature for
+// even large (10 MB) ICC profiles in, typically, a fraction of a millisecond.
+void pl_icc_profile_compute_signature(struct pl_icc_profile *profile);
+
+PL_API_END
+
 #endif // LIBPLACEBO_COLORSPACE_H_
