@@ -18,11 +18,19 @@
 #ifndef LIBPLACEBO_OPENGL_H_
 #define LIBPLACEBO_OPENGL_H_
 
-#include "gpu.h"
+#include <libplacebo/gpu.h>
+#include <libplacebo/swapchain.h>
 
-struct pl_opengl {
-    const struct pl_gpu *gpu;
-};
+PL_API_BEGIN
+
+// Note on thread safety: The thread safety of `pl_opengl` and any associated
+// GPU objects follows the same thread safety rules as the underlying OpenGL
+// context. In other words, they must only be called from the thread the OpenGL
+// context is current on.
+
+typedef const PL_STRUCT(pl_opengl) {
+    pl_gpu gpu;
+} *pl_opengl;
 
 struct pl_opengl_params {
     // Enable OpenGL debug report callbacks. May have little effect depending
@@ -30,13 +38,33 @@ struct pl_opengl_params {
     // debugging enabled.
     bool debug;
 
-    // Restrict specific features to e.g. work around driver bugs, or simply
-    // for testing purposes
-    pl_gpu_caps blacklist_caps; // capabilities to be excluded
-    int max_glsl_version;       // limit the maximum GLSL version
+    // Allow the use of (suspected) software rasterizers and renderers. These
+    // can be useful for debugging purposes, but normally, their use is
+    // undesirable when GPU-accelerated processing is expected.
+    bool allow_software;
+
+    // Restrict the maximum allowed GLSL version. (Mainly for testing)
+    int max_glsl_version;
+
+    // Optional. Required when importing/exporting dmabufs as textures.
+    void *egl_display;
+    void *egl_context;
+
+    // Optional callbacks to bind/release the OpenGL context on the current
+    // thread. If these are specified, then the resulting `pl_gpu` will have
+    // `pl_gpu_limits.thread_safe` enabled, and may therefore be used from any
+    // thread without first needing to bind the OpenGL context.
+    //
+    // If the user is re-using the same OpenGL context in non-libplacebo code,
+    // then these callbacks should include whatever synchronization is
+    // necessary to prevent simultaneous use between libplacebo and the user.
+    bool (*make_current)(void *priv);
+    void (*release_current)(void *priv);
+    void *priv;
 };
 
 // Default/recommended parameters
+#define pl_opengl_params(...) (&(struct pl_opengl_params) { __VA_ARGS__ })
 extern const struct pl_opengl_params pl_opengl_default_params;
 
 // Creates a new OpenGL renderer based on the given parameters. This will
@@ -49,12 +77,15 @@ extern const struct pl_opengl_params pl_opengl_default_params;
 //
 // Note that creating multiple `pl_opengl` instances from the same OpenGL
 // context is undefined behavior.
-const struct pl_opengl *pl_opengl_create(struct pl_context *ctx,
-                                         const struct pl_opengl_params *params);
+pl_opengl pl_opengl_create(pl_log log, const struct pl_opengl_params *params);
 
 // All resources allocated from the `pl_gpu` contained by this `pl_opengl` must
 // be explicitly destroyed by the user before calling `pl_opengl_destroy`.
-void pl_opengl_destroy(const struct pl_opengl **gl);
+void pl_opengl_destroy(pl_opengl *gl);
+
+// For a `pl_gpu` backed by `pl_opengl`, this function can be used to retrieve
+// the underlying `pl_opengl`. Returns NULL for any other type of `gpu`.
+pl_opengl pl_opengl_get(pl_gpu gpu);
 
 struct pl_opengl_framebuffer {
     // ID of the framebuffer, or 0 to use the context's default framebuffer.
@@ -89,47 +120,57 @@ struct pl_opengl_swapchain_params {
     void *priv;
 };
 
+#define pl_opengl_swapchain_params(...) (&(struct pl_opengl_swapchain_params) { __VA_ARGS__ })
+
 // Creates an instance of `pl_swapchain` tied to the active context.
 // Note: Due to OpenGL semantics, users *must* call `pl_swapchain_resize`
 // before attempting to use this swapchain, otherwise calls to
 // `pl_swapchain_start_frame` will fail.
-const struct pl_swapchain *pl_opengl_create_swapchain(const struct pl_opengl *gl,
+pl_swapchain pl_opengl_create_swapchain(pl_opengl gl,
                             const struct pl_opengl_swapchain_params *params);
 
 // Update the framebuffer description. After calling this function, users
 // *must* call `pl_swapchain_resize` before attempting to use the swapchain
 // again, otherwise calls to `pl_swapchain_start_frame` will fail.
-void pl_opengl_swapchain_update_fb(const struct pl_swapchain *sw,
+void pl_opengl_swapchain_update_fb(pl_swapchain sw,
                                    const struct pl_opengl_framebuffer *fb);
 
 struct pl_opengl_wrap_params {
-    // The GLuint texture object itself.
+    // The GLuint texture object itself. Optional. If no texture is provided,
+    // then only the opaque framebuffer `fbo` will be wrapped, leaving the
+    // resulting `pl_tex` object with some operations (such as sampling) being
+    // unsupported.
     unsigned int texture;
 
+    // The GLuint associated framebuffer. Optional. If this is not specified,
+    // then libplacebo will attempt creating a framebuffer from the provided
+    // texture object (if possible).
+    //
+    // Note: As a special case, if neither a texture nor an FBO are provided,
+    // this is equivalent to wrapping the OpenGL default framebuffer (id 0).
+    unsigned int framebuffer;
+
     // The image's dimensions (unused dimensions must be 0)
     int width;
     int height;
     int depth;
 
+    // Texture-specific fields:
+    //
+    // Note: These are only relevant if `texture` is provided.
+
     // The GLenum for the texture target to use, e.g. GL_TEXTURE_2D. Optional.
     // If this is left as 0, the target is inferred from the number of
     // dimensions. Users may want to set this to something specific like
     // GL_TEXTURE_EXTERNAL_OES depending on the nature of the texture.
     unsigned int target;
 
-    // The texture's GLint sized internal format (e.g. GL_RGBA16F)
+    // The texture's GLint sized internal format (e.g. GL_RGBA16F). Required.
     int iformat;
-
-    // The GLint texture min/mag filter. libplacebo does not distinguish
-    // between magnification/minification, so both must be the same. Optional,
-    // defaults to GL_LINEAR.
-    int filter;
-
-    // The GLint texture address mode. libplacebo does not distinguish by
-    // dimension, so all must be the same. Optional, defaults to GL_REPEAT.
-    int address_mode;
 };
 
+#define pl_opengl_wrap_params(...) (&(struct pl_opengl_wrap_params) { __VA_ARGS__ })
+
 // Wraps an external OpenGL object into a `pl_tex` abstraction. Due to the
 // internally synchronized nature of OpenGL, no explicit synchronization
 // is needed between libplacebo `pl_tex_` operations, and host accesses to
@@ -137,11 +178,10 @@ struct pl_opengl_wrap_params {
 // Note that this function transfers no ownership.
 //
 // This wrapper can be destroyed by simply calling `pl_tex_destroy` on it,
-// which will not destroy the underlying OpenGL texture.
+// which will *not* destroy the user-provided OpenGL texture or framebuffer.
 //
 // This function may fail, in which case it returns NULL.
-const struct pl_tex *pl_opengl_wrap(const struct pl_gpu *gpu,
-                                    const struct pl_opengl_wrap_params *params);
+pl_tex pl_opengl_wrap(pl_gpu gpu, const struct pl_opengl_wrap_params *params);
 
 // Analogous to `pl_opengl_wrap`, this function takes any `pl_tex` (including
 // ones created by `pl_tex_create`) and unwraps it to expose the underlying
@@ -153,8 +193,9 @@ const struct pl_tex *pl_opengl_wrap(const struct pl_gpu *gpu,
 //
 // For renderable/blittable textures, `out_fbo` will be updated to the ID of
 // the framebuffer attached to this texture, or 0 if there is none. (Optional)
-unsigned int pl_opengl_unwrap(const struct pl_gpu *gpu, const struct pl_tex *tex,
-                              unsigned int *out_target, int *out_iformat,
-                              unsigned int *out_fbo);
+unsigned int pl_opengl_unwrap(pl_gpu gpu, pl_tex tex, unsigned int *out_target,
+                              int *out_iformat, unsigned int *out_fbo);
+
+PL_API_END
 
 #endif // LIBPLACEBO_OPENGL_H_
