@@ -15,13 +15,15 @@
  * License along with libplacebo.  If not, see <http://www.gnu.org/licenses/>.
  */
 
+#ifndef LIBPLACEBO_UPLOAD_H_
+#define LIBPLACEBO_UPLOAD_H_
+
 #include <stdint.h>
 
 #include <libplacebo/gpu.h>
 #include <libplacebo/renderer.h>
 
-#ifndef LIBPLACEBO_UPLOAD_H_
-#define LIBPLACEBO_UPLOAD_H_
+PL_API_BEGIN
 
 // This file contains a utility function to assist in uploading data from host
 // memory to a texture. In particular, the texture will be suitable for use as
@@ -29,27 +31,32 @@
 
 // Description of the host representation of an image plane
 struct pl_plane_data {
-    enum pl_fmt_type type; // meaning of the data (must not be UINT or SINT)
-    int width, height;     // dimensions of the plane
-    int component_size[4]; // size in bits of each coordinate
-    int component_pad[4];  // ignored bits preceding each component
-    int component_map[4];  // semantic meaning of each component (pixel order)
-    size_t pixel_stride;   // offset in bytes between pixels (required)
-    size_t row_stride;     // offset in bytes between rows (optional)
+    enum pl_fmt_type type;  // meaning of the data (must not be UINT or SINT)
+    int width, height;      // dimensions of the plane
+    int component_size[4];  // size in bits of each coordinate
+    int component_pad[4];   // ignored bits preceding each component
+    int component_map[4];   // semantic meaning of each component (pixel order)
+    size_t pixel_stride;    // offset in bytes between pixels (required)
+    size_t row_stride;      // offset in bytes between rows (optional)
 
     // Similar to `pl_tex_transfer_params`, you can either upload from a raw
     // pointer address, or a buffer + offset. Again, the use of these two
     // mechanisms is mutually exclusive.
     //
     // 1. Uploading from host memory
-    const void *pixels;    // the actual data underlying this plane
+    const void *pixels;     // the actual data underlying this plane
 
-    // 2. Uploading from a buffer
-    const struct pl_buf *buf;   // the buffer to use (type must be PL_BUF_TEX_TRANSFER)
-    size_t buf_offset;          // offset of data within buffer, must be a
-                                // multiple of `pixel_stride` as well as of 4
+    // 2. Uploading from a buffer (requires `pl_gpu_limits.buf_transfer`)
+    pl_buf buf;             // the buffer to use
+    size_t buf_offset;      // offset of data within buffer, must be a
+                            // multiple of `pixel_stride` as well as of 4
 
-    // Note: When using this together with `pl_image`, there is some amount of
+    // Similar to `pl_tex_transfer_params.callback`, this allows turning the
+    // upload of a plane into an asynchronous upload. The same notes apply.
+    void (*callback)(void *priv);
+    void *priv;
+
+    // Note: When using this together with `pl_frame`, there is some amount of
     // overlap between `component_pad` and `pl_color_repr.bits`. Some key
     // differences between the two:
     //
@@ -70,7 +77,7 @@ struct pl_plane_data {
     // and working with non-byte-aligned pixels is awkward in general, the
     // convention is to represent yuv420p10 as 16-bit samples with either the
     // high or low bits set to 0. In this scenario, the `component_size` of the
-    // `pl_plane_repr` and `pl_bit_encoding.sample_depth` would be 16, while
+    // `pl_plane_data` and `pl_bit_encoding.sample_depth` would be 16, while
     // the `pl_bit_encoding.color_depth` would be 10 (and additionally, the
     // `pl_bit_encoding.bit_shift` would be either 0 or  6, depending on
     // whether the low or the high bits are used).
@@ -88,16 +95,25 @@ struct pl_plane_data {
 
 // Fills in the `component_size`, `component_pad` and `component_map` fields
 // based on the supplied mask for each component (in semantic order, i.e.
-// RGBA). If `mask` does not have a contiguous range of set bits, then the
-// result is undefined and probably not useful.
+// RGBA). Each element of `mask` must have a contiguous range of set bits.
 void pl_plane_data_from_mask(struct pl_plane_data *data, uint64_t mask[4]);
 
+// Helper function to take a `pl_plane_data` struct and try and improve its
+// alignment to make it more likely to correspond to a real `pl_fmt`. It does
+// this by attempting to round each component up to the nearest byte boundary.
+// This relies on the assumption (true in practice) that superfluous bits of
+// byte-misaligned formats are explicitly set to 0.
+//
+// The resulting shift must be consistent across all components, in which case
+// it's returned in `out_bits`. If no alignment was possible, `out_bits` is set
+// to {0}, and this function returns false.
+bool pl_plane_data_align(struct pl_plane_data *data, struct pl_bit_encoding *out_bits);
+
 // Helper function to find a suitable `pl_fmt` based on a pl_plane_data's
 // requirements. This is called internally by `pl_upload_plane`, but it's
 // exposed to users both as a convenience and so they may pre-emptively check
 // if a format would be supported without actually having to attempt the upload.
-const struct pl_fmt *pl_plane_find_fmt(const struct pl_gpu *gpu, int out_map[4],
-                                       const struct pl_plane_data *data);
+pl_fmt pl_plane_find_fmt(pl_gpu gpu, int out_map[4], const struct pl_plane_data *data);
 
 // Upload an image plane to a texture, and output the resulting `pl_plane`
 // struct to `out_plane` (optional). `tex` must be a valid pointer to a texture
@@ -107,7 +123,22 @@ const struct pl_fmt *pl_plane_find_fmt(const struct pl_gpu *gpu, int out_map[4],
 // The resulting texture is guaranteed to be `sampleable`, and it will also try
 // and maximize compatibility with the other `pl_renderer` requirements
 // (blittable, linear filterable, etc.).
-bool pl_upload_plane(const struct pl_gpu *gpu, struct pl_plane *out_plane,
-                     const struct pl_tex **tex, const struct pl_plane_data *data);
+//
+// Note: `out_plane->shift_x/y` are left uninitialized, and should be set
+// explicitly by the user.
+bool pl_upload_plane(pl_gpu gpu, struct pl_plane *out_plane,
+                     pl_tex *tex, const struct pl_plane_data *data);
+
+// Like `pl_upload_plane`, but only creates an uninitialized texture object
+// rather than actually performing an upload. This can be useful to, for
+// example, prepare textures to be used as the target of rendering.
+//
+// The resulting texture is guaranteed to be `renderable`, and it will also try
+// to maximize compatibility with the other `pl_renderer` requirements
+// (blittable, storable, etc.).
+bool pl_recreate_plane(pl_gpu gpu, struct pl_plane *out_plane,
+                       pl_tex *tex, const struct pl_plane_data *data);
+
+PL_API_END
 
 #endif // LIBPLACEBO_UPLOAD_H_
