@@ -22,8 +22,10 @@
 #include <libplacebo/gpu.h>
 #include <libplacebo/swapchain.h>
 
+PL_API_BEGIN
+
 // Structure representing a VkInstance. Using this is not required.
-struct pl_vk_inst {
+typedef const PL_STRUCT(pl_vk_inst) {
     VkInstance instance;
 
     // The Vulkan API version supported by this VkInstance.
@@ -41,7 +43,7 @@ struct pl_vk_inst {
     // layers enabled by libplacebo internally. May contain duplicates.
     const char * const *layers;
     int num_layers;
-};
+} *pl_vk_inst;
 
 struct pl_vk_inst_params {
     // If set, enable the debugging and validation layers. These should
@@ -87,16 +89,15 @@ struct pl_vk_inst_params {
     int num_opt_layers;
 };
 
+#define pl_vk_inst_params(...) (&(struct pl_vk_inst_params) { __VA_ARGS__ })
 extern const struct pl_vk_inst_params pl_vk_inst_default_params;
 
 // Helper function to simplify instance creation. The user could also bypass
 // these helpers and do it manually, but this function is provided as a
 // convenience. It also sets up a debug callback which forwards all vulkan
-// messages to the `pl_context` log callback.
-const struct pl_vk_inst *pl_vk_inst_create(struct pl_context *ctx,
-                                           const struct pl_vk_inst_params *params);
-
-void pl_vk_inst_destroy(const struct pl_vk_inst **inst);
+// messages to the `pl_log` callback.
+pl_vk_inst pl_vk_inst_create(pl_log log, const struct pl_vk_inst_params *params);
+void pl_vk_inst_destroy(pl_vk_inst *inst);
 
 struct pl_vulkan_queue {
     uint32_t index; // Queue family index
@@ -104,8 +105,8 @@ struct pl_vulkan_queue {
 };
 
 // Structure representing the actual vulkan device and associated GPU instance
-struct pl_vulkan {
-    const struct pl_gpu *gpu;
+typedef const PL_STRUCT(pl_vulkan) {
+    pl_gpu gpu;
 
     // The vulkan objects in use. The user may use this for their own purposes,
     // but please note that the lifetime is tied to the lifetime of the
@@ -125,7 +126,7 @@ struct pl_vulkan {
     int num_extensions;
 
     // The device features that were enabled at device creation time.
-    const VkPhysicalDeviceFeatures2KHR *features;
+    const VkPhysicalDeviceFeatures2 *features;
 
     // The explicit queue families we are using to provide a given capability,
     // or {0} if no appropriate dedicated queue family exists for this
@@ -143,14 +144,14 @@ struct pl_vulkan {
     // queue counts in list form. This list does not contain duplicates.
     const struct pl_vulkan_queue *queues;
     int num_queues;
-};
+} *pl_vulkan;
 
 struct pl_vulkan_params {
     // The vulkan instance. Optional, if NULL then libplacebo will internally
     // create a VkInstance with the settings from `instance_params`.
     //
-    // NOTE: The VkInstance provided by the user *MUST* be created with the
-    // `VK_KHR_get_physical_device_properties2` extension enabled!
+    // Note: The VkInstance provided by the user *MUST* be created with a
+    // VkApplicationInfo.apiVersion of VK_API_VERSION_1_1 or higher.
     VkInstance instance;
 
     // Pointer to `vkGetInstanceProcAddr`. If this is NULL, libplacebo will
@@ -186,11 +187,6 @@ struct pl_vulkan_params {
     // When choosing the device, only choose a device with this exact UUID.
     // This overrides `allow_software` and `device_name`. No effect if `device`
     // is set.
-    //
-    // Note: This relies on instance-level support for at least one of the
-    // VK_KHR_external_*_capabilities extensions (or vulkan 1.1). If this field
-    // is set when the instance does not support it, an error will be
-    // generated.
     uint8_t device_uuid[16];
 
     // When choosing the device, controls whether or not to also allow software
@@ -235,27 +231,36 @@ struct pl_vulkan_params {
     // features in the pNext chain, however these *must* all be
     // extension-specific structs, i.e. the use of "meta-structs" like
     // VkPhysicalDeviceVulkan11Features is not allowed.
-    const VkPhysicalDeviceFeatures2KHR *features;
+    const VkPhysicalDeviceFeatures2 *features;
 
     // --- Misc/debugging options
 
     // Restrict specific features to e.g. work around driver bugs, or simply
     // for testing purposes
-    pl_gpu_caps blacklist_caps; // capabilities to be excluded
     int max_glsl_version;       // limit the maximum GLSL version
-    bool disable_events;        // disables usage of VkEvent completely
     uint32_t max_api_version;   // limit the maximum vulkan API version
+
+    // Removed parameters (no effect)
+    bool disable_events PL_DEPRECATED;
 };
 
 // Default/recommended parameters. Should generally be safe and efficient.
+#define PL_VULKAN_DEFAULTS                              \
+    .async_transfer = true,                             \
+    .async_compute  = true,                             \
+    /* enabling multiple queues often decreases perf */ \
+    .queue_count    = 1,
+
+#define pl_vulkan_params(...) (&(struct pl_vulkan_params) { PL_VULKAN_DEFAULTS __VA_ARGS__ })
 extern const struct pl_vulkan_params pl_vulkan_default_params;
 
 // Creates a new vulkan device based on the given parameters and initializes
 // a new GPU. This function will internally initialize a VkDevice. There is
 // currently no way to share a vulkan device with the caller. If `params` is
 // left as NULL, it defaults to &pl_vulkan_default_params.
-const struct pl_vulkan *pl_vulkan_create(struct pl_context *ctx,
-                                         const struct pl_vulkan_params *params);
+//
+// Thread-safety: Safe
+pl_vulkan pl_vulkan_create(pl_log log, const struct pl_vulkan_params *params);
 
 // Destroys the vulkan device and all associated objects, except for the
 // VkInstance provided by the user.
@@ -267,13 +272,17 @@ const struct pl_vulkan *pl_vulkan_create(struct pl_context *ctx,
 // Also note that this function will block until all in-flight GPU commands are
 // finished processing. You can avoid this by manually calling `pl_gpu_finish`
 // before `pl_vulkan_destroy`.
-void pl_vulkan_destroy(const struct pl_vulkan **vk);
+void pl_vulkan_destroy(pl_vulkan *vk);
+
+// For a `pl_gpu` backed by `pl_vulkan`, this function can be used to retrieve
+// the underlying `pl_vulkan`. Returns NULL for any other type of `gpu`.
+pl_vulkan pl_vulkan_get(pl_gpu gpu);
 
 struct pl_vulkan_device_params {
     // The instance to use. Required!
     //
-    // NOTE: The VkInstance provided by the user *MUST* be created with the
-    // `VK_KHR_get_physical_device_properties2` extension enabled!
+    // Note: The VkInstance provided by the user *must* be created with a
+    // VkApplicationInfo.apiVersion of VK_API_VERSION_1_1 or higher.
     VkInstance instance;
 
     // Mirrored from `pl_vulkan_params`. All of these fields are optional.
@@ -284,21 +293,25 @@ struct pl_vulkan_device_params {
     bool allow_software;
 };
 
+#define pl_vulkan_device_params(...) (&(struct pl_vulkan_device_params) { __VA_ARGS__ })
+
 // Helper function to choose the best VkPhysicalDevice, given a VkInstance.
 // This uses the same logic as `pl_vulkan_create` uses internally. If no
 // matching device was found, this returns VK_NULL_HANDLE.
-VkPhysicalDevice pl_vulkan_choose_device(struct pl_context *ctx,
-                                         const struct pl_vulkan_device_params *params);
+VkPhysicalDevice pl_vulkan_choose_device(pl_log log,
+                              const struct pl_vulkan_device_params *params);
 
 struct pl_vulkan_swapchain_params {
     // The surface to use for rendering. Required, the user is in charge of
     // creating this. Must belong to the same VkInstance as `vk->instance`.
     VkSurfaceKHR surface;
 
-    // The image format and colorspace we should be using. Optional, if left
-    // as {0}, libplacebo will pick the best surface format based on what the
-    // GPU/surface seems to support.
-    VkSurfaceFormatKHR surface_format;
+    // When choosing the initial surface format, prefer HDR formats over SDR
+    // formats, if any is available.
+    //
+    // Deprecated in favor of `pl_swapchain_colorspace_hint`, which overrides
+    // this parameter if called.
+    bool prefer_hdr PL_DEPRECATED;
 
     // The preferred presentation mode. See the vulkan documentation for more
     // information about these. If the device/surface combination does not
@@ -327,17 +340,19 @@ struct pl_vulkan_swapchain_params {
     bool allow_suboptimal;
 };
 
+#define pl_vulkan_swapchain_params(...) (&(struct pl_vulkan_swapchain_params) { __VA_ARGS__ })
+
 // Creates a new vulkan swapchain based on an existing VkSurfaceKHR. Using this
 // function requires that the vulkan device was created with the
 // VK_KHR_swapchain extension. The easiest way of accomplishing this is to set
 // the `pl_vulkan_params.surface` explicitly at creation time.
-const struct pl_swapchain *pl_vulkan_create_swapchain(const struct pl_vulkan *vk,
+pl_swapchain pl_vulkan_create_swapchain(pl_vulkan vk,
                               const struct pl_vulkan_swapchain_params *params);
 
 // This will return true if the vulkan swapchain is internally detected
 // as being suboptimal (VK_SUBOPTIMAL_KHR). This might be of use to clients
 // who have `params->allow_suboptimal` enabled.
-bool pl_vulkan_swapchain_suboptimal(const struct pl_swapchain *sw);
+bool pl_vulkan_swapchain_suboptimal(pl_swapchain sw);
 
 // Vulkan interop API, for sharing a single VkDevice (and associated vulkan
 // resources) directly with the API user. The use of this API is a bit sketchy
@@ -346,8 +361,8 @@ bool pl_vulkan_swapchain_suboptimal(const struct pl_swapchain *sw);
 struct pl_vulkan_import_params {
     // The vulkan instance. Required.
     //
-    // NOTE: The VkInstance provided by the user *MUST* be created with the
-    // `VK_KHR_get_physical_device_properties2` extension enabled!
+    // Note: The VkInstance provided by the user *must* be created with a
+    // VkApplicationInfo.apiVersion of VK_API_VERSION_1_1 or higher.
     VkInstance instance;
 
     // Pointer to `vkGetInstanceProcAddr`. If this is NULL, libplacebo will
@@ -382,26 +397,28 @@ struct pl_vulkan_import_params {
     struct pl_vulkan_queue queue_compute;  // must support VK_QUEUE_COMPUTE_BIT
     struct pl_vulkan_queue queue_transfer; // must support VK_QUEUE_TRANSFER_BIT
 
-    // Enabled VkPhysicalDeviceFeatures. May be left as NULL, in which case
-    // libplacebo will assume no extra device features were enabled.
-    const VkPhysicalDeviceFeatures2KHR *features;
+    // Enabled VkPhysicalDeviceFeatures. The VkDevice provided by the user
+    // *must* be created with the `timelineSemaphore` feature enabled.
+    const VkPhysicalDeviceFeatures2 *features;
 
     // --- Misc/debugging options
 
     // Restrict specific features to e.g. work around driver bugs, or simply
     // for testing purposes. See `pl_vulkan_params` for a description of these.
-    pl_gpu_caps blacklist_caps;
     int max_glsl_version;
-    bool disable_events;
     uint32_t max_api_version;
+
+    // Removed parameters (no effect)
+    bool disable_events PL_DEPRECATED;
 };
 
+#define pl_vulkan_import_params(...) (&(struct pl_vulkan_import_params) { __VA_ARGS__ })
+
 // Import an existing VkDevice instead of creating a new one, and wrap it into
 // a `pl_vulkan` abstraction. It's safe to `pl_vulkan_destroy` this, which will
 // destroy application state related to libplacebo but leave the underlying
 // VkDevice intact.
-const struct pl_vulkan *pl_vulkan_import(struct pl_context *ctx,
-                                         const struct pl_vulkan_import_params *params);
+pl_vulkan pl_vulkan_import(pl_log log, const struct pl_vulkan_import_params *params);
 
 struct pl_vulkan_wrap_params {
     // The image itself. It *must* be usable concurrently by all of the queue
@@ -426,13 +443,16 @@ struct pl_vulkan_wrap_params {
     // of enabled usage flags.
     VkImageUsageFlags usage;
 
-    // The desired sampling / address modes of the resulting `pl_tex`. Note
-    // that PL_TEX_SAMPLE_LINEAR requires that `format` be a linearly
-    // sampleable texture format.
-    enum pl_tex_sample_mode sample_mode;
-    enum pl_tex_address_mode address_mode;
+    // See `pl_tex_params`
+    void *user_data;
+    pl_debug_tag debug_tag;
 };
 
+#define pl_vulkan_wrap_params(...) (&(struct pl_vulkan_wrap_params) {   \
+        .debug_tag = PL_DEBUG_TAG,                                      \
+        __VA_ARGS__                                                     \
+    })
+
 // Wraps an external VkImage into a pl_tex abstraction. By default, the image
 // is considered "held" by the user and must be released before calling any
 // pl_tex_* API calls on it (see `pl_vulkan_release`).
@@ -447,8 +467,7 @@ struct pl_vulkan_wrap_params {
 // `gpu`.
 //
 // This function may fail, in which case it returns NULL.
-const struct pl_tex *pl_vulkan_wrap(const struct pl_gpu *gpu,
-                                    const struct pl_vulkan_wrap_params *params);
+pl_tex pl_vulkan_wrap(pl_gpu gpu, const struct pl_vulkan_wrap_params *params);
 
 // For purely informative reasons, this contains a list of extensions and
 // device features that libplacebo *can* make use of. These are all strictly
@@ -460,7 +479,7 @@ const struct pl_tex *pl_vulkan_wrap(const struct pl_gpu *gpu,
 // than the more general purpose VkPhysicalDeviceVulkan11Features etc.
 extern const char * const pl_vulkan_recommended_extensions[];
 extern const int pl_vulkan_num_recommended_extensions;
-extern const VkPhysicalDeviceFeatures2KHR pl_vulkan_recommended_features;
+extern const VkPhysicalDeviceFeatures2 pl_vulkan_recommended_features;
 
 // Analogous to `pl_vulkan_wrap`, this function takes any `pl_tex` (including
 // ones created by `pl_tex_create`) and unwraps it to expose the underlying
@@ -470,39 +489,52 @@ extern const VkPhysicalDeviceFeatures2KHR pl_vulkan_recommended_features;
 //
 // `out_format` and `out_flags` will be updated to hold the VkImage's
 // format and usage flags. (Optional)
-VkImage pl_vulkan_unwrap(const struct pl_gpu *gpu, const struct pl_tex *tex,
+VkImage pl_vulkan_unwrap(pl_gpu gpu, pl_tex tex,
                          VkFormat *out_format, VkImageUsageFlags *out_flags);
 
-// "Hold" a shared image. This will transition the image into the layout and
-// access mode specified by the user, and fire the given semaphore (optional)
-// when this is done. This marks the image as held. Attempting to perform any
-// pl_tex_* operation (except pl_tex_destroy) on a held image is undefined
-// behavior.
+// Represents a vulkan semaphore/value pair (for compatibility with timeline
+// semaphores). When using normal, binary semaphores, `value` may be ignored.
+typedef struct pl_vulkan_sem {
+    VkSemaphore sem;
+    uint64_t value;
+} pl_vulkan_sem;
+
+// "Hold" a shared image. This will transition the image into the layout
+// specified by the user, and fire the given semaphore (required) when this is
+// done. This marks the image as held. Attempting to perform any pl_tex_*
+// operation (except pl_tex_destroy) on a held image is undefined behavior.
 //
 // Returns whether successful.
-bool pl_vulkan_hold(const struct pl_gpu *gpu, const struct pl_tex *tex,
-                    VkImageLayout layout, VkAccessFlags access,
-                    VkSemaphore sem_out);
+bool pl_vulkan_hold(pl_gpu gpu, pl_tex tex, VkImageLayout layout,
+                    pl_vulkan_sem sem_out);
 
 // This function is similar to `pl_vulkan_hold`, except that rather than
 // forcibly transitioning to a given layout, the user is instead informed about
-// the current layout and access and is in charge of transitioning it to their
-// own layout/access before using it. May be more convenient for some users.
+// the current layout and is in charge of transitioning it to their own layout
+// before using it. May be more convenient for some users.
 //
 // Returns whether successful.
-bool pl_vulkan_hold_raw(const struct pl_gpu *gpu, const struct pl_tex *tex,
-                        VkImageLayout *layout, VkAccessFlags *access,
-                        VkSemaphore sem_out);
+bool pl_vulkan_hold_raw(pl_gpu gpu, pl_tex tex, VkImageLayout *layout,
+                        pl_vulkan_sem sem_out);
 
-// "Release" a shared image, meaning it is no longer held. `layout` and
-// `access` describe the current state of the image at the point in time when
-// the user is releasing it. Performing any operation on the VkImage underlying
-// this `pl_tex` while it is not being held by the user is undefined behavior.
+// "Release" a shared image, meaning it is no longer held. `layout` describes
+// the current layout of the image at the point in time when the user is
+// releasing it. Performing any operation on the VkImage underlying this
+// `pl_tex` while it is not being held by the user is undefined behavior.
 //
 // If `sem_in` is specified, it must fire before libplacebo will actually use
 // or modify the image. (Optional)
-void pl_vulkan_release(const struct pl_gpu *gpu, const struct pl_tex *tex,
-                       VkImageLayout layout, VkAccessFlags access,
-                       VkSemaphore sem_in);
+//
+// Note: the lifetime of `sem_in` is indeterminate, and destroying it while the
+// texture is still depending on that semaphore is undefined behavior.
+//
+// Technically, the only way to be sure that it's safe to free is to use
+// `pl_gpu_finish()` or similar (e.g. `pl_vulkan_destroy` or
+// `vkDeviceWaitIdle`) after another operation involving `tex` has been emitted
+// (or the texture has been destroyed).
+void pl_vulkan_release(pl_gpu gpu, pl_tex tex, VkImageLayout layout,
+                       pl_vulkan_sem sem_in);
+
+PL_API_END
 
 #endif // LIBPLACEBO_VULKAN_H_
