@@ -19,21 +19,19 @@
 #include <math.h>
 
 #include "common.h"
-#include "context.h"
+#include "log.h"
 #include "shaders.h"
 
-struct pl_shader *pl_shader_alloc(struct pl_context *ctx,
-                                  const struct pl_shader_params *params)
+pl_shader pl_shader_alloc(pl_log log, const struct pl_shader_params *params)
 {
-    pl_assert(ctx);
-    struct pl_shader *sh = pl_alloc_ptr(ctx, sh);
+    pl_shader sh = pl_alloc_ptr(NULL, sh);
     *sh = (struct pl_shader) {
-        .ctx = ctx,
+        .log = log,
         .mutable = true,
     };
 
     // Ensure there's always at least one `tmp` object
-    PL_ARRAY_APPEND(sh, sh->tmp, pl_ref_new(ctx));
+    PL_ARRAY_APPEND(sh, sh->tmp, pl_ref_new(NULL));
 
     if (params)
         sh->res.params = *params;
@@ -41,9 +39,9 @@ struct pl_shader *pl_shader_alloc(struct pl_context *ctx,
     return sh;
 }
 
-void pl_shader_free(struct pl_shader **psh)
+void pl_shader_free(pl_shader *psh)
 {
-    struct pl_shader *sh = *psh;
+    pl_shader sh = *psh;
     if (!sh)
         return;
 
@@ -53,20 +51,22 @@ void pl_shader_free(struct pl_shader **psh)
     pl_free_ptr(psh);
 }
 
-void pl_shader_reset(struct pl_shader *sh, const struct pl_shader_params *params)
+void pl_shader_reset(pl_shader sh, const struct pl_shader_params *params)
 {
     for (int i = 0; i < sh->tmp.num; i++)
         pl_ref_deref(&sh->tmp.elem[i]);
 
     struct pl_shader new = {
-        .ctx = sh->ctx,
+        .log = sh->log,
         .mutable = true,
 
         // Preserve array allocations
-        .tmp.elem   = sh->tmp.elem,
-        .vars.elem  = sh->vars.elem,
-        .descs.elem = sh->descs.elem,
-        .vas.elem   = sh->vas.elem,
+        .tmp.elem       = sh->tmp.elem,
+        .vas.elem       = sh->vas.elem,
+        .vars.elem      = sh->vars.elem,
+        .descs.elem     = sh->descs.elem,
+        .consts.elem    = sh->consts.elem,
+        .steps.elem     = sh->steps.elem,
     };
 
     if (params)
@@ -77,15 +77,15 @@ void pl_shader_reset(struct pl_shader *sh, const struct pl_shader_params *params
         new.buffers[i] = (pl_str) { .buf = sh->buffers[i].buf };
 
     *sh = new;
-    PL_ARRAY_APPEND(sh, sh->tmp, pl_ref_new(sh->ctx));
+    PL_ARRAY_APPEND(sh, sh->tmp, pl_ref_new(NULL));
 }
 
-bool pl_shader_is_failed(const struct pl_shader *sh)
+bool pl_shader_is_failed(const pl_shader sh)
 {
     return sh->failed;
 }
 
-struct pl_glsl_desc sh_glsl(const struct pl_shader *sh)
+struct pl_glsl_version sh_glsl(const pl_shader sh)
 {
     if (SH_PARAMS(sh).glsl.version)
         return SH_PARAMS(sh).glsl;
@@ -93,29 +93,35 @@ struct pl_glsl_desc sh_glsl(const struct pl_shader *sh)
     if (SH_GPU(sh))
         return SH_GPU(sh)->glsl;
 
-    return (struct pl_glsl_desc) { .version = 130 };
+    return (struct pl_glsl_version) { .version = 130 };
 }
 
-bool sh_try_compute(struct pl_shader *sh, int bw, int bh, bool flex, size_t mem)
+bool sh_try_compute(pl_shader sh, int bw, int bh, bool flex, size_t mem)
 {
     pl_assert(bw && bh);
     int *sh_bw = &sh->res.compute_group_size[0];
     int *sh_bh = &sh->res.compute_group_size[1];
 
-    const struct pl_gpu *gpu = SH_GPU(sh);
-    if (!gpu || !(gpu->caps & PL_GPU_CAP_COMPUTE)) {
-        PL_TRACE(sh, "Disabling compute shader due to missing PL_GPU_CAP_COMPUTE");
+    struct pl_glsl_version glsl = sh_glsl(sh);
+    if (!glsl.compute) {
+        PL_TRACE(sh, "Disabling compute shader due to missing `compute` support");
         return false;
     }
 
-    if (sh->res.compute_shmem + mem > gpu->limits.max_shmem_size) {
+    if (sh->res.compute_shmem + mem > glsl.max_shmem_size) {
         PL_TRACE(sh, "Disabling compute shader due to insufficient shmem");
         return false;
     }
 
-    if (bw > gpu->limits.max_group_size[0] ||
-        bh > gpu->limits.max_group_size[1] ||
-        (bw * bh) > gpu->limits.max_group_threads)
+    if (sh->type == SH_FRAGMENT) {
+        PL_TRACE(sh, "Disabling compute shader because shader is already marked "
+                 "as fragment shader");
+        return false;
+    }
+
+    if (bw > glsl.max_group_size[0] ||
+        bh > glsl.max_group_size[1] ||
+        (bw * bh) > glsl.max_group_threads)
     {
         if (!flex) {
             PL_TRACE(sh, "Disabling compute shader due to exceeded group "
@@ -123,8 +129,8 @@ bool sh_try_compute(struct pl_shader *sh, int bw, int bh, bool flex, size_t mem)
             return false;
         } else {
             // Pick better group sizes
-            bw = PL_MIN(bw, gpu->limits.max_group_size[0]);
-            bh = gpu->limits.max_group_threads / bw;
+            bw = PL_MIN(bw, glsl.max_group_size[0]);
+            bh = glsl.max_group_threads / bw;
         }
     }
 
@@ -132,10 +138,10 @@ bool sh_try_compute(struct pl_shader *sh, int bw, int bh, bool flex, size_t mem)
 
     // If the current shader is either not a compute shader, or we have no
     // choice but to override the metadata, always do so
-    if (!sh->is_compute || (sh->flexible_work_groups && !flex)) {
+    if (sh->type != SH_COMPUTE || (sh->flexible_work_groups && !flex)) {
         *sh_bw = bw;
         *sh_bh = bh;
-        sh->is_compute = true;
+        sh->type = SH_COMPUTE;
         return true;
     }
 
@@ -143,7 +149,7 @@ bool sh_try_compute(struct pl_shader *sh, int bw, int bh, bool flex, size_t mem)
     if (sh->flexible_work_groups && flex) {
         *sh_bw = PL_MAX(*sh_bw, bw);
         *sh_bh = PL_MAX(*sh_bh, bh);
-        pl_assert(*sh_bw * *sh_bh <= gpu->limits.max_group_threads);
+        pl_assert(*sh_bw * *sh_bh <= glsl.max_group_threads);
         return true;
     }
 
@@ -163,39 +169,28 @@ bool sh_try_compute(struct pl_shader *sh, int bw, int bh, bool flex, size_t mem)
     return true;
 }
 
-bool pl_shader_is_compute(const struct pl_shader *sh)
+bool pl_shader_is_compute(const pl_shader sh)
 {
-    return sh->is_compute;
+    return sh->type == SH_COMPUTE;
 }
 
-bool pl_shader_output_size(const struct pl_shader *sh, int *w, int *h)
+bool pl_shader_output_size(const pl_shader sh, int *w, int *h)
 {
     if (!sh->output_w || !sh->output_h)
         return false;
 
-    *w = sh->output_w;
-    *h = sh->output_h;
+    *w = sh->transpose ? sh->output_h : sh->output_w;
+    *h = sh->transpose ? sh->output_w : sh->output_h;
     return true;
 }
 
-uint64_t pl_shader_signature(const struct pl_shader *sh)
-{
-    uint64_t res = 0;
-    for (int i = 0; i < PL_ARRAY_SIZE(sh->buffers); i++)
-        res ^= pl_str_hash(sh->buffers[i]);
-
-    // FIXME: also hash in the configuration of the descriptors/variables
-
-    return res;
-}
-
-ident_t sh_fresh(struct pl_shader *sh, const char *name)
+ident_t sh_fresh(pl_shader sh, const char *name)
 {
     return pl_asprintf(SH_TMP(sh), "_%s_%d_%u", PL_DEF(name, "var"),
                        sh->fresh++, SH_PARAMS(sh).id);
 }
 
-ident_t sh_var(struct pl_shader *sh, struct pl_shader_var sv)
+ident_t sh_var(pl_shader sh, struct pl_shader_var sv)
 {
     sv.var.name = sh_fresh(sh, sv.var.name);
     sv.data = pl_memdup(SH_TMP(sh), sv.data, pl_var_host_layout(0, &sv.var).size);
@@ -203,7 +198,13 @@ ident_t sh_var(struct pl_shader *sh, struct pl_shader_var sv)
     return (ident_t) sv.var.name;
 }
 
-ident_t sh_desc(struct pl_shader *sh, struct pl_shader_desc sd)
+static void merge_access(enum pl_desc_access *a, enum pl_desc_access b)
+{
+    if (*a != b)
+        *a = PL_DESC_ACCESS_READWRITE;
+}
+
+ident_t sh_desc(pl_shader sh, struct pl_shader_desc sd)
 {
     switch (sd.desc.type) {
     case PL_DESC_BUF_UNIFORM:
@@ -213,17 +214,25 @@ ident_t sh_desc(struct pl_shader *sh, struct pl_shader_desc sd)
         // Skip re-attaching the same buffer desc twice
         // FIXME: define aliases if the variable names differ
         for (int i = 0; i < sh->descs.num; i++) {
-            if (sh->descs.elem[i].binding.object == sd.binding.object)
+            if (sh->descs.elem[i].binding.object == sd.binding.object) {
+                merge_access(&sh->descs.elem[i].desc.access, sd.desc.access);
+                sh->descs.elem[i].memory |= sd.memory;
                 return (ident_t) sh->descs.elem[i].desc.name;
+            }
         }
 
         size_t bsize = sizeof(sd.buffer_vars[0]) * sd.num_buffer_vars;
         sd.buffer_vars = pl_memdup(SH_TMP(sh), sd.buffer_vars, bsize);
         break;
 
-    default:
+    case PL_DESC_SAMPLED_TEX:
+    case PL_DESC_STORAGE_IMG:
         pl_assert(!sd.num_buffer_vars);
         break;
+
+    case PL_DESC_INVALID:
+    case PL_DESC_TYPE_COUNT:
+        pl_unreachable();
     }
 
     sd.desc.name = sh_fresh(sh, sd.desc.name);
@@ -231,16 +240,87 @@ ident_t sh_desc(struct pl_shader *sh, struct pl_shader_desc sd)
     return (ident_t) sd.desc.name;
 }
 
-ident_t sh_attr_vec2(struct pl_shader *sh, const char *name,
+ident_t sh_const(pl_shader sh, struct pl_shader_const sc)
+{
+    if (sh->res.params.dynamic_constants && !sc.compile_time) {
+        return sh_var(sh, (struct pl_shader_var) {
+            .var = {
+                .name = sc.name,
+                .type = sc.type,
+                .dim_v = 1,
+                .dim_m = 1,
+                .dim_a = 1,
+            },
+            .data = sc.data,
+        });
+    }
+
+    sc.name = sh_fresh(sh, sc.name);
+
+    pl_gpu gpu = SH_GPU(sh);
+    if (gpu && gpu->limits.max_constants) {
+        sc.data = pl_memdup(SH_TMP(sh), sc.data, pl_var_type_size(sc.type));
+        PL_ARRAY_APPEND(sh, sh->consts, sc);
+        return (ident_t) sc.name;
+    }
+
+    // Fallback for GPUs without specialization constants
+    switch (sc.type) {
+    case PL_VAR_SINT:
+        GLSLH("const int %s = %d; \n", sc.name, *(int *) sc.data);
+        return (ident_t) sc.name;
+    case PL_VAR_UINT:
+        GLSLH("const uint %s = %uu; \n", sc.name, *(unsigned int *) sc.data);
+        return (ident_t) sc.name;
+    case PL_VAR_FLOAT:
+        GLSLH("const float %s = %f; \n", sc.name, *(float *) sc.data);
+        return (ident_t) sc.name;
+    case PL_VAR_INVALID:
+    case PL_VAR_TYPE_COUNT:
+        break;
+    }
+
+    pl_unreachable();
+}
+
+ident_t sh_const_int(pl_shader sh, const char *name, int val)
+{
+    return sh_const(sh, (struct pl_shader_const) {
+        .type = PL_VAR_SINT,
+        .name = name,
+        .data = &val,
+    });
+}
+
+ident_t sh_const_uint(pl_shader sh, const char *name, unsigned int val)
+{
+    return sh_const(sh, (struct pl_shader_const) {
+        .type = PL_VAR_UINT,
+        .name = name,
+        .data = &val,
+    });
+}
+
+ident_t sh_const_float(pl_shader sh, const char *name, float val)
+{
+    return sh_const(sh, (struct pl_shader_const) {
+        .type = PL_VAR_FLOAT,
+        .name = name,
+        .data = &val,
+    });
+}
+
+
+ident_t sh_attr_vec2(pl_shader sh, const char *name,
                      const struct pl_rect2df *rc)
 {
-    const struct pl_gpu *gpu = SH_GPU(sh);
+    pl_gpu gpu = SH_GPU(sh);
     if (!gpu) {
         SH_FAIL(sh, "Failed adding vertex attr '%s': No GPU available!", name);
         return NULL;
     }
 
-    const struct pl_fmt *fmt = pl_find_vertex_fmt(gpu, PL_FMT_FLOAT, 2);
+    pl_fmt fmt = pl_find_vertex_fmt(gpu, PL_FMT_FLOAT, 2);
     if (!fmt) {
         SH_FAIL(sh, "Failed adding vertex attr '%s': no vertex fmt!", name);
         return NULL;
@@ -266,19 +346,19 @@ ident_t sh_attr_vec2(struct pl_shader *sh, const char *name,
     return (ident_t) va.attr.name;
 }
 
-ident_t sh_bind(struct pl_shader *sh, const struct pl_tex *tex,
+ident_t sh_bind(pl_shader sh, pl_tex tex,
                 enum pl_tex_address_mode address_mode,
                 enum pl_tex_sample_mode sample_mode,
                 const char *name, const struct pl_rect2df *rect,
                 ident_t *out_pos, ident_t *out_size, ident_t *out_pt)
 {
-    if (!SH_GPU(sh)) {
-        SH_FAIL(sh, "Failed binding texture '%s': No GPU available!", name);
+    if (pl_tex_params_dimension(tex->params) != 2) {
+        SH_FAIL(sh, "Failed binding texture '%s': not a 2D texture!", name);
         return NULL;
     }
 
-    if (pl_tex_params_dimension(tex->params) != 2 || !tex->params.sampleable) {
-        SH_FAIL(sh, "Failed binding texture '%s': incompatible params!", name);
+    if (!tex->params.sampleable) {
+        SH_FAIL(sh, "Failed binding texture '%s': texture not sampleable!", name);
         return NULL;
     }
 
@@ -333,7 +413,7 @@ ident_t sh_bind(struct pl_shader *sh, const struct pl_tex *tex,
     return itex;
 }
 
-bool sh_buf_desc_append(void *alloc, const struct pl_gpu *gpu,
+bool sh_buf_desc_append(void *alloc, pl_gpu gpu,
                         struct pl_shader_desc *buf_desc,
                         struct pl_var_layout *out_layout,
                         const struct pl_var new_var)
@@ -352,7 +432,13 @@ bool sh_buf_desc_append(void *alloc, const struct pl_gpu *gpu,
         if (bv.layout.offset + bv.layout.size > gpu->limits.max_ssbo_size)
             return false;
         break;
-    default: abort();
+    case PL_DESC_INVALID:
+    case PL_DESC_SAMPLED_TEX:
+    case PL_DESC_STORAGE_IMG:
+    case PL_DESC_BUF_TEXEL_UNIFORM:
+    case PL_DESC_BUF_TEXEL_STORAGE:
+    case PL_DESC_TYPE_COUNT:
+        pl_unreachable();
     }
 
     if (out_layout)
@@ -371,7 +457,7 @@ size_t sh_buf_desc_size(const struct pl_shader_desc *buf_desc)
     return last->layout.offset + last->layout.size;
 }
 
-void sh_append(struct pl_shader *sh, enum pl_shader_buf buf, const char *fmt, ...)
+void sh_append(pl_shader sh, enum pl_shader_buf buf, const char *fmt, ...)
 {
     pl_assert(buf >= 0 && buf < SH_BUF_COUNT);
 
@@ -381,7 +467,7 @@ void sh_append(struct pl_shader *sh, enum pl_shader_buf buf, const char *fmt, ..
     va_end(ap);
 }
 
-void sh_append_str(struct pl_shader *sh, enum pl_shader_buf buf, pl_str str)
+void sh_append_str(pl_shader sh, enum pl_shader_buf buf, pl_str str)
 {
     pl_assert(buf >= 0 && buf < SH_BUF_COUNT);
     pl_str_append(sh, &sh->buffers[buf], str);
@@ -409,7 +495,7 @@ static const char *samplers2D[] = {
     [PL_SAMPLER_EXTERNAL]   = "samplerExternalOES",
 };
 
-ident_t sh_subpass(struct pl_shader *sh, const struct pl_shader *sub)
+ident_t sh_subpass(pl_shader sh, const pl_shader sub)
 {
     pl_assert(sh->mutable);
 
@@ -430,7 +516,7 @@ ident_t sh_subpass(struct pl_shader *sh, const struct pl_shader *sub)
         return NULL;
     }
 
-    if (sub->is_compute) {
+    if (sub->type == SH_COMPUTE) {
         int subw = sub->res.compute_group_size[0],
             subh = sub->res.compute_group_size[1];
         bool flex = sub->flexible_work_groups;
@@ -465,15 +551,17 @@ ident_t sh_subpass(struct pl_shader *sh, const struct pl_shader *sub)
     // Copy over all of the descriptors etc.
     for (int i = 0; i < sub->tmp.num; i++)
         PL_ARRAY_APPEND(sh, sh->tmp, pl_ref_dup(sub->tmp.elem[i]));
+    PL_ARRAY_CONCAT(sh, sh->vas, sub->vas);
     PL_ARRAY_CONCAT(sh, sh->vars, sub->vars);
     PL_ARRAY_CONCAT(sh, sh->descs, sub->descs);
-    PL_ARRAY_CONCAT(sh, sh->vas, sub->vas);
+    PL_ARRAY_CONCAT(sh, sh->consts, sub->consts);
+    PL_ARRAY_CONCAT(sh, sh->steps, sub->steps);
 
     return name;
 }
 
 // Finish the current shader body and return its function name
-static ident_t sh_split(struct pl_shader *sh)
+static ident_t sh_split(pl_shader sh)
 {
     pl_assert(sh->mutable);
 
@@ -504,15 +592,13 @@ static ident_t sh_split(struct pl_shader *sh)
     return name;
 }
 
-const struct pl_shader_res *pl_shader_finalize(struct pl_shader *sh)
+const struct pl_shader_res *pl_shader_finalize(pl_shader sh)
 {
     if (sh->failed)
         return NULL;
 
-    if (!sh->mutable) {
-        PL_WARN(sh, "Attempted to finalize a shader twice?");
+    if (!sh->mutable)
         return &sh->res;
-    }
 
     // Split the shader. This finalizes the body and adds it to the header
     sh->res.name = sh_split(sh);
@@ -524,6 +610,38 @@ const struct pl_shader_res *pl_shader_finalize(struct pl_shader *sh)
     pl_str *glsl = &sh->buffers[SH_BUF_PRELUDE];
     pl_str_append(sh, glsl, sh->buffers[SH_BUF_HEADER]);
 
+    // Generate the pretty description
+    sh->res.description = "(unknown shader)";
+    if (sh->steps.num) {
+        // Reuse this buffer
+        pl_str *desc = &sh->buffers[SH_BUF_BODY];
+        desc->len = 0;
+
+        for (int i = 0; i < sh->steps.num; i++) {
+            const char *step = sh->steps.elem[i];
+            if (!step)
+                continue;
+
+            // Group together duplicates. We're okay using a weak equality
+            // check here because all pass descriptions are static strings.
+            int count = 1;
+            for (int j = i+1; j < sh->steps.num; j++) {
+                if (sh->steps.elem[j] == step) {
+                    sh->steps.elem[j] = NULL;
+                    count++;
+                }
+            }
+
+            if (i > 0)
+                pl_str_append(sh, desc, pl_str0(", "));
+            pl_str_append(sh, desc, pl_str0(step));
+            if (count > 1)
+                pl_str_append_asprintf(sh, desc, " x%d", count);
+        }
+
+        sh->res.description = (char *) desc->buf;
+    }
+
     // Set the vas/vars/descs
     sh->res.vertex_attribs = sh->vas.elem;
     sh->res.num_vertex_attribs = sh->vas.num;
@@ -531,14 +649,18 @@ const struct pl_shader_res *pl_shader_finalize(struct pl_shader *sh)
     sh->res.num_variables = sh->vars.num;
     sh->res.descriptors = sh->descs.elem;
     sh->res.num_descriptors = sh->descs.num;
+    sh->res.constants = sh->consts.elem;
+    sh->res.num_constants = sh->consts.num;
+    sh->res.steps = sh->steps.elem;
+    sh->res.num_steps = sh->steps.num;
 
     // Update the result pointer and return
-    sh->res.glsl = glsl->buf;
+    sh->res.glsl = (char *) glsl->buf;
     sh->mutable = false;
     return &sh->res;
 }
 
-bool sh_require(struct pl_shader *sh, enum pl_shader_sig insig, int w, int h)
+bool sh_require(pl_shader sh, enum pl_shader_sig insig, int w, int h)
 {
     if (sh->failed) {
         SH_FAIL(sh, "Attempting to modify a failed shader!");
@@ -583,9 +705,9 @@ bool sh_require(struct pl_shader *sh, enum pl_shader_sig insig, int w, int h)
     return true;
 }
 
-void pl_shader_obj_destroy(struct pl_shader_obj **ptr)
+void pl_shader_obj_destroy(pl_shader_obj *ptr)
 {
-    struct pl_shader_obj *obj = *ptr;
+    pl_shader_obj obj = *ptr;
     if (!obj)
         return;
 
@@ -596,14 +718,14 @@ void pl_shader_obj_destroy(struct pl_shader_obj **ptr)
     pl_free(obj);
 }
 
-void *sh_require_obj(struct pl_shader *sh, struct pl_shader_obj **ptr,
+void *sh_require_obj(pl_shader sh, pl_shader_obj *ptr,
                      enum pl_shader_obj_type type, size_t priv_size,
-                     void (*uninit)(const struct pl_gpu *gpu, void *priv))
+                     void (*uninit)(pl_gpu gpu, void *priv))
 {
     if (!ptr)
         return NULL;
 
-    struct pl_shader_obj *obj = *ptr;
+    pl_shader_obj obj = *ptr;
     if (obj && obj->gpu != SH_GPU(sh)) {
         SH_FAIL(sh, "Passed pl_shader_obj belongs to different GPU!");
         return NULL;
@@ -627,45 +749,77 @@ void *sh_require_obj(struct pl_shader *sh, struct pl_shader_obj **ptr,
     return obj->priv;
 }
 
-ident_t sh_prng(struct pl_shader *sh, bool temporal, ident_t *p_state)
+ident_t sh_prng(pl_shader sh, bool temporal, ident_t *p_state)
 {
-    // Initialize the PRNG. This is friendly for wide usage and returns in
-    // a very pleasant-looking distribution across frames even if the difference
-    // between input coordinates is very small. This is based on BlumBlumShub,
-    // with some modifications for speed / aesthetics.
-    // cf. https://briansharpe.wordpress.com/2011/10/01/gpu-texture-free-noise/
-    ident_t randfun = sh_fresh(sh, "random"), permute = sh_fresh(sh, "permute");
-    GLSLH("float %s(float x) {                      \n"
-          "    x = (34.0 * x + 1.0) * x;            \n"
-          "    return fract(x * 1.0/289.0) * 289.0; \n" // (almost) mod 289
-          "}                                        \n"
-          "float %s(inout float state) {            \n"
-          "    state = %s(state);                   \n"
-          "    return fract(state * 1.0/41.0);      \n"
-          "}\n", permute, randfun, permute);
+    ident_t randfun = sh_fresh(sh, "rand"),
+            state = sh_fresh(sh, "state");
 
-    // Phi is the most irrational number, so it's a good candidate for
-    // generating seed values to the PRNG
-    static const double phi = 1.618033988749895;
+    if (sh_glsl(sh).version >= 130) {
+
+        // Based on pcg3d (http://jcgt.org/published/0009/03/02/)
+        GLSLP("#define prng_t uvec3\n");
+        GLSLH("vec3 %s(inout uvec3 s) {                     \n"
+              "    s = 1664525u * s + uvec3(1013904223u);   \n"
+              "    s.x += s.y * s.z;                        \n"
+              "    s.y += s.z * s.x;                        \n"
+              "    s.z += s.x * s.y;                        \n"
+              "    s ^= s >> 16u;                           \n"
+              "    s.x += s.y * s.z;                        \n"
+              "    s.y += s.z * s.x;                        \n"
+              "    s.z += s.x * s.y;                        \n"
+              "    return vec3(s) * 1.0/float(0xFFFFFFFFu); \n"
+              "}                                            \n",
+              randfun);
+
+        const char *seed = "0u";
+        if (temporal) {
+            seed = sh_var(sh, (struct pl_shader_var) {
+                .var  = pl_var_uint("seed"),
+                .data = &(unsigned int){ SH_PARAMS(sh).index },
+                .dynamic = true,
+            });
+        };
+
+        GLSL("uvec3 %s = uvec3(gl_FragCoord.xy, %s); \n", state, seed);
+
+    } else {
+
+        // Based on SGGP (https://briansharpe.wordpress.com/2011/10/01/gpu-texture-free-noise/)
+        ident_t permute = sh_fresh(sh, "permute");
+        GLSLP("#define prng_t float\n");
+        GLSLH("float %s(float x) {                          \n"
+              "    x = (34.0 * x + 1.0) * x;                \n"
+              "    return fract(x * 1.0/289.0) * 289.0;     \n"
+              "}                                            \n"
+              "vec3 %s(inout float s) {                     \n"
+              "    vec3 ret;                                \n"
+              "    ret.x = %s(s);                           \n"
+              "    ret.y = %s(ret.x);                       \n"
+              "    ret.z = %s(ret.y);                       \n"
+              "    s = ret.z;                               \n"
+              "    return fract(ret * 1.0/41.0);            \n"
+              "}                                            \n",
+              permute, randfun, permute, permute, permute);
+
+        static const double phi = 1.618033988749895;
+        const char *seed = "0.0";
+        if (temporal) {
+            seed = sh_var(sh, (struct pl_shader_var) {
+                .var  = pl_var_float("seed"),
+                .data = &(float){ modff(phi * SH_PARAMS(sh).index, &(float){0}) },
+                .dynamic = true,
+            });
+        };
+
+        GLSL("vec3 %s_m = vec3(fract(gl_FragCoord.xy * vec2(%f)), %s);  \n"
+             "%s_m += vec3(1.0);                                        \n"
+             "float %s = %s(%s(%s(%s_m.x) + %s_m.y) + %s_m.z);          \n",
+             state, phi, seed,
+             state,
+             state, permute, permute, permute, state, state, state);
 
-    const char *seed = "0.0";
-    if (temporal) {
-        float seedval = modff(phi * SH_PARAMS(sh).index, &(float){0});
-        seed = sh_var(sh, (struct pl_shader_var) {
-            .var  = pl_var_float("seed"),
-            .data = &seedval,
-            .dynamic = true,
-        });
     }
 
-    ident_t state = sh_fresh(sh, "prng");
-    GLSL("vec2 %s_init = fract(gl_FragCoord.xy * vec2(%f)); \n"
-         "vec3 %s_m = vec3(%s_init, %s) + vec3(1.0);        \n"
-         "float %s = %s(%s(%s(%s_m.x) + %s_m.y) + %s_m.z);  \n",
-         state, phi,
-         state, state, seed,
-         state, permute, permute, permute, state, state, state);
-
     if (p_state)
         *p_state = state;
 
@@ -677,11 +831,11 @@ ident_t sh_prng(struct pl_shader *sh, bool temporal, ident_t *p_state)
 // Defines a LUT position helper macro. This translates from an absolute texel
 // scale (0.0 - 1.0) to the texture coordinate scale for the corresponding
 // sample in a texture of dimension `lut_size`.
-static ident_t sh_lut_pos(struct pl_shader *sh, int lut_size)
+static ident_t sh_lut_pos(pl_shader sh, int lut_size)
 {
     ident_t name = sh_fresh(sh, "LUT_POS");
-    GLSLH("#define %s(x) mix(%f, %f, (x)) \n",
-          name, 0.5 / lut_size, 1.0 - 0.5 / lut_size);
+    GLSLH("#define %s(x) mix(%s, %s, (x)) \n",
+          name, SH_FLOAT(0.5 / lut_size), SH_FLOAT(1.0 - 0.5 / lut_size));
     return name;
 }
 
@@ -691,14 +845,15 @@ struct sh_lut_obj {
     bool linear;
     int width, height, depth, comps;
     uint64_t signature;
+    bool error; // reset if params change
 
     // weights, depending on the method
-    const struct pl_tex *tex;
+    pl_tex tex;
     pl_str str;
     void *data;
 };
 
-static void sh_lut_uninit(const struct pl_gpu *gpu, void *ptr)
+static void sh_lut_uninit(pl_gpu gpu, void *ptr)
 {
     struct sh_lut_obj *lut = ptr;
     pl_tex_destroy(gpu, &lut->tex);
@@ -709,13 +864,13 @@ static void sh_lut_uninit(const struct pl_gpu *gpu, void *ptr)
 }
 
 // Maximum number of floats to embed as a literal array (when using SH_LUT_AUTO)
-#define SH_LUT_MAX_LITERAL 256
+#define SH_LUT_MAX_LITERAL_SOFT 64
+#define SH_LUT_MAX_LITERAL_HARD 256
 
-ident_t sh_lut(struct pl_shader *sh, const struct sh_lut_params *params)
+ident_t sh_lut(pl_shader sh, const struct sh_lut_params *params)
 {
-    const struct pl_gpu *gpu = SH_GPU(sh);
+    pl_gpu gpu = SH_GPU(sh);
     void *tmp = NULL;
-    ident_t ret = NULL;
 
     pl_assert(params->width > 0 && params->height >= 0 && params->depth >= 0);
     pl_assert(params->comps > 0);
@@ -730,9 +885,23 @@ ident_t sh_lut(struct pl_shader *sh, const struct sh_lut_params *params)
     uint32_t max_tex_dim[] = {
         gpu ? gpu->limits.max_tex_1d_dim : 0,
         gpu ? gpu->limits.max_tex_2d_dim : 0,
-        gpu ? gpu->limits.max_tex_3d_dim : 0,
+        (gpu && gpu->glsl.version > 100) ? gpu->limits.max_tex_3d_dim : 0,
     };
 
+    struct sh_lut_obj *lut = SH_OBJ(sh, params->object, PL_SHADER_OBJ_LUT,
+                                    struct sh_lut_obj, sh_lut_uninit);
+
+    if (!lut)
+        return NULL;
+
+    bool update = params->update || lut->signature != params->signature ||
+                  params->type != lut->type || params->linear != lut->linear ||
+                  params->width != lut->width || params->height != lut->height ||
+                  params->depth != lut->depth || params->comps != lut->comps;
+
+    if (lut->error && !update)
+        return NULL; // suppress error spam until something changes
+
     // Try picking the right number of dimensions for the texture LUT. This
     // allows e.g. falling back to 2D textures if 1D textures are unsupported.
     for (int d = dims; d <= PL_ARRAY_SIZE(max_tex_dim); d++) {
@@ -760,7 +929,7 @@ next_dim: ; // `continue` out of the inner loop
     if (params->linear)
         texcaps |= PL_FMT_CAP_LINEAR;
 
-    const struct pl_fmt *texfmt = NULL;
+    pl_fmt texfmt = NULL;
     if (texdim) {
         texfmt = pl_find_fmt(gpu, fmt_type[params->type], params->comps,
                              params->type == PL_VAR_FLOAT ? 16 : 32,
@@ -768,14 +937,6 @@ next_dim: ; // `continue` out of the inner loop
                              texcaps);
     }
 
-    struct sh_lut_obj *lut = SH_OBJ(sh, params->object, PL_SHADER_OBJ_LUT,
-                                    struct sh_lut_obj, sh_lut_uninit);
-
-    if (!lut) {
-        SH_FAIL(sh, "Failed initializing LUT object!");
-        goto error;
-    }
-
     enum sh_lut_method method = params->method;
 
     // The linear sampling code currently only supports 1D linear interpolation
@@ -783,38 +944,43 @@ next_dim: ; // `continue` out of the inner loop
         if (texfmt) {
             method = SH_LUT_TEXTURE;
         } else {
-            SH_FAIL(sh, "Can't emulate linear LUTs for 2D/3D LUTs and no "
-                    "texture support available!");
+            PL_ERR(sh, "Can't emulate linear LUTs for 2D/3D LUTs and no "
+                  "texture support available!");
             goto error;
         }
     }
 
-    // Older GLSL forbids literal array constructors
-    bool can_literal = sh_glsl(sh).version > 110;
+    bool can_uniform = gpu && gpu->limits.max_variable_comps >= size * params->comps;
+    bool can_literal = sh_glsl(sh).version > 110; // needed for literal arrays
+    can_literal &= size <= SH_LUT_MAX_LITERAL_HARD && !params->dynamic;
 
-    // Pick the best method
-    if (!method && size <= SH_LUT_MAX_LITERAL && !params->dynamic && can_literal)
-        method = SH_LUT_LITERAL; // use literals for small constant LUTs
+    // Deselect unsupported methods
+    if (method == SH_LUT_UNIFORM && !can_uniform)
+        method = SH_LUT_AUTO;
+    if (method == SH_LUT_LITERAL && !can_literal)
+        method = SH_LUT_AUTO;
+    if (method == SH_LUT_TEXTURE && !texfmt)
+        method = SH_LUT_AUTO;
 
+    // Sorted by priority
+    if (!method && can_literal && size <= SH_LUT_MAX_LITERAL_SOFT)
+        method = SH_LUT_LITERAL;
     if (!method && texfmt)
-        method = SH_LUT_TEXTURE; // use textures if a texfmt exists
-
-    // Use an input variable as a last fallback
-    if (!method)
+        method = SH_LUT_TEXTURE;
+    if (!method && can_uniform)
         method = SH_LUT_UNIFORM;
+    if (!method && can_literal)
+        method = SH_LUT_LITERAL;
 
-    // Forcibly reinitialize the existing LUT if needed
-    bool update = params->update || lut->signature != params->signature;
-    if (method != lut->method || params->type != lut->type ||
-        params->linear != lut->linear || params->width != lut->width ||
-        params->height != lut->height || params->depth != lut->depth ||
-        params->comps != lut->comps)
-    {
-        PL_DEBUG(sh, "LUT cache invalidated, regenerating..");
-        update = true;
+    if (!method) {
+        PL_ERR(sh, "Can't generate LUT: no compatible methods!");
+        goto error;
     }
 
+    // Reinitialize the existing LUT if needed
+    update |= method != lut->method;
     if (update) {
+        PL_DEBUG(sh, "LUT cache invalidated, regenerating..");
         size_t buf_size = size * params->comps * pl_var_type_size(params->type);
         tmp = pl_zalloc(NULL, buf_size);
         params->fill(tmp, params);
@@ -822,12 +988,12 @@ next_dim: ; // `continue` out of the inner loop
         switch (method) {
         case SH_LUT_TEXTURE: {
             if (!texdim) {
-                SH_FAIL(sh, "Texture LUT exceeds texture dimensions!");
+                PL_ERR(sh, "Texture LUT exceeds texture dimensions!");
                 goto error;
             }
 
             if (!texfmt) {
-                SH_FAIL(sh, "Found no compatible texture format for LUT!");
+                PL_ERR(sh, "Found no compatible texture format for LUT!");
                 goto error;
             }
 
@@ -839,16 +1005,17 @@ next_dim: ; // `continue` out of the inner loop
                 .sampleable     = true,
                 .host_writable  = params->dynamic,
                 .initial_data   = params->dynamic ? NULL : tmp,
+                .debug_tag      = PL_DEBUG_TAG,
             };
 
             bool ok;
             if (params->dynamic) {
                 ok = pl_tex_recreate(gpu, &lut->tex, &tex_params);
                 if (ok) {
-                    ok = pl_tex_upload(gpu, &(struct pl_tex_transfer_params) {
+                    ok = pl_tex_upload(gpu, pl_tex_transfer_params(
                         .tex = lut->tex,
                         .ptr = tmp,
-                    });
+                    ));
                 }
             } else {
                 // Can't use pl_tex_recreate because of `initial_data`
@@ -858,7 +1025,7 @@ next_dim: ; // `continue` out of the inner loop
             }
 
             if (!ok) {
-                SH_FAIL(sh, "Failed creating LUT texture!");
+                PL_ERR(sh, "Failed creating LUT texture!");
                 goto error;
             }
             break;
@@ -902,7 +1069,9 @@ next_dim: ; // `continue` out of the inner loop
                                                  c > 0 ? "," : "",
                                                  ((int *) tmp)[i+c]);
                         break;
-                    default: abort();
+                    case PL_VAR_INVALID:
+                    case PL_VAR_TYPE_COUNT:
+                        pl_unreachable();
                     }
                 }
                 if (params->comps > 1)
@@ -911,7 +1080,8 @@ next_dim: ; // `continue` out of the inner loop
             break;
         }
 
-        case SH_LUT_AUTO: abort();
+        case SH_LUT_AUTO:
+            pl_unreachable();
         }
 
         lut->method = method;
@@ -1009,7 +1179,8 @@ next_dim: ; // `continue` out of the inner loop
         GLSLH(");\n");
         break;
 
-    default: abort();
+    case SH_LUT_AUTO:
+        pl_unreachable();
     }
 
     if (arr_name) {
@@ -1040,15 +1211,18 @@ next_dim: ; // `continue` out of the inner loop
         }
     }
 
-    pl_assert(name);
-    ret = name;
-    // fall through
-error:
+    lut->error = false;
     pl_free(tmp);
-    return ret;
+    pl_assert(name);
+    return name;
+
+error:
+    lut->error = true;
+    pl_free(tmp);
+    return NULL;
 }
 
-const char *sh_bvec(const struct pl_shader *sh, int dims)
+const char *sh_bvec(const pl_shader sh, int dims)
 {
     static const char *bvecs[] = {
         [1] = "bool",
