@@ -18,12 +18,7 @@
 #include <math.h>
 #include "shaders.h"
 
-const struct pl_deband_params pl_deband_default_params = {
-    .iterations = 1,
-    .threshold  = 4.0,
-    .radius     = 16.0,
-    .grain      = 6.0,
-};
+const struct pl_deband_params pl_deband_default_params = { PL_DEBAND_DEFAULTS };
 
 static inline struct pl_tex_params src_params(const struct pl_sample_src *src)
 {
@@ -44,7 +39,7 @@ enum filter {
 };
 
 // Helper function to compute the src/dst sizes and upscaling ratios
-static bool setup_src(struct pl_shader *sh, const struct pl_sample_src *src,
+static bool setup_src(pl_shader sh, const struct pl_sample_src *src,
                       ident_t *src_tex, ident_t *pos, ident_t *size, ident_t *pt,
                       float *ratio_x, float *ratio_y, uint8_t *comp_mask,
                       float *scale, bool resizeable, const char **fn,
@@ -54,7 +49,7 @@ static bool setup_src(struct pl_shader *sh, const struct pl_sample_src *src,
     float src_w, src_h;
     enum pl_tex_sample_mode sample_mode;
     if (src->tex) {
-        const struct pl_fmt *fmt = src->tex->params.format;
+        pl_fmt fmt = src->tex->params.format;
         bool can_linear = fmt->caps & PL_FMT_CAP_LINEAR;
         pl_assert(pl_tex_params_dimension(src->tex->params) == 2);
         sig = PL_SHADER_SIG_NONE;
@@ -164,7 +159,7 @@ static bool setup_src(struct pl_shader *sh, const struct pl_sample_src *src,
         }
 
         if (fn)
-            *fn = sh_tex_fn(sh, (struct pl_tex_params) { .w = 1, .d = 1 }); // 2D
+            *fn = sh_tex_fn(sh, (struct pl_tex_params) { .w = 1, .h = 1 }); // 2D
 
         sh->sampler_type = src->sampler;
 
@@ -176,7 +171,8 @@ static bool setup_src(struct pl_shader *sh, const struct pl_sample_src *src,
         case PL_FMT_SNORM: sh->sampler_prefix = ' '; break;
         case PL_FMT_UINT: sh->sampler_prefix = 'u'; break;
         case PL_FMT_SINT: sh->sampler_prefix = 's'; break;
-        default: abort();
+        case PL_FMT_TYPE_COUNT:
+            pl_unreachable();
         }
 
         *src_tex = "src_tex";
@@ -186,20 +182,21 @@ static bool setup_src(struct pl_shader *sh, const struct pl_sample_src *src,
     return true;
 }
 
-void pl_shader_deband(struct pl_shader *sh, const struct pl_sample_src *src,
+void pl_shader_deband(pl_shader sh, const struct pl_sample_src *src,
                       const struct pl_deband_params *params)
 {
     float scale;
     ident_t tex, pos, pt;
     const char *fn;
     if (!setup_src(sh, src, &tex, &pos, NULL, &pt, NULL, NULL, NULL, &scale,
-                   true, &fn, LINEAR))
+                   false, &fn, LINEAR))
         return;
 
-    GLSL("vec4 color;\n");
-    GLSL("// pl_shader_deband\n");
-    GLSL("{\n");
     params = PL_DEF(params, &pl_deband_default_params);
+    sh_describe(sh, "debanding");
+    GLSL("vec4 color;           \n"
+         "// pl_shader_deband   \n"
+         "{                     \n");
 
     ident_t prng, state;
     prng = sh_prng(sh, true, &state);
@@ -209,49 +206,52 @@ void pl_shader_deband(struct pl_shader *sh, const struct pl_sample_src *src,
          "color = %s(%s, pos); \n",
          pos, fn, tex);
 
-    // Helper function: Compute a stochastic approximation of the avg color
-    // around a pixel, given a specified radius
-    ident_t average = sh_fresh(sh, "average");
-    GLSLH("vec4 %s(vec2 pos, float range, inout float %s) {     \n"
-          // Compute a random angle and distance
-          "    float dist = %s * range;                         \n"
-          "    float dir  = %s * %f;                            \n"
-          "    vec2 o = dist * vec2(cos(dir), sin(dir));        \n"
-          // Sample at quarter-turn intervals around the source pixel
-          "    vec4 sum = vec4(0.0);                            \n"
-          "    sum += %s(%s, pos + %s * vec2( o.x,  o.y)); \n"
-          "    sum += %s(%s, pos + %s * vec2(-o.x,  o.y)); \n"
-          "    sum += %s(%s, pos + %s * vec2(-o.x, -o.y)); \n"
-          "    sum += %s(%s, pos + %s * vec2( o.x, -o.y)); \n"
-          // Return the (normalized) average
-          "    return 0.25 * sum;                               \n"
-          "}\n",
-          average, state, prng, prng, M_PI * 2,
-          fn, tex, pt, fn, tex, pt, fn, tex, pt, fn, tex, pt);
+    if (params->iterations > 0) {
+        // Helper function: Compute a stochastic approximation of the avg color
+        // around a pixel, given a specified radius
+        ident_t average = sh_fresh(sh, "average");
+        GLSLH("vec4 %s(vec2 pos, float range, inout prng_t %s) {\n"
+              // Compute a random angle and distance
+              "    vec2 dd = %s.xy * vec2(range, %f);           \n"
+              "    vec2 o = dd.x * vec2(cos(dd.y), sin(dd.y));  \n"
+              // Sample at quarter-turn intervals around the source pixel
+              "    vec4 sum = vec4(0.0);                        \n"
+              "    sum += %s(%s, pos + %s * vec2( o.x,  o.y));  \n"
+              "    sum += %s(%s, pos + %s * vec2(-o.x,  o.y));  \n"
+              "    sum += %s(%s, pos + %s * vec2(-o.x, -o.y));  \n"
+              "    sum += %s(%s, pos + %s * vec2( o.x, -o.y));  \n"
+              // Return the (normalized) average
+              "    return 0.25 * sum;                               \n"
+              "}\n",
+              average, state, prng, M_PI * 2,
+              fn, tex, pt, fn, tex, pt, fn, tex, pt, fn, tex, pt);
 
-    // For each iteration, compute the average at a given distance and
-    // pick it instead of the color if the difference is below the threshold.
-    for (int i = 1; i <= params->iterations; i++) {
-        GLSL("avg = %s(pos, %f, %s);                                    \n"
-             "diff = abs(color - avg);                                  \n"
-             "color = mix(avg, color, %s(greaterThan(diff, vec4(%f)))); \n",
-             average, i * params->radius, state,
-             sh_bvec(sh, 4), params->threshold / (1000 * i * scale));
+        ident_t radius = sh_const_float(sh, "radius", params->radius);
+        ident_t threshold = sh_const_float(sh, "threshold",
+                                           params->threshold / (1000 * scale));
+
+        // For each iteration, compute the average at a given distance and
+        // pick it instead of the color if the difference is below the threshold.
+        for (int i = 1; i <= params->iterations; i++) {
+            GLSL("avg = %s(pos, %d.0 * %s, %s);                                     \n"
+                 "diff = abs(color - avg);                                          \n"
+                 "color = mix(avg, color, %s(greaterThan(diff, vec4(%s / %d.0))));  \n",
+                 average, i, radius, state, sh_bvec(sh, 4), threshold, i);
+        }
     }
 
-    GLSL("color *= vec4(%f);\n", scale);
+    GLSL("color *= vec4(%s);\n", SH_FLOAT(scale));
 
     // Add some random noise to smooth out residual differences
     if (params->grain > 0) {
-        GLSL("vec3 noise = vec3(%s, %s, %s);         \n"
-             "color.rgb += %f * (noise - vec3(0.5)); \n",
-             prng, prng, prng, params->grain / 1000.0);
+        GLSL( "color.rgb += %s * (%s - vec3(0.5)); \n",
+             SH_FLOAT(params->grain / 1000.0), prng);
     }
 
     GLSL("}\n");
 }
 
-bool pl_shader_sample_direct(struct pl_shader *sh, const struct pl_sample_src *src)
+bool pl_shader_sample_direct(pl_shader sh, const struct pl_sample_src *src)
 {
     float scale;
     ident_t tex, pos;
@@ -261,12 +261,12 @@ bool pl_shader_sample_direct(struct pl_shader *sh, const struct pl_sample_src *s
         return false;
 
     GLSL("// pl_shader_sample_direct          \n"
-         "vec4 color = vec4(%f) * %s(%s, %s); \n",
-         scale, fn, tex, pos);
+         "vec4 color = vec4(%s) * %s(%s, %s); \n",
+         SH_FLOAT(scale), fn, tex, pos);
     return true;
 }
 
-bool pl_shader_sample_nearest(struct pl_shader *sh, const struct pl_sample_src *src)
+bool pl_shader_sample_nearest(pl_shader sh, const struct pl_sample_src *src)
 {
     float scale;
     ident_t tex, pos;
@@ -275,13 +275,14 @@ bool pl_shader_sample_nearest(struct pl_shader *sh, const struct pl_sample_src *
                    true, &fn, NEAREST))
         return false;
 
+    sh_describe(sh, "nearest");
     GLSL("// pl_shader_sample_nearest         \n"
-         "vec4 color = vec4(%f) * %s(%s, %s); \n",
-         scale, fn, tex, pos);
+         "vec4 color = vec4(%s) * %s(%s, %s); \n",
+         SH_FLOAT(scale), fn, tex, pos);
     return true;
 }
 
-bool pl_shader_sample_bilinear(struct pl_shader *sh, const struct pl_sample_src *src)
+bool pl_shader_sample_bilinear(pl_shader sh, const struct pl_sample_src *src)
 {
     float scale;
     ident_t tex, pos;
@@ -290,13 +291,14 @@ bool pl_shader_sample_bilinear(struct pl_shader *sh, const struct pl_sample_src
                    true, &fn, LINEAR))
         return false;
 
+    sh_describe(sh, "bilinear");
     GLSL("// pl_shader_sample_bilinear        \n"
-         "vec4 color = vec4(%f) * %s(%s, %s); \n",
-         scale, fn, tex, pos);
+         "vec4 color = vec4(%s) * %s(%s, %s); \n",
+         SH_FLOAT(scale), fn, tex, pos);
     return true;
 }
 
-static void bicubic_calcweights(struct pl_shader *sh, const char *t, const char *s)
+static void bicubic_calcweights(pl_shader sh, const char *t, const char *s)
 {
     // Explanation of how bicubic scaling with only 4 texel fetches is done:
     //   http://www.mate.tue.nl/mate/pdfs/10318.pdf
@@ -314,7 +316,7 @@ static void bicubic_calcweights(struct pl_shader *sh, const char *t, const char
          t, s, s);
 }
 
-bool pl_shader_sample_bicubic(struct pl_shader *sh, const struct pl_sample_src *src)
+bool pl_shader_sample_bicubic(pl_shader sh, const struct pl_sample_src *src)
 {
     ident_t tex, pos, size, pt;
     float rx, ry, scale;
@@ -328,8 +330,9 @@ bool pl_shader_sample_bicubic(struct pl_shader *sh, const struct pl_sample_src *
                  "will most likely result in nasty aliasing!");
     }
 
+    sh_describe(sh, "bicubic");
     GLSL("// pl_shader_sample_bicubic                   \n"
-         "vec4 color = vec4(0.0);                       \n"
+         "vec4 color;                                   \n"
          "{                                             \n"
          "vec2 pos  = %s;                               \n"
          "vec2 pt   = %s;                               \n"
@@ -352,13 +355,57 @@ bool pl_shader_sample_bicubic(struct pl_shader *sh, const struct pl_sample_src *
          "vec4 bg = %s(%s, pos + cdelta.zw);        \n"
          "vec4 aa = mix(bg, br, parmy.b);           \n"
          // x-interpolation
-         "color = vec4(%f) * mix(aa, ab, parmx.b);  \n"
+         "color = vec4(%s) * mix(aa, ab, parmx.b);  \n"
          "}                                         \n",
-         fn, tex, fn, tex, fn, tex, fn, tex, scale);
+         fn, tex, fn, tex, fn, tex, fn, tex, SH_FLOAT(scale));
+
     return true;
 }
 
-static bool filter_compat(const struct pl_filter *filter, float inv_scale,
+bool pl_shader_sample_oversample(pl_shader sh, const struct pl_sample_src *src,
+                                 float threshold)
+{
+    ident_t tex, pos, size, pt;
+    float rx, ry, scale;
+    const char *fn;
+    if (!setup_src(sh, src, &tex, &pos, &size, &pt, &rx, &ry, NULL, &scale,
+                   true, &fn, LINEAR))
+        return false;
+
+    ident_t ratio = sh_var(sh, (struct pl_shader_var) {
+        .var = pl_var_vec2("ratio"),
+        .data = &(float[2]) { rx, ry },
+    });
+
+    // Round the position to the nearest pixel
+    sh_describe(sh, "oversample");
+    GLSL("// pl_shader_sample_oversample                \n"
+         "vec4 color;                                   \n"
+         "{                                             \n"
+         "vec2 pt = %s;                                 \n"
+         "vec2 pos = %s - vec2(0.5) * pt;               \n"
+         "vec2 fcoord = fract(pos * %s - vec2(0.5));    \n"
+         "vec2 coeff = fcoord * %s;                     \n",
+         pt, pos, size, ratio);
+
+    if (threshold > 0.0) {
+        threshold = PL_MIN(threshold, 1.0);
+        ident_t thresh = sh_const_float(sh, "threshold", threshold);
+        GLSL("coeff = (coeff - %s) / (1.0 - 2.0 * %s);  \n",
+             thresh, thresh);
+    }
+
+    // Compute the right output blend of colors
+    GLSL("coeff = clamp(coeff, 0.0, 1.0);               \n"
+         "pos += (coeff - fcoord) * pt;                 \n"
+         "color = vec4(%s) * %s(%s, pos);               \n"
+         "}                                             \n",
+         SH_FLOAT(scale), fn, tex);
+
+    return true;
+}
+
+static bool filter_compat(pl_filter filter, float inv_scale,
                           int lut_entries, float cutoff,
                           const struct pl_filter_config *params)
 {
@@ -378,9 +425,9 @@ static bool filter_compat(const struct pl_filter *filter, float inv_scale,
 // If `in` is NULL, samples directly
 // If `in` is set, takes the pixel from inX[idx] where X is the component,
 // `in` is the given identifier, and `idx` must be defined by the caller
-static void polar_sample(struct pl_shader *sh, const struct pl_filter *filter,
-                         const char *fn, ident_t tex, ident_t lut, int x, int y,
-                         uint8_t comp_mask, ident_t in)
+static void polar_sample(pl_shader sh, pl_filter filter, const char *fn,
+                         ident_t tex, ident_t lut, ident_t cutoff, ident_t radius,
+                         int x, int y, uint8_t comp_mask, ident_t in)
 {
     // Since we can't know the subpixel position in advance, assume a
     // worst case scenario
@@ -395,12 +442,12 @@ static void polar_sample(struct pl_shader *sh, const struct pl_filter *filter,
     // Check for samples that might be skippable
     bool maybe_skippable = dmax >= filter->radius_cutoff - M_SQRT2;
     if (maybe_skippable)
-        GLSL("if (d < %f) {\n", filter->radius_cutoff);
+        GLSL("if (d < %s) {\n", cutoff);
 
     // Get the weight for this pixel
-    GLSL("w = %s(d * 1.0/%f); \n"
+    GLSL("w = %s(d * 1.0/%s); \n"
          "wsum += w;          \n",
-         lut, filter->radius);
+         lut, radius);
 
     if (in) {
         for (uint8_t comps = comp_mask; comps;) {
@@ -423,12 +470,12 @@ static void polar_sample(struct pl_shader *sh, const struct pl_filter *filter,
 }
 
 struct sh_sampler_obj {
-    const struct pl_filter *filter;
-    struct pl_shader_obj *lut;
-    struct pl_shader_obj *pass2; // for pl_shader_sample_ortho
+    pl_filter filter;
+    pl_shader_obj lut;
+    pl_shader_obj pass2; // for pl_shader_sample_ortho
 };
 
-static void sh_sampler_uninit(const struct pl_gpu *gpu, void *ptr)
+static void sh_sampler_uninit(pl_gpu gpu, void *ptr)
 {
     struct sh_sampler_obj *obj = ptr;
     pl_shader_obj_destroy(&obj->lut);
@@ -440,14 +487,13 @@ static void sh_sampler_uninit(const struct pl_gpu *gpu, void *ptr)
 static void fill_polar_lut(void *data, const struct sh_lut_params *params)
 {
     const struct sh_sampler_obj *obj = params->priv;
-    const struct pl_filter *filt = obj->filter;
+    pl_filter filt = obj->filter;
 
     pl_assert(params->width == filt->params.lut_entries && params->comps == 1);
     memcpy(data, filt->weights, params->width * sizeof(float));
 }
 
-bool pl_shader_sample_polar(struct pl_shader *sh,
-                            const struct pl_sample_src *src,
+bool pl_shader_sample_polar(pl_shader sh, const struct pl_sample_src *src,
                             const struct pl_sample_filter_params *params)
 {
     pl_assert(params);
@@ -456,10 +502,7 @@ bool pl_shader_sample_polar(struct pl_shader *sh,
         return false;
     }
 
-    const struct pl_gpu *gpu = SH_GPU(sh);
-    pl_assert(gpu);
-
-    bool has_compute = gpu->caps & PL_GPU_CAP_COMPUTE && !params->no_compute;
+    bool has_compute = sh_glsl(sh).compute && !params->no_compute;
     has_compute &= sh_glsl(sh).version >= 130; // needed for round()
     if (!src->tex && has_compute) {
         // FIXME: Could maybe solve this by communicating the wbase from
@@ -510,12 +553,12 @@ bool pl_shader_sample_polar(struct pl_shader *sh,
 
     if (update) {
         pl_filter_free(&obj->filter);
-        obj->filter = pl_filter_generate(sh->ctx, &(struct pl_filter_params) {
+        obj->filter = pl_filter_generate(sh->log, pl_filter_params(
             .config         = params->filter,
             .lut_entries    = lut_entries,
             .filter_scale   = inv_scale,
             .cutoff         = cutoff,
-        });
+        ));
 
         if (!obj->filter) {
             // This should never happen, but just in case ..
@@ -524,12 +567,14 @@ bool pl_shader_sample_polar(struct pl_shader *sh,
         }
     }
 
+    sh_describe(sh, "polar scaling");
     GLSL("// pl_shader_sample_polar                     \n"
          "vec4 color = vec4(0.0);                       \n"
          "{                                             \n"
          "vec2 pos = %s, size = %s, pt = %s;            \n"
          "vec2 fcoord = fract(pos * size - vec2(0.5));  \n"
          "vec2 base = pos - pt * fcoord;                \n"
+         "vec2 center = base + pt * vec2(0.5);          \n"
          "float w, d, wsum = 0.0;                       \n"
          "int idx;                                      \n"
          "vec4 c;                                       \n",
@@ -542,7 +587,7 @@ bool pl_shader_sample_polar(struct pl_shader *sh,
     // Determined experimentally on modern AMD and Nvidia hardware. 32 is a
     // good tradeoff for the horizontal work group size. Apart from that,
     // just use as many threads as possible.
-    const int bw = 32, bh = gpu->limits.max_group_threads / bw;
+    const int bw = 32, bh = sh_glsl(sh).max_group_threads / bw;
 
     // We need to sample everything from base_min to base_max, so make sure
     // we have enough room in shmem
@@ -558,7 +603,7 @@ bool pl_shader_sample_polar(struct pl_shader *sh,
     // using a texture-based LUT is better. For the fragment shader fallback
     // code, which is primarily texture bound, the extra cost of LUT
     // interpolation is worth the reduction in texel fetches.
-    ident_t lut = sh_lut(sh, &(struct sh_lut_params) {
+    ident_t lut = sh_lut(sh, sh_lut_params(
         .object = &obj->lut,
         .method = is_compute ? SH_LUT_TEXTURE : SH_LUT_AUTO,
         .type = PL_VAR_FLOAT,
@@ -568,13 +613,16 @@ bool pl_shader_sample_polar(struct pl_shader *sh,
         .update = update,
         .fill = fill_polar_lut,
         .priv = obj,
-    });
+    ));
 
     if (!lut) {
         SH_FAIL(sh, "Failed initializing polar LUT!");
         return false;
     }
 
+    ident_t cutoff_c = sh_const_float(sh, "radius_cutoff", obj->filter->radius_cutoff);
+    ident_t radius_c = sh_const_float(sh, "radius", obj->filter->radius);
+
     if (is_compute) {
         // Compute shader kernel
         GLSL("vec2 wpos = %s_map(gl_WorkGroupID * gl_WorkGroupSize);        \n"
@@ -582,17 +630,31 @@ bool pl_shader_sample_polar(struct pl_shader *sh,
              "ivec2 rel = ivec2(round((base - wbase) * size));              \n",
              pos);
 
+        ident_t iw_c = sh_const(sh, (struct pl_shader_const) {
+            .type = PL_VAR_SINT,
+            .compile_time = true,
+            .name ="iw",
+            .data = &iw,
+        });
+
+        ident_t ih_c = sh_const(sh, (struct pl_shader_const) {
+            .type = PL_VAR_SINT,
+            .compile_time = true,
+            .name = "ih",
+            .data = &ih,
+        });
+
         // Load all relevant texels into shmem
-        GLSL("for (int y = int(gl_LocalInvocationID.y); y < %d; y += %d) {  \n"
-             "for (int x = int(gl_LocalInvocationID.x); x < %d; x += %d) {  \n"
+        GLSL("for (int y = int(gl_LocalInvocationID.y); y < %s; y += %d) {  \n"
+             "for (int x = int(gl_LocalInvocationID.x); x < %s; x += %d) {  \n"
              "c = %s(%s, wbase + pt * vec2(x - %d, y - %d));                \n",
-             ih, bh, iw, bw, fn, src_tex, offset, offset);
+             ih_c, bh, iw_c, bw, fn, src_tex, offset, offset);
 
         in = sh_fresh(sh, "in");
         for (uint8_t comps = comp_mask; comps;) {
             uint8_t c = __builtin_ctz(comps);
-            GLSLH("shared float %s%d[%d];   \n", in, c, ih * iw);
-            GLSL("%s%d[%d * y + x] = c[%d]; \n", in, c, iw, c);
+            GLSLH("shared float %s%d[%s * %s]; \n", in, c, ih_c, iw_c);
+            GLSL("%s%d[%s * y + x] = c[%d]; \n", in, c, iw_c, c);
             comps &= ~(1 << c);
         }
 
@@ -602,9 +664,10 @@ bool pl_shader_sample_polar(struct pl_shader *sh,
         // Dispatch the actual samples
         for (int y = 1 - bound; y <= bound; y++) {
             for (int x = 1 - bound; x <= bound; x++) {
-                GLSL("idx = %d * rel.y + rel.x + %d;\n",
-                     iw, iw * (y + offset) + x + offset);
-                polar_sample(sh, obj->filter, fn, src_tex, lut, x, y, comp_mask, in);
+                GLSL("idx = %s * rel.y + rel.x + %s * %d + %d; \n",
+                     iw_c, iw_c, y + offset, x + offset);
+                polar_sample(sh, obj->filter, fn, src_tex, lut, cutoff_c, radius_c,
+                             x, y, comp_mask, in);
             }
         }
     } else {
@@ -615,40 +678,77 @@ bool pl_shader_sample_polar(struct pl_shader *sh,
             comps &= ~(1 << c);
         }
 
-        // Iterate over the LUT space in groups of 4 texels at a time, and
-        // decide for each texel group whether to use gathering or direct
-        // sampling.
-        for (int y = 1 - bound; y <= bound; y += 2) {
-            for (int x = 1 - bound; x <= bound; x += 2) {
+        // For maximum efficiency, we want to use textureGather() if
+        // possible, rather than direct sampling. Since this is not
+        // always possible/sensible, we need to possibly intermix gathering
+        // with regular sampling. This requires keeping track of which
+        // pixels in the next row were already gathered by the previous
+        // row.
+        uint32_t gathered_cur = 0x0, gathered_next = 0x0;
+        const float radius2 = PL_SQUARE(obj->filter->radius_cutoff);
+        const int base = bound - 1;
+
+        if (base + bound >= 8 * sizeof(gathered_cur)) {
+            SH_FAIL(sh, "Polar radius %f exceeds implementation capacity!",
+                    obj->filter->radius_cutoff);
+            return false;
+        }
+
+        for (int y = 1 - bound; y <= bound; y++) {
+            for (int x = 1 - bound; x <= bound; x++) {
+                // Skip already gathered texels
+                uint32_t bit = 1llu << (base + x);
+                if (gathered_cur & bit)
+                    continue;
+
                 // Using texture gathering is only more efficient than direct
                 // sampling in the case where we expect to be able to use all
                 // four gathered texels, without having to discard any. So
-                // only do it if we suspsect it will be a win rather than a
+                // only do it if we suspect it will be a win rather than a
                 // loss.
-                bool use_gather = sqrt(x*x + y*y) < obj->filter->radius_cutoff;
+                int xx = x*x, xx1 = (x+1)*(x+1);
+                int yy = y*y, yy1 = (y+1)*(y+1);
+                bool use_gather = PL_MAX(xx, xx1) + PL_MAX(yy, yy1) < radius2;
+                use_gather &= PL_MAX(x, y) <= sh_glsl(sh).max_gather_offset;
+                use_gather &= PL_MIN(x, y) >= sh_glsl(sh).min_gather_offset;
+                use_gather &= !src->tex || src->tex->params.format->gatherable;
 
-                // Make sure all required features are supported
-                use_gather &= sh_glsl(sh).version >= 400;
-                use_gather &= gpu->limits.max_gather_offset != 0;
-                use_gather &= PL_MAX(x, y) <= gpu->limits.max_gather_offset;
-                use_gather &= PL_MIN(x, y) >= gpu->limits.min_gather_offset;
+                // Gathering from components other than the R channel requires
+                // support for GLSL 400, which introduces the overload of
+                // textureGather* that allows specifying the component.
+                //
+                // This is also the minimum requirement if we don't know the
+                // texture format capabilities, for the sampler2D interface
+                if (comp_mask != 0x1 || !src->tex)
+                    use_gather &= sh_glsl(sh).version >= 400;
 
                 if (!use_gather) {
                     // Switch to direct sampling instead
-                    for (int yy = y; yy <= bound && yy <= y + 1; yy++) {
-                        for (int xx = x; xx <= bound && xx <= x + 1; xx++) {
-                            polar_sample(sh, obj->filter, fn, src_tex, lut,
-                                         xx, yy, comp_mask, NULL);
-                        }
-                    }
-                    continue; // next group of 4
+                    polar_sample(sh, obj->filter, fn, src_tex, lut, cutoff_c,
+                                 radius_c, x, y, comp_mask, NULL);
+                    continue;
                 }
 
                 // Gather the four surrounding texels simultaneously
                 for (uint8_t comps = comp_mask; comps;) {
                     uint8_t c = __builtin_ctz(comps);
-                    GLSL("in%d = textureGatherOffset(%s, base, "
-                         "ivec2(%d, %d), %d);\n", c, src_tex, x, y, c);
+                    if (x || y) {
+                        if (c) {
+                            GLSL("in%d = textureGatherOffset(%s, center, "
+                                 "ivec2(%d, %d), %d);\n",
+                                 c, src_tex, x, y, c);
+                        } else {
+                            GLSL("in0 = textureGatherOffset(%s, center, "
+                                 "ivec2(%d, %d));\n", src_tex, x, y);
+                        }
+                    } else {
+                        if (c) {
+                            GLSL("in%d = textureGather(%s, center, %d);\n",
+                                 c, src_tex, c);
+                        } else {
+                            GLSL("in0 = textureGather(%s, center);\n", src_tex);
+                        }
+                    }
                     comps &= ~(1 << c);
                 }
 
@@ -662,30 +762,40 @@ bool pl_shader_sample_polar(struct pl_shader *sh,
                         continue; // next subpixel
 
                     GLSL("idx = %d;\n", p);
-                    polar_sample(sh, obj->filter, fn, src_tex, lut,
-                                 x+xo[p], y+yo[p], comp_mask, "in");
+                    polar_sample(sh, obj->filter, fn, src_tex, lut, cutoff_c,
+                                 radius_c, x+xo[p], y+yo[p], comp_mask, "in");
                 }
+
+                // Mark the other next row's pixels as already gathered
+                gathered_next |= bit | (bit << 1);
+                x++; // skip adjacent pixel
             }
+
+            // Prepare for new row
+            gathered_cur = gathered_next;
+            gathered_next = 0;
         }
     }
 
-    GLSL("color = vec4(%f / wsum) * color; \n"
-         "}                                \n",
-         scale);
+    GLSL("color = vec4(%s / wsum) * color; \n", SH_FLOAT(scale));
+    if (!(comp_mask & (1 << PL_CHANNEL_A)))
+        GLSL("color.a = 1.0; \n");
+
+    GLSL("}\n");
     return true;
 }
 
 static void fill_ortho_lut(void *data, const struct sh_lut_params *params)
 {
     const struct sh_sampler_obj *obj = params->priv;
-    const struct pl_filter *filt = obj->filter;
+    pl_filter filt = obj->filter;
     size_t entries = filt->params.lut_entries * filt->row_stride;
 
     pl_assert(params->width * params->height * params->comps == entries);
     memcpy(data, filt->weights, entries * sizeof(float));
 }
 
-bool pl_shader_sample_ortho(struct pl_shader *sh, int pass,
+bool pl_shader_sample_ortho(pl_shader sh, int pass,
                             const struct pl_sample_src *src,
                             const struct pl_sample_filter_params *params)
 {
@@ -695,7 +805,7 @@ bool pl_shader_sample_ortho(struct pl_shader *sh, int pass,
         return false;
     }
 
-    const struct pl_gpu *gpu = SH_GPU(sh);
+    pl_gpu gpu = SH_GPU(sh);
     pl_assert(gpu);
 
     struct pl_sample_src srcfix = *src;
@@ -708,9 +818,6 @@ bool pl_shader_sample_ortho(struct pl_shader *sh, int pass,
         srcfix.rect.y0 = 0;
         srcfix.rect.y1 = srcfix.new_h = src_params(src).h;
         break;
-    case PL_SEP_PASSES:
-    default:
-        abort();
     }
 
     uint8_t comp_mask;
@@ -754,13 +861,13 @@ bool pl_shader_sample_ortho(struct pl_shader *sh, int pass,
 
     if (update) {
         pl_filter_free(&obj->filter);
-        obj->filter = pl_filter_generate(sh->ctx, &(struct pl_filter_params) {
+        obj->filter = pl_filter_generate(sh->log, pl_filter_params(
             .config             = params->filter,
             .lut_entries        = lut_entries,
             .filter_scale       = inv_scale,
             .max_row_size       = gpu->limits.max_tex_2d_dim / 4,
             .row_stride_align   = 4,
-        });
+        ));
 
         if (!obj->filter) {
             // This should never happen, but just in case ..
@@ -771,7 +878,7 @@ bool pl_shader_sample_ortho(struct pl_shader *sh, int pass,
 
     int N = obj->filter->row_size; // number of samples to convolve
     int width = obj->filter->row_stride / 4; // width of the LUT texture
-    ident_t lut = sh_lut(sh, &(struct sh_lut_params) {
+    ident_t lut = sh_lut(sh, sh_lut_params(
         .object = &obj->lut,
         .type = PL_VAR_FLOAT,
         .width = width,
@@ -781,22 +888,28 @@ bool pl_shader_sample_ortho(struct pl_shader *sh, int pass,
         .update = update,
         .fill = fill_ortho_lut,
         .priv = obj,
-    });
+    ));
     if (!lut) {
         SH_FAIL(sh, "Failed initializing separated LUT!");
         return false;
     }
 
-    const float dir[PL_SEP_PASSES][2] = {
-        [PL_SEP_HORIZ] = {1.0, 0.0},
-        [PL_SEP_VERT]  = {0.0, 1.0},
+    const int dir[PL_SEP_PASSES][2] = {
+        [PL_SEP_HORIZ] = {1, 0},
+        [PL_SEP_VERT]  = {0, 1},
     };
 
+    static const char *names[PL_SEP_PASSES] = {
+        [PL_SEP_HORIZ] = "ortho scaling (horiz)",
+        [PL_SEP_VERT]  = "ortho scaling (vert)",
+    };
+
+    sh_describe(sh, names[pass]);
     GLSL("// pl_shader_sample_ortho                        \n"
          "vec4 color = vec4(0.0);                          \n"
          "{                                                \n"
          "vec2 pos = %s, size = %s, pt = %s;               \n"
-         "vec2 dir = vec2(%f, %f);                         \n"
+         "vec2 dir = vec2(%d.0, %d.0);                     \n"
          "pt *= dir;                                       \n"
          "vec2 fcoord2 = fract(pos * size - vec2(0.5));    \n"
          "float fcoord = dot(fcoord2, dir);                \n"
@@ -842,11 +955,14 @@ bool pl_shader_sample_ortho(struct pl_shader *sh, int pass,
     }
 
     if (use_ar) {
-        GLSL("color = mix(color, clamp(color, lo, hi), %f);\n",
-             params->antiring);
+        GLSL("color = mix(color, clamp(color, lo, hi), %s);\n",
+             sh_const_float(sh, "antiring", params->antiring));
     }
 
-    GLSL("color *= vec4(%f);\n", scale);
+    GLSL("color *= vec4(%s);\n", SH_FLOAT(scale));
+    if (!(comp_mask & (1 << PL_CHANNEL_A)))
+        GLSL("color.a = 1.0; \n");
+
     GLSL("}\n");
     return true;
 }
