@@ -124,13 +124,21 @@ int main()
         struct pl_color_repr repr = {
             .levels = PL_COLOR_LEVELS_LIMITED,
             .sys = sys,
+            .bits = {
+                // synthetic test
+                .color_depth = 8,
+                .sample_depth = 10,
+            },
         };
 
+        float scale = pl_color_repr_normalize(&repr);
         struct pl_transform3x3 yuv2rgb = pl_color_repr_decode(&repr, NULL);
-        static const float white_ycbcr[3] = { 235/255., 128/255., 128/255. };
-        static const float black_ycbcr[3] = {  16/255., 128/255., 128/255. };
-        static const float white_other[3] = { 235/255., 235/255., 235/255. };
-        static const float black_other[3] = {  16/255.,  16/255.,  16/255. };
+        pl_matrix3x3_scale(&yuv2rgb.mat, scale);
+
+        static const float white_ycbcr[3] = { 235/1023., 128/1023., 128/1023. };
+        static const float black_ycbcr[3] = {  16/1023., 128/1023., 128/1023. };
+        static const float white_other[3] = { 235/1023., 235/1023., 235/1023. };
+        static const float black_other[3] = {  16/1023.,  16/1023.,  16/1023. };
 
         float white[3], black[3];
         for (int i = 0; i < 3; i++) {
@@ -234,19 +242,16 @@ int main()
     };
 
     pl_color_space_infer(&hlg);
-    REQUIRE(hlg.light == PL_COLOR_LIGHT_SCENE_HLG);
+    REQUIRE(hlg.hdr.max_luma == 1000.0f);
 
     struct pl_color_space unknown = {0};
     struct pl_color_space display = {
         .primaries = PL_COLOR_PRIM_BT_709,
-        .transfer = PL_COLOR_TRC_GAMMA22,
-        .light = PL_COLOR_LIGHT_DISPLAY,
-        .sig_peak = 1.0,
-        .sig_avg = 0.25,
-        .sig_scale = 1.0,
+        .transfer = PL_COLOR_TRC_BT_1886,
     };
 
     pl_color_space_infer(&unknown);
+    pl_color_space_infer(&display);
     REQUIRE(pl_color_space_equal(&unknown, &display));
 
     float x, y;
