@@ -2,17 +2,18 @@
 
 int main()
 {
-    struct pl_context *ctx = pl_test_context();
-    for (const struct pl_named_filter_config *conf = pl_named_filters;
-         conf->filter; conf++)
-    {
+    pl_log log = pl_test_logger();
+    for (const struct pl_filter_preset *conf = pl_filter_presets; conf->name; conf++) {
+        if (!conf->filter)
+            continue;
+
         struct pl_filter_params params = {
             .config      = *conf->filter,
             .lut_entries = 128,
         };
 
         printf("Testing filter '%s'\n", conf->name);
-        const struct pl_filter *flt = pl_filter_generate(ctx, &params);
+        pl_filter flt = pl_filter_generate(log, &params);
         REQUIRE(flt);
 
         if (params.config.polar) {
@@ -35,5 +36,5 @@ int main()
 
         pl_filter_free(&flt);
     }
-    pl_context_destroy(&ctx);
+    pl_log_destroy(&log);
 }
