@@ -1,7 +1,7 @@
 #include "tests.h"
 #include "shaders.h"
 
-static void pl_buffer_tests(const struct pl_gpu *gpu)
+static void pl_buffer_tests(pl_gpu gpu)
 {
     const size_t buf_size = 1024;
     if (buf_size > gpu->limits.max_buf_size)
@@ -14,7 +14,7 @@ static void pl_buffer_tests(const struct pl_gpu *gpu)
     for (int i = 0; i < buf_size; i++)
         test_src[i] = (RANDOM * 256);
 
-    const struct pl_buf *buf = NULL, *tbuf = NULL;
+    pl_buf buf = NULL, tbuf = NULL;
 
     printf("test buffer static creation and readback\n");
     buf = pl_buf_create(gpu, &(struct pl_buf_params) {
@@ -61,7 +61,7 @@ static void pl_buffer_tests(const struct pl_gpu *gpu)
     pl_buf_destroy(gpu, &buf);
     pl_buf_destroy(gpu, &tbuf);
 
-    if (gpu->caps & PL_GPU_CAP_MAPPED_BUFFERS) {
+    if (buf_size <= gpu->limits.max_mapped_size) {
         printf("test host mapped buffer readback\n");
         buf = pl_buf_create(gpu, &(struct pl_buf_params) {
             .size = buf_size,
@@ -84,7 +84,7 @@ static void test_cb(void *priv)
     *flag = true;
 }
 
-static void pl_test_roundtrip(const struct pl_gpu *gpu, const struct pl_tex *tex[2],
+static void pl_test_roundtrip(pl_gpu gpu, pl_tex tex[2],
                               uint8_t *src, uint8_t *dst)
 {
     if (!tex[0] || !tex[1]) {
@@ -96,7 +96,7 @@ static void pl_test_roundtrip(const struct pl_gpu *gpu, const struct pl_tex *tex
     texels *= tex[0]->params.h ? tex[0]->params.h : 1;
     texels *= tex[0]->params.d ? tex[0]->params.d : 1;
 
-    const struct pl_fmt *fmt = tex[0]->params.format;
+    pl_fmt fmt = tex[0]->params.format;
     size_t bytes = texels * fmt->texel_size;
     memset(src, 0, bytes);
     memset(dst, 0, bytes);
@@ -104,7 +104,7 @@ static void pl_test_roundtrip(const struct pl_gpu *gpu, const struct pl_tex *tex
     for (size_t i = 0; i < bytes; i++)
         src[i] = (RANDOM * 256);
 
-    struct pl_timer *ul, *dl;
+    pl_timer ul, dl;
     ul = pl_timer_create(gpu);
     dl = pl_timer_create(gpu);
 
@@ -114,14 +114,14 @@ static void pl_test_roundtrip(const struct pl_gpu *gpu, const struct pl_tex *tex
         .tex = tex[0],
         .ptr = src,
         .timer = ul,
-        .callback = (gpu->caps & PL_GPU_CAP_CALLBACKS) ? test_cb : NULL,
+        .callback = gpu->limits.callbacks ? test_cb : NULL,
         .priv = &ran_ul,
     }));
 
     // Test blitting, if possible for this format
-    const struct pl_tex *dst_tex = tex[0];
+    pl_tex dst_tex = tex[0];
     if (tex[0]->params.blit_src && tex[1]->params.blit_dst) {
-        pl_tex_clear(gpu, tex[1], (float[4]){0.0}); // for testing
+        pl_tex_clear_ex(gpu, tex[1], (union pl_clear_color){0}); // for testing
         pl_tex_blit(gpu, &(struct pl_tex_blit_params) {
             .src = tex[0],
             .dst = tex[1],
@@ -133,12 +133,12 @@ static void pl_test_roundtrip(const struct pl_gpu *gpu, const struct pl_tex *tex
         .tex = dst_tex,
         .ptr = dst,
         .timer = dl,
-        .callback = (gpu->caps & PL_GPU_CAP_CALLBACKS) ? test_cb : NULL,
+        .callback = gpu->limits.callbacks ? test_cb : NULL,
         .priv = &ran_dl,
     }));
 
     pl_gpu_finish(gpu);
-    if (gpu->caps & PL_GPU_CAP_CALLBACKS)
+    if (gpu->limits.callbacks)
         REQUIRE(ran_ul && ran_dl);
 
     if (fmt->emulated && fmt->type == PL_FMT_FLOAT) {
@@ -157,14 +157,14 @@ static void pl_test_roundtrip(const struct pl_gpu *gpu, const struct pl_tex *tex
     pl_timer_destroy(gpu, &dl);
 }
 
-static void pl_texture_tests(const struct pl_gpu *gpu)
+static void pl_texture_tests(pl_gpu gpu)
 {
     const size_t max_size = 16*16*16 * 4 *sizeof(double);
     uint8_t *test_src = malloc(max_size * 2);
     uint8_t *test_dst = test_src + max_size;
 
     for (int f = 0; f < gpu->num_formats; f++) {
-        const struct pl_fmt *fmt = gpu->formats[f];
+        pl_fmt fmt = gpu->formats[f];
         if (fmt->opaque || !(fmt->caps & PL_FMT_CAP_HOST_READABLE))
             continue;
 
@@ -177,15 +177,16 @@ static void pl_texture_tests(const struct pl_gpu *gpu)
             .blit_dst      = (fmt->caps & PL_FMT_CAP_BLITTABLE),
             .host_writable = true,
             .host_readable = true,
+            .debug_tag     = PL_DEBUG_TAG,
         };
 
-        const struct pl_tex *tex[2];
+        pl_tex tex[2];
 
         if (gpu->limits.max_tex_1d_dim >= 16) {
             printf("... 1D\n");
             struct pl_tex_params params = ref_params;
             params.w = 16;
-            if (!(gpu->caps & PL_GPU_CAP_BLITTABLE_1D_3D))
+            if (!gpu->limits.blittable_1d_3d)
                 params.blit_src = params.blit_dst = false;
             for (int i = 0; i < PL_ARRAY_SIZE(tex); i++)
                 tex[i] = pl_tex_create(gpu, &params);
@@ -209,7 +210,7 @@ static void pl_texture_tests(const struct pl_gpu *gpu)
             printf("... 3D\n");
             struct pl_tex_params params = ref_params;
             params.w = params.h = params.d = 16;
-            if (!(gpu->caps & PL_GPU_CAP_BLITTABLE_1D_3D))
+            if (!gpu->limits.blittable_1d_3d)
                 params.blit_src = params.blit_dst = false;
             for (int i = 0; i < PL_ARRAY_SIZE(tex); i++)
                 tex[i] = pl_tex_create(gpu, &params);
@@ -222,7 +223,7 @@ static void pl_texture_tests(const struct pl_gpu *gpu)
     free(test_src);
 }
 
-static void pl_shader_tests(const struct pl_gpu *gpu)
+static void pl_shader_tests(pl_gpu gpu)
 {
     if (gpu->glsl.version < 410)
         return;
@@ -245,7 +246,7 @@ static void pl_shader_tests(const struct pl_gpu *gpu)
         "    out_color = vec4(frag_color, 1.0);     \n"
         "}";
 
-    const struct pl_fmt *fbo_fmt;
+    pl_fmt fbo_fmt;
     enum pl_fmt_caps caps = PL_FMT_CAP_RENDERABLE | PL_FMT_CAP_BLITTABLE |
                             PL_FMT_CAP_LINEAR;
 
@@ -256,7 +257,7 @@ static void pl_shader_tests(const struct pl_gpu *gpu)
 #define FBO_W 16
 #define FBO_H 16
 
-    const struct pl_tex *fbo;
+    pl_tex fbo;
     fbo = pl_tex_create(gpu, &(struct pl_tex_params) {
         .format         = fbo_fmt,
         .w              = FBO_W,
@@ -268,9 +269,9 @@ static void pl_shader_tests(const struct pl_gpu *gpu)
     });
     REQUIRE(fbo);
 
-    pl_tex_clear(gpu, fbo, (float[4]){0});
+    pl_tex_clear_ex(gpu, fbo, (union pl_clear_color){0});
 
-    const struct pl_fmt *vert_fmt;
+    pl_fmt vert_fmt;
     vert_fmt = pl_find_vertex_fmt(gpu, PL_FMT_FLOAT, 3);
     REQUIRE(vert_fmt);
 
@@ -281,10 +282,10 @@ static void pl_shader_tests(const struct pl_gpu *gpu)
         {{ 1.0,  1.0}, {1, 1, 0}},
     };
 
-    const struct pl_pass *pass;
+    pl_pass pass;
     pass = pl_pass_create(gpu, &(struct pl_pass_params) {
         .type           = PL_PASS_RASTER,
-        .target_dummy   = *fbo,
+        .target_format  = fbo_fmt,
         .vertex_shader  = vert_shader,
         .glsl_shader    = frag_shader,
 
@@ -304,9 +305,13 @@ static void pl_shader_tests(const struct pl_gpu *gpu)
         }},
     });
     REQUIRE(pass);
-    REQUIRE(pass->params.cached_program_len);
+    if (pass->params.cached_program || pass->params.cached_program_len) {
+        // Ensure both are set if either one is set
+        REQUIRE(pass->params.cached_program);
+        REQUIRE(pass->params.cached_program_len);
+    }
 
-    struct pl_timer *timer = pl_timer_create(gpu);
+    pl_timer timer = pl_timer_create(gpu);
     pl_pass_run(gpu, &(struct pl_pass_run_params) {
         .pass           = pass,
         .target         = fbo,
@@ -346,7 +351,7 @@ static void pl_shader_tests(const struct pl_gpu *gpu)
 
     if (sizeof(vertices) <= gpu->limits.max_vbo_size) {
         // Test the use of an explicit vertex buffer
-        const struct pl_buf *vert = pl_buf_create(gpu, &(struct pl_buf_params) {
+        pl_buf vert = pl_buf_create(gpu, &(struct pl_buf_params) {
             .size = sizeof(vertices),
             .initial_data = vertices,
             .drawable = true,
@@ -379,8 +384,8 @@ static void pl_shader_tests(const struct pl_gpu *gpu)
     TEST_FBO_PATTERN(1e-6, "%s", "using indexed rendering");
 
     // Test the use of pl_dispatch
-    struct pl_dispatch *dp = pl_dispatch_create(gpu->ctx, gpu);
-    struct pl_shader *sh = pl_dispatch_begin(dp);
+    pl_dispatch dp = pl_dispatch_create(gpu->log, gpu);
+    pl_shader sh = pl_dispatch_begin(dp);
     REQUIRE(pl_shader_custom(sh, &(struct pl_custom_shader) {
         .body       = "color = vec4(col, 1.0);",
         .input      = PL_SHADER_SIG_NONE,
@@ -411,33 +416,58 @@ static void pl_shader_tests(const struct pl_gpu *gpu)
 
     TEST_FBO_PATTERN(1e-6, "%s", "using custom vertices");
 
-    const struct pl_tex *src;
+    pl_tex src;
     src = pl_tex_create(gpu, &(struct pl_tex_params) {
         .format         = fbo_fmt,
         .w              = FBO_W,
         .h              = FBO_H,
+        .storable       = fbo->params.storable,
         .sampleable     = true,
         .initial_data   = data,
     });
 
+    if (fbo->params.storable) {
+        // Test 1x1 blit, to make sure the scaling code runs
+        REQUIRE(pl_tex_blit_compute(gpu, dp, &(struct pl_tex_blit_params) {
+            .src = src,
+            .dst = fbo,
+            .src_rc = {0, 0, 0, 1, 1, 1},
+            .dst_rc = {0, 0, 0, FBO_W, FBO_H, 1},
+            .sample_mode = PL_TEX_SAMPLE_NEAREST,
+        }));
+
+        // Test non-resizing blit, which uses the efficient imageLoad path
+        REQUIRE(pl_tex_blit_compute(gpu, dp, &(struct pl_tex_blit_params) {
+            .src = src,
+            .dst = fbo,
+            .src_rc = {0, 0, 0, FBO_W, FBO_H, 1},
+            .dst_rc = {0, 0, 0, FBO_W, FBO_H, 1},
+            .sample_mode = PL_TEX_SAMPLE_NEAREST,
+        }));
+
+        TEST_FBO_PATTERN(1e-6, "%s", "pl_tex_blit_compute");
+    }
+
     // Test encoding/decoding of all gamma functions, color spaces, etc.
     for (enum pl_color_transfer trc = 0; trc < PL_COLOR_TRC_COUNT; trc++) {
         sh = pl_dispatch_begin(dp);
-        pl_shader_sample_nearest(sh, &(struct pl_sample_src) { .tex = src });
-        pl_shader_delinearize(sh, trc);
-        pl_shader_linearize(sh, trc);
-        REQUIRE(pl_dispatch_finish(dp, &(struct pl_dispatch_params) {
+        pl_shader_sample_nearest(sh, pl_sample_src( .tex = src ));
+        pl_shader_delinearize(sh, pl_color_space( .transfer = trc ));
+        pl_shader_linearize(sh, pl_color_space( .transfer = trc ));
+        REQUIRE(pl_dispatch_finish(dp, pl_dispatch_params(
             .shader = &sh,
             .target = fbo,
-        }));
+        )));
 
         float epsilon = pl_color_transfer_is_hdr(trc) ? 1e-4 : 1e-6;
         TEST_FBO_PATTERN(epsilon, "transfer function %d", (int) trc);
     }
 
     for (enum pl_color_system sys = 0; sys < PL_COLOR_SYSTEM_COUNT; sys++) {
+        if (sys == PL_COLOR_SYSTEM_DOLBYVISION)
+            continue; // requires metadata
         sh = pl_dispatch_begin(dp);
-        pl_shader_sample_nearest(sh, &(struct pl_sample_src) { .tex = src });
+        pl_shader_sample_nearest(sh, pl_sample_src( .tex = src ));
         pl_shader_encode_color(sh, &(struct pl_color_repr) { .sys = sys });
         pl_shader_decode_color(sh, &(struct pl_color_repr) { .sys = sys }, NULL);
         REQUIRE(pl_dispatch_finish(dp, &(struct pl_dispatch_params) {
@@ -457,7 +487,6 @@ static void pl_shader_tests(const struct pl_gpu *gpu)
             // edge cases for some reason
             // TODO: figure out why!
             continue;
-            break;
 
         default: epsilon = 1e-6; break;
         }
@@ -465,21 +494,6 @@ static void pl_shader_tests(const struct pl_gpu *gpu)
         TEST_FBO_PATTERN(epsilon, "color system %d", (int) sys);
     }
 
-    for (enum pl_color_light light = 0; light < PL_COLOR_LIGHT_COUNT; light++) {
-        sh = pl_dispatch_begin(dp);
-        struct pl_color_space src_space = { .light = light };
-        struct pl_color_space dst_space = { 0 };
-        pl_shader_sample_nearest(sh, &(struct pl_sample_src) { .tex = src });
-        pl_shader_color_map(sh, NULL, src_space, dst_space, NULL, false);
-        pl_shader_color_map(sh, NULL, dst_space, src_space, NULL, false);
-        REQUIRE(pl_dispatch_finish(dp, &(struct pl_dispatch_params) {
-            .shader = &sh,
-            .target = fbo,
-        }));
-
-        TEST_FBO_PATTERN(1e-6, "light %d", (int) light);
-    }
-
     // Repeat this a few times to test the caching
     for (int i = 0; i < 10; i++) {
         if (i == 5) {
@@ -491,7 +505,7 @@ static void pl_shader_tests(const struct pl_gpu *gpu)
             REQUIRE(pl_dispatch_save(dp, cache) == size);
 
             pl_dispatch_destroy(&dp);
-            dp = pl_dispatch_create(gpu->ctx, gpu);
+            dp = pl_dispatch_create(gpu->log, gpu);
             pl_dispatch_load(dp, cache);
 
 #ifndef MSAN
@@ -509,20 +523,16 @@ static void pl_shader_tests(const struct pl_gpu *gpu)
         sh = pl_dispatch_begin(dp);
 
         // For testing, force the use of CS if possible
-        if (gpu->caps & PL_GPU_CAP_COMPUTE) {
-            sh->is_compute = true;
+        if (gpu->glsl.compute) {
+            sh->type = SH_COMPUTE;
             sh->res.compute_group_size[0] = 8;
             sh->res.compute_group_size[1] = 8;
         }
 
-        pl_shader_deband(sh,
-            &(struct pl_sample_src) {
-                .tex            = src,
-            },
-            &(struct pl_deband_params) {
-                .iterations     = 0,
-                .grain          = 0.0,
-        });
+        pl_shader_deband(sh, pl_sample_src( .tex = src ), pl_deband_params(
+            .iterations     = 0,
+            .grain          = 0.0,
+        ));
 
         REQUIRE(pl_dispatch_finish(dp, &(struct pl_dispatch_params) {
             .shader = &sh,
@@ -533,10 +543,12 @@ static void pl_shader_tests(const struct pl_gpu *gpu)
 
     // Test peak detection and readback if possible
     sh = pl_dispatch_begin(dp);
-    pl_shader_sample_nearest(sh, &(struct pl_sample_src) { .tex = src });
+    pl_shader_sample_nearest(sh, pl_sample_src( .tex = src ));
 
-    struct pl_shader_obj *peak_state = NULL;
-    if (pl_shader_detect_peak(sh, pl_color_space_monitor, &peak_state, NULL)) {
+    pl_shader_obj peak_state = NULL;
+    struct pl_color_space csp_gamma22 = { .transfer = PL_COLOR_TRC_GAMMA22 };
+    struct pl_peak_detect_params peak_params = { .minimum_peak = 0.01 };
+    if (pl_shader_detect_peak(sh, csp_gamma22, &peak_state, &peak_params)) {
         REQUIRE(pl_dispatch_compute(dp, &(struct pl_dispatch_compute_params) {
             .shader = &sh,
             .width = fbo->params.w,
@@ -552,13 +564,13 @@ static void pl_shader_tests(const struct pl_gpu *gpu)
             for (int x = 0; x < FBO_W; x++) {
                 float *color = &data[(y * FBO_W + x) * 4];
                 float smax = powf(PL_MAX(color[0], color[1]), 2.2);
+                smax = (1 - 1e-3f) * smax + 1e-3f;
                 float slog = logf(PL_MAX(smax, 0.001));
                 real_peak = PL_MAX(smax, real_peak);
                 real_avg += slog;
             }
         }
 
-        real_peak *= 1.0 + pl_peak_detect_default_params.overshoot_margin;
         real_avg = expf(real_avg / (FBO_W * FBO_H));
         printf("real peak: %f, real average: %f\n", real_peak, real_avg);
         REQUIRE(feq(peak, real_peak, 1e-4));
@@ -571,9 +583,9 @@ static void pl_shader_tests(const struct pl_gpu *gpu)
 #ifdef PL_HAVE_LCMS
     // Test the use of ICC profiles if available
     sh = pl_dispatch_begin(dp);
-    pl_shader_sample_nearest(sh, &(struct pl_sample_src) { .tex = src });
+    pl_shader_sample_nearest(sh, pl_sample_src( .tex = src ));
 
-    struct pl_shader_obj *icc = NULL;
+    pl_shader_obj icc = NULL;
     struct pl_icc_color_space src_color = { .color = pl_color_space_bt709 };
     struct pl_icc_color_space dst_color = { .color = pl_color_space_srgb };
     struct pl_icc_result out;
@@ -590,30 +602,43 @@ static void pl_shader_tests(const struct pl_gpu *gpu)
     pl_shader_obj_destroy(&icc);
 #endif
 
-    // Test AV1 grain synthesis
-    struct pl_shader_obj *grain = NULL;
+    // Test film grain synthesis
+    pl_shader_obj grain = NULL;
+    struct pl_film_grain_params grain_params = {
+        .tex = src,
+        .components = 3,
+        .component_mapping = { 0, 1, 2},
+        .repr = &(struct pl_color_repr) {
+            .sys = PL_COLOR_SYSTEM_BT_709,
+            .levels = PL_COLOR_LEVELS_LIMITED,
+            .bits = { .color_depth = 10, .sample_depth = 10 },
+        },
+    };
+
     for (int i = 0; i < 2; i++) {
-        struct pl_av1_grain_params grain_params = {
-            .data = av1_grain_data,
-            .tex = src,
-            .components = 3,
-            .component_mapping = { 0, 1, 2 },
-            .repr = &(struct pl_color_repr) {
-                .sys = PL_COLOR_SYSTEM_BT_709,
-                .levels = PL_COLOR_LEVELS_LIMITED,
-                .bits = { .color_depth = 10, .sample_depth = 10 },
-            },
-        };
-        grain_params.data.grain_seed = rand();
-        grain_params.data.overlap = !!i;
+        grain_params.data.type = PL_FILM_GRAIN_AV1;
+        grain_params.data.params.av1 = av1_grain_data;
+        grain_params.data.params.av1.overlap = !!i;
+        grain_params.data.seed = rand();
 
         sh = pl_dispatch_begin(dp);
-        pl_shader_av1_grain(sh, &grain, &grain_params);
+        pl_shader_film_grain(sh, &grain, &grain_params);
         REQUIRE(pl_dispatch_finish(dp, &(struct pl_dispatch_params) {
             .shader = &sh,
             .target = fbo,
         }));
     }
+
+    grain_params.data.type = PL_FILM_GRAIN_H274;
+    grain_params.data.params.h274 = h274_grain_data;
+    grain_params.data.seed = rand();
+
+    sh = pl_dispatch_begin(dp);
+    pl_shader_film_grain(sh, &grain, &grain_params);
+    REQUIRE(pl_dispatch_finish(dp, &(struct pl_dispatch_params) {
+        .shader = &sh,
+        .target = fbo,
+    }));
     pl_shader_obj_destroy(&grain);
 
     // Test custom shaders
@@ -645,23 +670,36 @@ static void pl_shader_tests(const struct pl_gpu *gpu)
         .target = fbo,
     }));
 
+    // Test dolbyvision
+    if (gpu->glsl.version >= 130) {
+        struct pl_color_repr repr = {
+            .sys = PL_COLOR_SYSTEM_DOLBYVISION,
+            .dovi = &dovi_meta,
+        };
+
+        sh = pl_dispatch_begin(dp);
+        pl_shader_sample_direct(sh, pl_sample_src( .tex = src ));
+        pl_shader_decode_color(sh, &repr, NULL);
+        REQUIRE(pl_dispatch_finish(dp, &(struct pl_dispatch_params) {
+            .shader = &sh,
+            .target = fbo,
+        }));
+    }
+
     pl_dispatch_destroy(&dp);
     pl_tex_destroy(gpu, &src);
     pl_tex_destroy(gpu, &fbo);
 }
 
-static void pl_scaler_tests(const struct pl_gpu *gpu)
+static void pl_scaler_tests(pl_gpu gpu)
 {
-    const struct pl_fmt *src_fmt = pl_find_fmt(gpu, PL_FMT_FLOAT, 1, 16, 32,
-                                               PL_FMT_CAP_LINEAR);
-
-    const struct pl_fmt *fbo_fmt = pl_find_fmt(gpu, PL_FMT_FLOAT, 1, 16, 32,
-                                               PL_FMT_CAP_RENDERABLE);
+    pl_fmt src_fmt = pl_find_fmt(gpu, PL_FMT_FLOAT, 1, 16, 32, PL_FMT_CAP_LINEAR);
+    pl_fmt fbo_fmt = pl_find_fmt(gpu, PL_FMT_FLOAT, 1, 16, 32, PL_FMT_CAP_RENDERABLE);
     if (!src_fmt || !fbo_fmt)
         return;
 
     float *fbo_data = NULL;
-    struct pl_shader_obj *lut = NULL;
+    pl_shader_obj lut = NULL;
 
     static float data_5x5[5][5] = {
         { 0, 0, 0, 0, 0 },
@@ -671,7 +709,7 @@ static void pl_scaler_tests(const struct pl_gpu *gpu)
         { 0, 0, 0, 0, 0 },
     };
 
-    const struct pl_tex *dot5x5 = pl_tex_create(gpu, &(struct pl_tex_params) {
+    pl_tex dot5x5 = pl_tex_create(gpu, &(struct pl_tex_params) {
         .w              = 5,
         .h              = 5,
         .format         = src_fmt,
@@ -684,33 +722,27 @@ static void pl_scaler_tests(const struct pl_gpu *gpu)
         .h              = 100,
         .format         = fbo_fmt,
         .renderable     = true,
-        .storable       = !!(fbo_fmt->caps & PL_FMT_CAP_STORABLE),
-        .host_readable  = true,
+        .storable       = fbo_fmt->caps & PL_FMT_CAP_STORABLE,
+        .host_readable  = fbo_fmt->caps & PL_FMT_CAP_HOST_READABLE,
     };
 
-    const struct pl_tex *fbo = pl_tex_create(gpu, &fbo_params);
-    if (!fbo) {
-        printf("Failed creating readable FBO... falling back to non-readable\n");
-        fbo_params.host_readable = false;
-        fbo = pl_tex_create(gpu, &fbo_params);
-    }
-
-    struct pl_dispatch *dp = pl_dispatch_create(gpu->ctx, gpu);
+    pl_tex fbo = pl_tex_create(gpu, &fbo_params);
+    pl_dispatch dp = pl_dispatch_create(gpu->log, gpu);
     if (!dot5x5 || !fbo || !dp)
         goto error;
 
-    struct pl_shader *sh = pl_dispatch_begin(dp);
+    pl_shader sh = pl_dispatch_begin(dp);
     REQUIRE(pl_shader_sample_polar(sh,
-        &(struct pl_sample_src) {
+        pl_sample_src(
             .tex        = dot5x5,
             .new_w      = fbo->params.w,
             .new_h      = fbo->params.h,
-        },
-        &(struct pl_sample_filter_params) {
+        ),
+        pl_sample_filter_params(
             .filter     = pl_filter_ewa_lanczos,
             .lut        = &lut,
             .no_compute = !fbo->params.storable,
-        }
+        )
     ));
     REQUIRE(pl_dispatch_finish(dp, &(struct pl_dispatch_params) {
         .shader = &sh,
@@ -843,7 +875,7 @@ static const char *test_luts[] = {
 
 };
 
-static bool frame_passthrough(const struct pl_gpu *gpu, const struct pl_tex **tex,
+static bool frame_passthrough(pl_gpu gpu, pl_tex *tex,
                               const struct pl_source_frame *src, struct pl_frame *out_frame)
 {
     const struct pl_frame *frame = src->frame_data;
@@ -856,18 +888,23 @@ static enum pl_queue_status get_frame_ptr(struct pl_source_frame *out_frame,
 {
     const struct pl_source_frame **pframe = qparams->priv;
     if (!(*pframe)->frame_data)
-        return QUEUE_EOF;
+        return PL_QUEUE_EOF;
 
     *out_frame = *(*pframe)++;
-    return QUEUE_OK;
+    return PL_QUEUE_OK;
 }
 
-static void pl_render_tests(const struct pl_gpu *gpu)
+static void render_info_cb(void *priv, const struct pl_render_info *info)
 {
-    const struct pl_tex *img5x5_tex = NULL, *fbo = NULL;
-    struct pl_renderer *rr = NULL;
+    printf("{%d} Executed shader: %s\n", info->index,
+           info->pass->shader->description);
+}
+
+static void pl_render_tests(pl_gpu gpu)
+{
+    pl_tex img5x5_tex = NULL, fbo = NULL;
+    pl_renderer rr = NULL;
 
-    float *fbo_data = NULL;
     static float data_5x5[5][5] = {
         { 0.0, 0.0, 0.0, 0.0, 0.0 },
         { 0.0, 0.0, 0.0, 0.0, 0.0 },
@@ -894,8 +931,8 @@ static void pl_render_tests(const struct pl_gpu *gpu)
     if (!pl_upload_plane(gpu, &img5x5, &img5x5_tex, &img5x5_data))
         goto error;
 
-    rr = pl_renderer_create(gpu->ctx, gpu);
-    pl_tex_clear(gpu, fbo, (float[4]){0});
+    rr = pl_renderer_create(gpu->log, gpu);
+    pl_tex_clear_ex(gpu, fbo, (union pl_clear_color){0});
 
     struct pl_frame image = {
         .num_planes     = 1,
@@ -904,7 +941,7 @@ static void pl_render_tests(const struct pl_gpu *gpu)
             .sys        = PL_COLOR_SYSTEM_BT_709,
             .levels     = PL_COLOR_LEVELS_FULL,
         },
-        .color          = pl_color_space_bt709,
+        .color          = pl_color_space_srgb,
         .crop           = {-1.0, 0.0, width - 1.0, height},
     };
 
@@ -919,42 +956,37 @@ static void pl_render_tests(const struct pl_gpu *gpu)
         .repr = {
             .sys        = PL_COLOR_SYSTEM_RGB,
             .levels     = PL_COLOR_LEVELS_FULL,
+            .bits.color_depth = 32,
         },
         .color          = pl_color_space_srgb,
     };
 
     REQUIRE(pl_render_image(rr, &image, &target, NULL));
 
-    fbo_data = malloc(fbo->params.w * fbo->params.h * sizeof(float[4]));
-    REQUIRE(pl_tex_download(gpu, &(struct pl_tex_transfer_params) {
-        .tex            = fbo,
-        .ptr            = fbo_data,
-    }));
-
     // TODO: embed a reference texture and ensure it matches
 
     // Test a bunch of different params
 #define TEST(SNAME, STYPE, DEFAULT, FIELD, LIMIT)                       \
     do {                                                                \
         for (int i = 0; i <= LIMIT; i++) {                              \
+            printf("testing `" #STYPE "." #FIELD " = %d`\n", i);        \
             struct pl_render_params params = pl_render_default_params;  \
             params.force_dither = true;                                 \
             struct STYPE tmp = DEFAULT;                                 \
             tmp.FIELD = i;                                              \
             params.SNAME = &tmp;                                        \
-            for (int p = 0; p < 5; p++) {                               \
-                REQUIRE(pl_render_image(rr, &image, &target, &params)); \
-                pl_gpu_flush(gpu);                                      \
-            }                                                           \
+            REQUIRE(pl_render_image(rr, &image, &target, &params));     \
+            pl_gpu_flush(gpu);                                          \
         }                                                               \
     } while (0)
 
 #define TEST_PARAMS(NAME, FIELD, LIMIT) \
     TEST(NAME##_params, pl_##NAME##_params, pl_##NAME##_default_params, FIELD, LIMIT)
 
-    for (int i = 0; i < pl_num_filter_presets; i++) {
+    for (int i = 0; i < pl_num_scale_filters; i++) {
         struct pl_render_params params = pl_render_default_params;
-        params.upscaler = pl_filter_presets[i].filter;
+        params.upscaler = pl_scale_filters[i].filter;
+        printf("testing `params.upscaler = /* %s */`\n", pl_scale_filters[i].name);
         REQUIRE(pl_render_image(rr, &image, &target, &params));
         pl_gpu_flush(gpu);
     }
@@ -962,17 +994,15 @@ static void pl_render_tests(const struct pl_gpu *gpu)
     TEST_PARAMS(deband, iterations, 3);
     TEST_PARAMS(sigmoid, center, 1);
     TEST_PARAMS(color_map, intent, PL_INTENT_ABSOLUTE_COLORIMETRIC);
-    TEST_PARAMS(color_map, gamut_warning, 1);
     TEST_PARAMS(dither, method, PL_DITHER_WHITE_NOISE);
     TEST_PARAMS(dither, temporal, true);
     TEST(cone_params, pl_cone_params, pl_vision_deuteranomaly, strength, 0);
 
-    // Test HDR stuff
-    image.color.sig_scale = 10.0;
-    target.color.sig_scale = 2.0;
-    TEST_PARAMS(color_map, tone_mapping_algo, PL_TONE_MAPPING_BT_2390);
-    TEST_PARAMS(color_map, desaturation_strength, 1);
-    image.color.sig_scale = target.color.sig_scale = 0.0;
+    // Test HDR tone mapping
+    image.color = pl_color_space_hdr10;
+    TEST_PARAMS(color_map, tone_mapping_mode, PL_TONE_MAP_MODE_COUNT - 1);
+    TEST_PARAMS(color_map, gamut_mode, PL_GAMUT_MODE_COUNT - 1);
+    image.color = pl_color_space_srgb;
 
     // Test some misc stuff
     struct pl_render_params params = pl_render_default_params;
@@ -990,9 +1020,14 @@ static void pl_render_tests(const struct pl_gpu *gpu)
     REQUIRE(pl_render_image(rr, &image, &target, &params));
     params = pl_render_default_params;
 
-    image.av1_grain = av1_grain_data;
+    // Test film grain synthesis
+    image.film_grain.type = PL_FILM_GRAIN_AV1;
+    image.film_grain.params.av1 = av1_grain_data,
     REQUIRE(pl_render_image(rr, &image, &target, &params));
-    image.av1_grain = (struct pl_av1_grain_data) {0};
+    image.film_grain.type = PL_FILM_GRAIN_H274;
+    image.film_grain.params.h274 = h274_grain_data,
+    REQUIRE(pl_render_image(rr, &image, &target, &params));
+    image.film_grain = (struct pl_film_grain_data) {0};
 
     // Test mpv-style custom shaders
     for (int i = 0; i < PL_ARRAY_SIZE(user_shader_tests); i++) {
@@ -1001,7 +1036,7 @@ static void pl_render_tests(const struct pl_gpu *gpu)
         hook = pl_mpv_user_shader_parse(gpu, user_shader_tests[i],
                                         strlen(user_shader_tests[i]));
 
-        if (gpu->caps & PL_GPU_CAP_COMPUTE) {
+        if (gpu->glsl.compute) {
             REQUIRE(hook);
         } else {
             // Not all shaders compile without compute shader support
@@ -1021,9 +1056,15 @@ static void pl_render_tests(const struct pl_gpu *gpu)
     for (int i = 0; i < PL_ARRAY_SIZE(test_luts); i++) {
         printf("testing custom lut %d\n", i);
         struct pl_custom_lut *lut;
-        lut = pl_lut_parse_cube(gpu->ctx, test_luts[i], strlen(test_luts[i]));
+        lut = pl_lut_parse_cube(gpu->log, test_luts[i], strlen(test_luts[i]));
         REQUIRE(lut);
 
+        bool has_3dlut = gpu->limits.max_tex_3d_dim && gpu->glsl.version > 100;
+        if (lut->size[2] && !has_3dlut) {
+            pl_lut_free(&lut);
+            continue;
+        }
+
         // Test all three at the same time to reduce the number of tests
         image.lut = target.lut = params.lut = lut;
 
@@ -1040,9 +1081,16 @@ static void pl_render_tests(const struct pl_gpu *gpu)
     // Test overlays
     image.num_overlays = 1;
     image.overlays = &(struct pl_overlay) {
-        .plane = img5x5,
-        .rect = {0, 0, 2, 2},
+        .tex = img5x5.texture,
         .mode = PL_OVERLAY_NORMAL,
+        .num_parts = 2,
+        .parts = (struct pl_overlay_part[]) {{
+            .src = {0, 0, 2, 2},
+            .dst = {30, 100, 40, 200},
+        }, {
+            .src = {2, 2, 5, 5},
+            .dst = {1000, -1, 3, 5},
+        }},
     };
     REQUIRE(pl_render_image(rr, &image, &target, &params));
     params.disable_fbos = true;
@@ -1052,18 +1100,30 @@ static void pl_render_tests(const struct pl_gpu *gpu)
 
     target.num_overlays = 1;
     target.overlays = &(struct pl_overlay) {
-        .plane = img5x5,
-        .rect = {5, 5, 15, 15},
+        .tex = img5x5.texture,
         .mode = PL_OVERLAY_MONOCHROME,
-        .base_color = {1.0, 0.5, 0.0},
+        .num_parts = 1,
+        .parts = &(struct pl_overlay_part) {
+            .src = {5, 5, 15, 15},
+            .dst = {5, 5, 15, 15},
+            .color = {1.0, 0.5, 0.0},
+        },
     };
     REQUIRE(pl_render_image(rr, &image, &target, &params));
+    REQUIRE(pl_render_image(rr, NULL, &target, &params));
     target.num_overlays = 0;
 
+    // Test rotation
+    for (pl_rotation rot = 0; rot < PL_ROTATION_360; rot += PL_ROTATION_90) {
+        image.rotation = rot;
+        REQUIRE(pl_render_image(rr, &image, &target, &params));
+    }
+
     // Attempt frame mixing, using the mixer queue helper
     printf("testing frame mixing \n");
     struct pl_render_params mix_params = {
         .frame_mixer = &pl_filter_mitchell_clamp,
+        .info_callback = render_info_cb,
     };
 
     struct pl_queue_params qparams = {
@@ -1083,22 +1143,27 @@ static void pl_render_tests(const struct pl_gpu *gpu)
         };
     }
 
-    struct pl_queue *queue = pl_queue_create(gpu);
+    pl_queue queue = pl_queue_create(gpu);
     enum pl_queue_status ret;
 
-    // Test pre-pushing all frames, with delayed EOF
-    for (int i = 0; i < NUM_MIX_FRAMES; i++)
-        pl_queue_push(queue, &srcframes[i]);
+    // Test pre-pushing all frames, with delayed EOF.
+    for (int i = 0; i < NUM_MIX_FRAMES; i++) {
+        const struct pl_source_frame *src = &srcframes[i];
+        if (i > 10) // test pushing in reverse order
+            src = &srcframes[NUM_MIX_FRAMES - 1 - i];
+        if (!pl_queue_push_block(queue, 1, src)) // mini-sleep
+            pl_queue_push(queue, src); // push it anyway, for testing
+    }
 
     struct pl_frame_mix mix;
-    while ((ret = pl_queue_update(queue, &mix, &qparams)) != QUEUE_EOF) {
-        if (ret == QUEUE_MORE) {
+    while ((ret = pl_queue_update(queue, &mix, &qparams)) != PL_QUEUE_EOF) {
+        if (ret == PL_QUEUE_MORE) {
             REQUIRE(qparams.pts > 0.0);
             pl_queue_push(queue, NULL); // push delayed EOF
             continue;
         }
 
-        REQUIRE(ret == QUEUE_OK);
+        REQUIRE(ret == PL_QUEUE_OK);
         REQUIRE(pl_render_image_mix(rr, &mix, &target, &mix_params));
 
         // Simulate advancing vsync
@@ -1118,8 +1183,8 @@ static void pl_render_tests(const struct pl_gpu *gpu)
     };
 
     pl_queue_reset(queue);
-    while ((ret = pl_queue_update(queue, &mix, &qparams)) != QUEUE_EOF) {
-        REQUIRE(ret == QUEUE_OK);
+    while ((ret = pl_queue_update(queue, &mix, &qparams)) != PL_QUEUE_EOF) {
+        REQUIRE(ret == PL_QUEUE_OK);
         REQUIRE(mix.num_frames <= 2);
         REQUIRE(pl_render_image_mix(rr, &mix, &target, &mix_params));
         qparams.pts += qparams.vsync_duration;
@@ -1127,12 +1192,11 @@ static void pl_render_tests(const struct pl_gpu *gpu)
 
     // Test large PTS jump
     pl_queue_reset(queue);
-    REQUIRE(pl_queue_update(queue, &mix, &qparams) == QUEUE_EOF);
+    REQUIRE(pl_queue_update(queue, &mix, &qparams) == PL_QUEUE_EOF);
 
     pl_queue_destroy(&queue);
 
 error:
-    free(fbo_data);
     pl_renderer_destroy(&rr);
     pl_tex_destroy(gpu, &img5x5_tex);
     pl_tex_destroy(gpu, &fbo);
@@ -1143,9 +1207,9 @@ static struct pl_hook_res noop_hook(void *priv, const struct pl_hook_params *par
     return (struct pl_hook_res) {0};
 }
 
-static void pl_ycbcr_tests(const struct pl_gpu *gpu)
+static void pl_ycbcr_tests(pl_gpu gpu)
 {
-    struct pl_renderer *rr = pl_renderer_create(gpu->ctx, gpu);
+    pl_renderer rr = pl_renderer_create(gpu->log, gpu);
     if (!rr)
         return;
 
@@ -1163,16 +1227,17 @@ static void pl_ycbcr_tests(const struct pl_gpu *gpu)
             .component_map = {i},
             .pixel_stride = sizeof(uint16_t),
             .row_stride = PL_ALIGN2(width * sizeof(uint16_t),
-                                    gpu->limits.align_tex_xfer_stride),
+                                    gpu->limits.align_tex_xfer_pitch),
         };
     }
 
-    const struct pl_fmt *fmt = pl_plane_find_fmt(gpu, NULL, &data[0]);
-    if (!fmt || !(fmt->caps & (PL_FMT_CAP_RENDERABLE | PL_FMT_CAP_HOST_READABLE)))
+    pl_fmt fmt = pl_plane_find_fmt(gpu, NULL, &data[0]);
+    enum pl_fmt_caps caps = PL_FMT_CAP_RENDERABLE | PL_FMT_CAP_HOST_READABLE;
+    if (!fmt || (fmt->caps & caps) != caps)
         return;
 
-    const struct pl_tex *src_tex[3] = {0};
-    const struct pl_tex *dst_tex[3] = {0};
+    pl_tex src_tex[3] = {0};
+    pl_tex dst_tex[3] = {0};
     struct pl_frame img = {
         .num_planes = 3,
         .repr = pl_color_repr_hdtv,
@@ -1248,7 +1313,7 @@ static void pl_ycbcr_tests(const struct pl_gpu *gpu)
         REQUIRE(pl_tex_download(gpu, &(struct pl_tex_transfer_params) {
             .tex = dst_tex[i],
             .ptr = dst_buffer,
-            .stride_w = data[i].row_stride / data[i].pixel_stride,
+            .row_pitch = data[i].row_stride,
         }));
 
         for (int y = 0; y < data[i].height; y++) {
@@ -1272,7 +1337,7 @@ error:
     }
 }
 
-static void pl_test_export_import(const struct pl_gpu *gpu,
+static void pl_test_export_import(pl_gpu gpu,
                                   enum pl_handle_type handle_type)
 {
     // Test texture roundtrip
@@ -1281,23 +1346,22 @@ static void pl_test_export_import(const struct pl_gpu *gpu,
         !(gpu->import_caps.tex & handle_type))
         goto skip_tex;
 
-    const struct pl_fmt *fmt = pl_find_fmt(gpu, PL_FMT_UNORM, 1, 0, 0,
-                                           PL_FMT_CAP_BLITTABLE);
+    pl_fmt fmt = pl_find_fmt(gpu, PL_FMT_UNORM, 1, 0, 0, PL_FMT_CAP_BLITTABLE);
     if (!fmt)
         goto skip_tex;
 
     printf("testing texture import/export\n");
 
-    const struct pl_tex *export = pl_tex_create(gpu, &(struct pl_tex_params) {
+    pl_tex export = pl_tex_create(gpu, &(struct pl_tex_params) {
         .w = 32,
         .h = 32,
         .format = fmt,
         .export_handle = handle_type,
     });
     REQUIRE(export);
-    REQUIRE(export->shared_mem.handle.fd > -1);
+    REQUIRE_HANDLE(export->shared_mem, handle_type);
 
-    const struct pl_tex *import = pl_tex_create(gpu, &(struct pl_tex_params) {
+    pl_tex import = pl_tex_create(gpu, &(struct pl_tex_params) {
         .w = 32,
         .h = 32,
         .format = fmt,
@@ -1319,14 +1383,14 @@ skip_tex: ;
 
     printf("testing buffer import/export\n");
 
-    const struct pl_buf *exp_buf = pl_buf_create(gpu, &(struct pl_buf_params) {
+    pl_buf exp_buf = pl_buf_create(gpu, &(struct pl_buf_params) {
         .size = 32,
         .export_handle = handle_type,
     });
     REQUIRE(exp_buf);
-    REQUIRE(exp_buf->shared_mem.handle.fd > -1);
+    REQUIRE_HANDLE(exp_buf->shared_mem, handle_type);
 
-    const struct pl_buf *imp_buf = pl_buf_create(gpu, &(struct pl_buf_params) {
+    pl_buf imp_buf = pl_buf_create(gpu, &(struct pl_buf_params) {
         .size = 32,
         .import_handle = handle_type,
         .shared_mem = exp_buf->shared_mem,
@@ -1337,7 +1401,7 @@ skip_tex: ;
     pl_buf_destroy(gpu, &exp_buf);
 }
 
-static void pl_test_host_ptr(const struct pl_gpu *gpu)
+static void pl_test_host_ptr(pl_gpu gpu)
 {
     if (!(gpu->import_caps.buf & PL_HANDLE_HOST_PTR))
         return;
@@ -1345,7 +1409,7 @@ static void pl_test_host_ptr(const struct pl_gpu *gpu)
 #ifdef __unix__
 
     printf("testing host ptr\n");
-    REQUIRE(gpu->caps & PL_GPU_CAP_MAPPED_BUFFERS);
+    REQUIRE(gpu->limits.max_mapped_size);
 
     const size_t size = 2 << 20;
     const size_t offset = 2 << 10;
@@ -1355,8 +1419,7 @@ static void pl_test_host_ptr(const struct pl_gpu *gpu)
     for (int i = 0; i < size; i++)
         data[i] = (uint8_t) i;
 
-    const struct pl_buf *buf = pl_buf_create(gpu, &(struct pl_buf_params) {
-        .type = PL_BUF_TEX_TRANSFER,
+    pl_buf buf = pl_buf_create(gpu, &(struct pl_buf_params) {
         .size = slice,
         .import_handle = PL_HANDLE_HOST_PTR,
         .shared_mem = {
@@ -1376,7 +1439,7 @@ static void pl_test_host_ptr(const struct pl_gpu *gpu)
 #endif // unix
 }
 
-static void gpu_shader_tests(const struct pl_gpu *gpu)
+static void gpu_shader_tests(pl_gpu gpu)
 {
     pl_buffer_tests(gpu);
     pl_texture_tests(gpu);
@@ -1388,7 +1451,7 @@ static void gpu_shader_tests(const struct pl_gpu *gpu)
     REQUIRE(!pl_gpu_is_failed(gpu));
 }
 
-static void gpu_interop_tests(const struct pl_gpu *gpu)
+static void gpu_interop_tests(pl_gpu gpu)
 {
     pl_test_export_import(gpu, PL_HANDLE_DMA_BUF);
     pl_test_host_ptr(gpu);
