@@ -3,38 +3,37 @@
 #include "vulkan/gpu.h"
 #include <vulkan/vulkan.h>
 
-static void vulkan_interop_tests(const struct pl_vulkan *pl_vk,
+static void vulkan_interop_tests(pl_vulkan pl_vk,
                                  enum pl_handle_type handle_type)
 {
-    const struct pl_gpu *gpu = pl_vk->gpu;
+    pl_gpu gpu = pl_vk->gpu;
     printf("testing vulkan interop for handle type 0x%x\n", handle_type);
 
     if (gpu->export_caps.buf & handle_type) {
-        const struct pl_buf *buf = pl_buf_create(gpu, &(struct pl_buf_params) {
+        pl_buf buf = pl_buf_create(gpu, pl_buf_params(
             .size = 1024,
             .export_handle = handle_type,
-        });
+        ));
 
         REQUIRE(buf);
-        REQUIRE(buf->shared_mem.handle.fd > -1);
+        REQUIRE_HANDLE(buf->shared_mem, handle_type);
         REQUIRE(buf->shared_mem.size >= buf->params.size);
         REQUIRE(pl_buf_export(gpu, buf));
         pl_buf_destroy(gpu, &buf);
     }
 
-    const struct pl_fmt *fmt = pl_find_fmt(gpu, PL_FMT_UNORM, 1, 0, 0,
-                                           PL_FMT_CAP_BLITTABLE);
+    pl_fmt fmt = pl_find_fmt(gpu, PL_FMT_UNORM, 1, 0, 0, PL_FMT_CAP_BLITTABLE);
     if (!fmt)
         return;
 
     if (gpu->export_caps.sync & handle_type) {
-        const struct pl_sync *sync = pl_sync_create(gpu, handle_type);
-        const struct pl_tex *tex = pl_tex_create(gpu, &(struct pl_tex_params) {
+        pl_sync sync = pl_sync_create(gpu, handle_type);
+        pl_tex tex = pl_tex_create(gpu, pl_tex_params(
             .w = 32,
             .h = 32,
             .format = fmt,
             .blit_dst = true,
-        });
+        ));
 
         REQUIRE(sync);
         REQUIRE(tex);
@@ -49,12 +48,10 @@ static void vulkan_interop_tests(const struct pl_vulkan *pl_vk,
         // Re-use our internal helpers to signal this VkSemaphore
         struct vk_ctx *vk = PL_PRIV(pl_vk);
         struct vk_cmd *cmd = vk_cmd_begin(vk, vk->pool_graphics);
-        VkSemaphore signal;
         REQUIRE(cmd);
-        pl_vk_sync_unwrap(sync, NULL, &signal);
-        vk_cmd_sig(cmd, signal);
-        vk_cmd_queue(vk, &cmd);
-        REQUIRE(vk_flush_commands(vk));
+        struct pl_sync_vk *sync_vk = PL_PRIV(sync);
+        vk_cmd_sig(cmd, (pl_vulkan_sem){ sync_vk->signal });
+        REQUIRE(vk_cmd_submit(vk, &cmd));
 
         // Do something with the image again to "import" it
         pl_tex_clear(gpu, tex, (float[4]){0});
@@ -66,17 +63,17 @@ static void vulkan_interop_tests(const struct pl_vulkan *pl_vk,
     }
 }
 
-static void vulkan_swapchain_tests(const struct pl_vulkan *vk, VkSurfaceKHR surf)
+static void vulkan_swapchain_tests(pl_vulkan vk, VkSurfaceKHR surf)
 {
     if (!surf)
         return;
 
     printf("testing vulkan swapchain\n");
-    const struct pl_gpu *gpu = vk->gpu;
-    const struct pl_swapchain *sw;
-    sw = pl_vulkan_create_swapchain(vk, &(struct pl_vulkan_swapchain_params) {
+    pl_gpu gpu = vk->gpu;
+    pl_swapchain sw;
+    sw = pl_vulkan_create_swapchain(vk, pl_vulkan_swapchain_params(
         .surface = surf,
-    });
+    ));
     REQUIRE(sw);
 
     // Attempt actually initializing the swapchain
@@ -109,23 +106,23 @@ static void vulkan_swapchain_tests(const struct pl_vulkan *vk, VkSurfaceKHR surf
 
 int main()
 {
-    struct pl_context *ctx = pl_test_context();
-    const struct pl_vk_inst *inst;
-    inst = pl_vk_inst_create(ctx, &(struct pl_vk_inst_params) {
+    pl_log log = pl_test_logger();
+    pl_vk_inst inst = pl_vk_inst_create(log, pl_vk_inst_params(
         .debug = true,
         .debug_extra = true,
+        .get_proc_addr = vkGetInstanceProcAddr,
         .opt_extensions = (const char *[]){
             VK_KHR_SURFACE_EXTENSION_NAME,
             "VK_EXT_headless_surface", // in case it isn't defined
         },
         .num_opt_extensions = 2,
-    });
+    ));
 
     if (!inst)
         return SKIP;
 
-    PL_VK_LOAD_FUN(inst->instance, EnumeratePhysicalDevices, vkGetInstanceProcAddr);
-    PL_VK_LOAD_FUN(inst->instance, GetPhysicalDeviceProperties, vkGetInstanceProcAddr);
+    PL_VK_LOAD_FUN(inst->instance, EnumeratePhysicalDevices, inst->get_proc_addr);
+    PL_VK_LOAD_FUN(inst->instance, GetPhysicalDeviceProperties, inst->get_proc_addr);
 
     uint32_t num = 0;
     EnumeratePhysicalDevices(inst->instance, &num, NULL);
@@ -140,7 +137,7 @@ int main()
     VkSurfaceKHR surf = NULL;
 
 #ifdef VK_EXT_headless_surface
-    PL_VK_LOAD_FUN(inst->instance, CreateHeadlessSurfaceEXT, vkGetInstanceProcAddr);
+    PL_VK_LOAD_FUN(inst->instance, CreateHeadlessSurfaceEXT, inst->get_proc_addr);
     if (CreateHeadlessSurfaceEXT) {
         VkHeadlessSurfaceCreateInfoEXT info = {
             .sType = VK_STRUCTURE_TYPE_HEADLESS_SURFACE_CREATE_INFO_EXT,
@@ -153,42 +150,57 @@ int main()
 
     // Make sure choosing any device works
     VkPhysicalDevice dev;
-    dev = pl_vulkan_choose_device(ctx, &(struct pl_vulkan_device_params) {
+    dev = pl_vulkan_choose_device(log, pl_vulkan_device_params(
         .instance = inst->instance,
+        .get_proc_addr = inst->get_proc_addr,
         .allow_software = true,
         .surface = surf,
-    });
-    REQUIRE(dev);
+    ));
+    if (!dev)
+        return SKIP;
 
     // Test all attached devices
     for (int i = 0; i < num; i++) {
         VkPhysicalDeviceProperties props = {0};
         GetPhysicalDeviceProperties(devices[i], &props);
+#ifndef CI_ALLOW_SW
+        if (props.deviceType == VK_PHYSICAL_DEVICE_TYPE_CPU) {
+            printf("Skipping device %d: %s\n", i, props.deviceName);
+            continue;
+        }
+#endif
         printf("Testing device %d: %s\n", i, props.deviceName);
 
         // Make sure we can choose this device by name
-        dev = pl_vulkan_choose_device(ctx, &(struct pl_vulkan_device_params) {
+        dev = pl_vulkan_choose_device(log, pl_vulkan_device_params(
             .instance = inst->instance,
+            .get_proc_addr = inst->get_proc_addr,
             .device_name = props.deviceName,
-        });
+        ));
         REQUIRE(dev == devices[i]);
 
-        struct pl_vulkan_params params = pl_vulkan_default_params;
-        params.instance = inst->instance;
-        params.device = devices[i];
-        params.queue_count = 8; // test inter-queue stuff
-        params.surface = surf;
+        struct pl_vulkan_params params = *pl_vulkan_params(
+            .instance = inst->instance,
+            .get_proc_addr = inst->get_proc_addr,
+            .device = devices[i],
+            .queue_count = 8, // test inter-queue stuff
+            .surface = surf,
+        );
 
-        const struct pl_vulkan *vk = pl_vulkan_create(ctx, &params);
+        pl_vulkan vk = pl_vulkan_create(log, &params);
         if (!vk)
             continue;
 
         gpu_shader_tests(vk->gpu);
         vulkan_swapchain_tests(vk, surf);
 
+        // Print heap statistics
+        pl_vk_print_heap(vk->gpu, PL_LOG_DEBUG);
+
         // Test importing this context via the vulkan interop API
-        struct pl_vulkan_import_params iparams = {
+        pl_vulkan vk2 = pl_vulkan_import(log, pl_vulkan_import_params(
             .instance = vk->instance,
+            .get_proc_addr = inst->get_proc_addr,
             .phys_device = vk->phys_device,
             .device = vk->device,
 
@@ -198,8 +210,7 @@ int main()
             .queue_graphics = vk->queue_graphics,
             .queue_compute = vk->queue_compute,
             .queue_transfer = vk->queue_transfer,
-        };
-        const struct pl_vulkan *vk2 = pl_vulkan_import(ctx, &iparams);
+        ));
         REQUIRE(vk2);
         pl_vulkan_destroy(&vk2);
 
@@ -218,7 +229,7 @@ int main()
         // Re-run the same export/import tests with async queues disabled
         params.async_compute = false;
         params.async_transfer = false;
-        vk = pl_vulkan_create(ctx, &params);
+        vk = pl_vulkan_create(log, &params);
         REQUIRE(vk); // it succeeded the first time
 
 #ifdef PL_HAVE_UNIX
@@ -233,11 +244,11 @@ int main()
         pl_vulkan_destroy(&vk);
 
         // Reduce log spam after first tested device
-        pl_test_set_verbosity(ctx, PL_LOG_INFO);
+        pl_log_level_update(log, PL_LOG_INFO);
     }
 
     vkDestroySurfaceKHR(inst->instance, surf, NULL);
     pl_vk_inst_destroy(&inst);
-    pl_context_destroy(&ctx);
+    pl_log_destroy(&log);
     free(devices);
 }
