@@ -17,7 +17,7 @@
 
 #include <strings.h>
 
-#include "context.h"
+#include "log.h"
 #include "common.h"
 #include "gpu.h"
 
@@ -79,8 +79,7 @@ void pl_plane_data_from_mask(struct pl_plane_data *data, uint64_t mask[4])
     }
 }
 
-bool pl_plane_data_align(struct pl_plane_data *data,
-                         struct pl_bit_encoding *out_bits)
+bool pl_plane_data_align(struct pl_plane_data *data, struct pl_bit_encoding *out_bits)
 {
     struct pl_plane_data aligned = *data;
     struct pl_bit_encoding bits = {0};
@@ -153,8 +152,7 @@ misaligned:
     return false;
 }
 
-const struct pl_fmt *pl_plane_find_fmt(const struct pl_gpu *gpu, int out_map[4],
-                                       const struct pl_plane_data *data)
+pl_fmt pl_plane_find_fmt(pl_gpu gpu, int out_map[4], const struct pl_plane_data *data)
 {
     int dummy[4] = {0};
     out_map = PL_DEF(out_map, dummy);
@@ -168,13 +166,15 @@ const struct pl_fmt *pl_plane_find_fmt(const struct pl_gpu *gpu, int out_map[4],
     }
 
     for (int n = 0; n < gpu->num_formats; n++) {
-        const struct pl_fmt *fmt = gpu->formats[n];
+        pl_fmt fmt = gpu->formats[n];
         if (fmt->opaque || fmt->num_components < num)
             continue;
         if (fmt->type != data->type || fmt->texel_size != data->pixel_stride)
             continue;
         if (!(fmt->caps & PL_FMT_CAP_SAMPLEABLE))
             continue;
+        if (data->row_stride % fmt->texel_align)
+            continue; // reject misaligned formats
 
         int idx = 0;
 
@@ -201,8 +201,8 @@ next_fmt: ; // acts as `continue`
     return NULL;
 }
 
-bool pl_upload_plane(const struct pl_gpu *gpu, struct pl_plane *out_plane,
-                     const struct pl_tex **tex, const struct pl_plane_data *data)
+bool pl_upload_plane(pl_gpu gpu, struct pl_plane *out_plane,
+                     pl_tex *tex, const struct pl_plane_data *data)
 {
     pl_assert(!data->buf ^ !data->pixels); // exactly one
 
@@ -211,15 +211,8 @@ bool pl_upload_plane(const struct pl_gpu *gpu, struct pl_plane *out_plane,
         pl_assert(data->buf_offset == PL_ALIGN(data->buf_offset, data->pixel_stride));
     }
 
-    size_t row_stride = PL_DEF(data->row_stride, data->pixel_stride * data->width);
-    unsigned int stride_texels = row_stride / data->pixel_stride;
-    if (stride_texels * data->pixel_stride != row_stride) {
-        PL_ERR(gpu, "data->row_stride must be a multiple of data->pixel_stride!");
-        return false;
-    }
-
     int out_map[4];
-    const struct pl_fmt *fmt = pl_plane_find_fmt(gpu, out_map, data);
+    pl_fmt fmt = pl_plane_find_fmt(gpu, out_map, data);
     if (!fmt) {
         PL_ERR(gpu, "Failed picking any compatible texture format for a plane!");
         return false;
@@ -227,14 +220,14 @@ bool pl_upload_plane(const struct pl_gpu *gpu, struct pl_plane *out_plane,
         // TODO: try soft-converting to a supported format using e.g zimg?
     }
 
-    bool ok = pl_tex_recreate(gpu, tex, &(struct pl_tex_params) {
+    bool ok = pl_tex_recreate(gpu, tex, pl_tex_params(
         .w = data->width,
         .h = data->height,
         .format = fmt,
         .sampleable = true,
         .host_writable = true,
         .blit_src = fmt->caps & PL_FMT_CAP_BLITTABLE,
-    });
+    ));
 
     if (!ok) {
         PL_ERR(gpu, "Failed initializing plane texture!");
@@ -251,36 +244,36 @@ bool pl_upload_plane(const struct pl_gpu *gpu, struct pl_plane *out_plane,
         }
     }
 
-    return pl_tex_upload(gpu, &(struct pl_tex_transfer_params) {
+    return pl_tex_upload(gpu, pl_tex_transfer_params(
         .tex        = *tex,
-        .stride_w   = stride_texels,
+        .row_pitch  = data->row_stride,
         .ptr        = (void *) data->pixels,
         .buf        = data->buf,
         .buf_offset = data->buf_offset,
         .callback   = data->callback,
         .priv       = data->priv,
-    });
+    ));
 }
 
-bool pl_recreate_plane(const struct pl_gpu *gpu, struct pl_plane *out_plane,
-                       const struct pl_tex **tex, const struct pl_plane_data *data)
+bool pl_recreate_plane(pl_gpu gpu, struct pl_plane *out_plane,
+                       pl_tex *tex, const struct pl_plane_data *data)
 {
     int out_map[4];
-    const struct pl_fmt *fmt = pl_plane_find_fmt(gpu, out_map, data);
+    pl_fmt fmt = pl_plane_find_fmt(gpu, out_map, data);
     if (!fmt) {
         PL_ERR(gpu, "Failed picking any compatible texture format for a plane!");
         return false;
     }
 
-    bool ok = pl_tex_recreate(gpu, tex, &(struct pl_tex_params) {
+    bool ok = pl_tex_recreate(gpu, tex, pl_tex_params(
         .w = data->width,
         .h = data->height,
         .format = fmt,
         .renderable = true,
-        .host_readable = true,
+        .host_readable = fmt->caps & PL_FMT_CAP_HOST_READABLE,
         .blit_dst = fmt->caps & PL_FMT_CAP_BLITTABLE,
         .storable = fmt->caps & PL_FMT_CAP_STORABLE,
-    });
+    ));
 
     if (!ok) {
         PL_ERR(gpu, "Failed initializing plane texture!");
