@@ -37,6 +37,8 @@ void vk_dev_callback(struct vk_ctx *vk, vk_cb callback,
 
 // Helper wrapper around command buffers that also track dependencies,
 // callbacks and synchronization primitives
+//
+// Thread-safety: Unsafe
 struct vk_cmd {
     struct vk_cmdpool *pool; // pool it was allocated from
     VkQueue queue;           // the submission queue (for recording/pending)
@@ -46,16 +48,15 @@ struct vk_cmd {
     // this command can be executed. These are *not* owned by the vk_cmd
     PL_ARRAY(VkSemaphore) deps;
     PL_ARRAY(VkPipelineStageFlags) depstages;
+    PL_ARRAY(uint64_t) depvalues;
     // The signals represent semaphores that fire once the command finishes
     // executing. These are also not owned by the vk_cmd
     PL_ARRAY(VkSemaphore) sigs;
+    PL_ARRAY(uint64_t) sigvalues;
     // Since VkFences are useless, we have to manually track "callbacks"
     // to fire once the VkFence completes. These are used for multiple purposes,
     // ranging from garbage collection (resource deallocation) to fencing.
     PL_ARRAY(struct vk_callback) callbacks;
-    // Abstract objects associated with this command. Can be used to
-    // selectively flush.
-    PL_ARRAY(const void *) objs;
 };
 
 // Associate a callback with the completion of the current command. This
@@ -65,52 +66,39 @@ void vk_cmd_callback(struct vk_cmd *cmd, vk_cb callback,
 
 // Associate a raw dependency for the current command. This semaphore must
 // signal by the corresponding stage before the command may execute.
-void vk_cmd_dep(struct vk_cmd *cmd, VkSemaphore dep, VkPipelineStageFlags stage);
-
-// Associate an object with a command. This can be used to partially flush
-// commands only involving the object in question.
-void vk_cmd_obj(struct vk_cmd *cmd, const void *obj);
+void vk_cmd_dep(struct vk_cmd *cmd, VkPipelineStageFlags stage, pl_vulkan_sem dep);
 
 // Associate a raw signal with the current command. This semaphore will signal
 // after the command completes.
-void vk_cmd_sig(struct vk_cmd *cmd, VkSemaphore sig);
+void vk_cmd_sig(struct vk_cmd *cmd, pl_vulkan_sem sig);
 
-enum vk_wait_type {
-    VK_WAIT_NONE,    // no synchronization needed
-    VK_WAIT_BARRIER, // synchronization via pipeline barriers
-    VK_WAIT_EVENT,   // synchronization via events
+// Synchronization scope
+struct vk_sync_scope {
+    uint64_t value;             // last timeline semaphore value
+    VkQueue queue;              // source queue of last access
+    VkPipelineStageFlags stage; // stage bitmask of last access
+    VkAccessFlags access;       // access type bitmask
 };
 
-// Signal abstraction: represents an abstract synchronization mechanism.
-// Internally, this may either resolve as a semaphore or an event depending
-// on whether the appropriate conditions are met.
-struct vk_signal;
+// Synchronization primitive
+struct vk_sem {
+    // timeline semaphores, together with a pair of structs respectively
+    // describing the last read and write access, separately
+    VkSemaphore semaphore;
+    struct vk_sync_scope read, write;
+};
 
-// Generates a signal after the execution of all previous commands matching the
-// given the pipeline stage. The signal is owned by the caller, and must be
-// consumed eith vk_cmd_wait or released with vk_signal_cancel in order to
-// free the resources.
-struct vk_signal *vk_cmd_signal(struct vk_ctx *vk, struct vk_cmd *cmd,
-                                VkPipelineStageFlags stage);
+bool vk_sem_init(struct vk_ctx *vk, struct vk_sem *sem, pl_debug_tag debug_tag);
+void vk_sem_uninit(struct vk_ctx *vk, struct vk_sem *sem);
 
-// Consumes a previously generated signal. This signal must fire by the
-// indicated stage before the command can run. This function takes over
-// ownership of the signal (and the signal will be released/reused
-// automatically)
+// Updates the `vk_sem` state for a given access. If `is_trans` is set, this
+// access is treated as a write (since it alters the resource's state).
 //
-// The return type indicates what the caller needs to do:
-//   VK_SIGNAL_NONE:    no further handling needed, caller can use TOP_OF_PIPE
-//   VK_SIGNAL_BARRIER: caller must use pipeline barrier from last stage
-//   VK_SIGNAL_EVENT:   caller must use VkEvent from last stage
-//                      (never returned if out_event is NULL)
-enum vk_wait_type vk_cmd_wait(struct vk_ctx *vk, struct vk_cmd *cmd,
-                              struct vk_signal **sigptr,
-                              VkPipelineStageFlags stage,
-                              VkEvent *out_event);
-
-// Destroys a currently pending signal, for example if the resource is no
-// longer relevant.
-void vk_signal_destroy(struct vk_ctx *vk, struct vk_signal **sig);
+// Returns a struct describing the previous access to a resource. A pipeline
+// barrier is only required if the previous access scope is nonzero.
+struct vk_sync_scope vk_sem_barrier(struct vk_ctx *vk, struct vk_cmd *cmd,
+                                    struct vk_sem *sem, VkPipelineStageFlags stage,
+                                    VkAccessFlags access, bool is_trans);
 
 // Command pool / queue family hybrid abstraction
 struct vk_cmdpool {
@@ -136,9 +124,9 @@ void vk_cmdpool_destroy(struct vk_ctx *vk, struct vk_cmdpool *pool);
 // Returns NULL on failure.
 struct vk_cmd *vk_cmd_begin(struct vk_ctx *vk, struct vk_cmdpool *pool);
 
-// Finish recording a command buffer and queue it for execution. This function
+// Finish recording a command buffer and submit it for execution. This function
 // takes over ownership of **cmd, and sets *cmd to NULL in doing so.
-bool vk_cmd_queue(struct vk_ctx *vk, struct vk_cmd **cmd);
+bool vk_cmd_submit(struct vk_ctx *vk, struct vk_cmd **cmd);
 
 // Block until some commands complete executing. This is the only function that
 // actually processes the callbacks. Will wait at most `timeout` nanoseconds
@@ -151,14 +139,6 @@ bool vk_cmd_queue(struct vk_ctx *vk, struct vk_cmd **cmd);
 // never flushed!
 bool vk_poll_commands(struct vk_ctx *vk, uint64_t timeout);
 
-// Flush all currently queued commands. Returns whether successful. Failed
-// commands will be implicitly dropped.
-bool vk_flush_commands(struct vk_ctx *vk);
-
-// Like `vk_flush_commands`, but only flushes up to the last command involving
-// `obj`, inclusive. If `obj` is NULL, behaves as `vk_flush_commands`.
-bool vk_flush_obj(struct vk_ctx *vk, const void *obj);
-
 // Rotate through queues in each command pool. Call this once per frame, after
 // submitting all of the command buffers for that frame. Calling this more
 // often than that is possible but bad for performance.
