@@ -17,8 +17,8 @@
 
 #pragma once
 
-#include "gpu.h"
 #include "common.h"
+#include "gpu.h"
 
 struct vk_format {
     VkFormat tfmt;     // internal vulkan format enum (textures)
@@ -29,3 +29,6 @@ struct vk_format {
 };
 
 extern const struct vk_format vk_formats[];
+
+// Add all supported formats to the `pl_gpu` format list
+void vk_setup_formats(struct pl_gpu *gpu);
