@@ -21,6 +21,7 @@
 #include "utils.h"
 #include "gpu.h"
 #include "swapchain.h"
+#include "pl_thread.h"
 
 struct sem_pair {
     VkSemaphore in;
@@ -28,8 +29,10 @@ struct sem_pair {
 };
 
 struct priv {
+    pl_mutex lock;
     struct vk_ctx *vk;
     VkSurfaceKHR surf;
+    PL_ARRAY(VkSurfaceFormatKHR) formats;
 
     // current swapchain and metadata:
     struct pl_vulkan_swapchain_params params;
@@ -38,22 +41,23 @@ struct priv {
     VkSwapchainKHR old_swapchain;
     int cur_width, cur_height;
     int swapchain_depth;
-    int frames_in_flight;   // number of frames currently queued
-    bool suboptimal;        // true once VK_SUBOPTIMAL_KHR is returned
+    pl_rc_t frames_in_flight;       // number of frames currently queued
+    bool suboptimal;                // true once VK_SUBOPTIMAL_KHR is returned
+    bool needs_recreate;            // swapchain needs to be recreated
     struct pl_color_repr color_repr;
     struct pl_color_space color_space;
     struct pl_hdr_metadata hdr_metadata;
 
     // state of the images:
-    PL_ARRAY(const struct pl_tex *) images; // pl_tex wrappers for the VkImages
-    PL_ARRAY(struct sem_pair) sems;         // pool of semaphores used to synchronize images
-    int idx_sems;                           // index of next free semaphore pair
-    int last_imgidx;                        // the image index last acquired (for submit)
+    PL_ARRAY(pl_tex) images;        // pl_tex wrappers for the VkImages
+    PL_ARRAY(struct sem_pair) sems; // pool of semaphores used to synchronize images
+    int idx_sems;                   // index of next free semaphore pair
+    int last_imgidx;                // the image index last acquired (for submit)
 };
 
 static struct pl_sw_fns vulkan_swapchain;
 
-static bool vk_map_color_space(VkColorSpaceKHR space, struct pl_color_space *out)
+static bool map_color_space(VkColorSpaceKHR space, struct pl_color_space *out)
 {
     switch (space) {
     // Note: This is technically against the spec, but more often than not
@@ -144,56 +148,23 @@ static bool vk_map_color_space(VkColorSpaceKHR space, struct pl_color_space *out
     }
 }
 
-static bool pick_surf_format(const struct pl_gpu *gpu, const struct vk_ctx *vk,
-                             VkSurfaceKHR surf, bool prefer_hdr,
-                             VkSurfaceFormatKHR *out_format,
-                             struct pl_color_space *out_space)
+static bool pick_surf_format(pl_swapchain sw, const struct pl_color_space *hint)
 {
+    struct priv *p = PL_PRIV(sw);
+    struct vk_ctx *vk = p->vk;
+    pl_gpu gpu = sw->gpu;
+
     int best_score = 0, best_id;
-    VkSurfaceFormatKHR *formats = NULL;
-    int num = 0;
-
-    // Specific format requested by user
-    if (out_format->format) {
-        if (vk_map_color_space(out_format->colorSpace, out_space)) {
-            PL_INFO(gpu, "Using user-supplied surface configuration: %s + %s",
-                    vk_fmt_name(out_format->format),
-                    vk_csp_name(out_format->colorSpace));
-            return true;
-        } else {
-            PL_ERR(gpu, "User-supplied surface format unsupported: %s",
-                   vk_fmt_name(out_format->format));
-        }
-    }
-
-    VK(vk->GetPhysicalDeviceSurfaceFormatsKHR(vk->physd, surf, &num, NULL));
-    formats = pl_calloc_ptr(NULL, num, formats);
-    VK(vk->GetPhysicalDeviceSurfaceFormatsKHR(vk->physd, surf, &num, formats));
-
-    PL_INFO(gpu, "Available surface configurations:");
-    for (int i = 0; i < num; i++) {
-        PL_INFO(gpu, "    %d: %-40s %s", i,
-                vk_fmt_name(formats[i].format),
-                vk_csp_name(formats[i].colorSpace));
-    }
-
-    for (int i = 0; i < num; i++) {
-        // A value of VK_FORMAT_UNDEFINED means we can pick anything we want
-        if (formats[i].format == VK_FORMAT_UNDEFINED) {
-            *out_format = (VkSurfaceFormatKHR) {
-                .colorSpace = VK_COLOR_SPACE_SRGB_NONLINEAR_KHR,
-                .format = VK_FORMAT_R8G8B8A8_UNORM,
-            };
-            pl_free(formats);
-            return true;
-        }
+    bool wide_gamut = pl_color_primaries_is_wide_gamut(hint->primaries);
+    bool prefer_hdr = pl_color_transfer_is_hdr(hint->transfer);
 
+    for (int i = 0; i < p->formats.num; i++) {
         // Color space / format whitelist
         struct pl_color_space space;
-        if (!vk_map_color_space(formats[i].colorSpace, &space))
+        if (!map_color_space(p->formats.elem[i].colorSpace, &space))
             continue;
 
-        switch (formats[i].format) {
+        switch (p->formats.elem[i].format) {
         // Only accept floating point formats for linear curves
         case VK_FORMAT_R16G16B16_SFLOAT:
         case VK_FORMAT_R16G16B16A16_SFLOAT:
@@ -232,9 +203,9 @@ static bool pick_surf_format(const struct pl_gpu *gpu, const struct vk_ctx *vk,
 
         // Make sure we can wrap this format to a meaningful, valid pl_fmt
         for (int n = 0; n < gpu->num_formats; n++) {
-            const struct pl_fmt *plfmt = gpu->formats[n];
+            pl_fmt plfmt = gpu->formats[n];
             const struct vk_format **pvkfmt = PL_PRIV(plfmt);
-            if ((*pvkfmt)->tfmt != formats[i].format)
+            if ((*pvkfmt)->tfmt != p->formats.elem[i].format)
                 continue;
 
             enum pl_fmt_caps render_caps = 0;
@@ -247,19 +218,26 @@ static bool pick_surf_format(const struct pl_gpu *gpu, const struct vk_ctx *vk,
             int score = 0;
             for (int c = 0; c < 3; c++)
                 score += plfmt->component_depth[c];
+            if (pl_color_primaries_is_wide_gamut(space.primaries) == wide_gamut)
+                score += 1000;
+            if (space.primaries == hint->primaries)
+                score += 2000;
             if (pl_color_transfer_is_hdr(space.transfer) == prefer_hdr)
                 score += 10000;
+            if (space.transfer == hint->transfer)
+                score += 20000;
 
             switch (plfmt->type) {
-            case PL_FMT_UNORM: score += 3000; break;
-            case PL_FMT_SNORM: score += 2000; break;
-            case PL_FMT_FLOAT: score += 1000; break;
-            default: break;
+            case PL_FMT_UNKNOWN: break;
+            case PL_FMT_UINT: break;
+            case PL_FMT_SINT: break;
+            case PL_FMT_UNORM: score += 500; break;
+            case PL_FMT_SNORM: score += 400; break;
+            case PL_FMT_FLOAT: score += 300; break;
+            case PL_FMT_TYPE_COUNT: pl_unreachable();
             };
 
             if (score > best_score) {
-                *out_format = formats[i];
-                *out_space = space;
                 best_score = score;
                 best_id = i;
                 break;
@@ -267,49 +245,89 @@ static bool pick_surf_format(const struct pl_gpu *gpu, const struct vk_ctx *vk,
         }
     }
 
-    if (best_score)
-        PL_INFO(gpu, "Picked surface configuration %d", best_id);
+    if (!best_score) {
+        PL_ERR(vk, "Failed picking any valid, renderable surface format!");
+        return false;
+    }
 
-    // fall through
-error:
-    if (!best_score)
-        PL_FATAL(vk, "Failed picking any valid, renderable surface format!");
-    pl_free(formats);
-    return best_score > 0;
+    VkSurfaceFormatKHR new_sfmt = p->formats.elem[best_id];
+    if (p->protoInfo.imageFormat != new_sfmt.format ||
+        p->protoInfo.imageColorSpace != new_sfmt.colorSpace)
+    {
+        PL_INFO(vk, "Picked surface configuration %d: %s + %s", best_id,
+                vk_fmt_name(new_sfmt.format),
+                vk_csp_name(new_sfmt.colorSpace));
+
+        p->protoInfo.imageFormat = new_sfmt.format;
+        p->protoInfo.imageColorSpace = new_sfmt.colorSpace;
+        p->needs_recreate = true;
+    }
+
+    return true;
 }
 
-const struct pl_swapchain *pl_vulkan_create_swapchain(const struct pl_vulkan *plvk,
+static void set_hdr_metadata(struct priv *p, const struct pl_hdr_metadata *metadata)
+{
+    struct vk_ctx *vk = p->vk;
+    if (!vk->SetHdrMetadataEXT)
+        return;
+
+    // Ignore no-op changes
+    if (pl_hdr_metadata_equal(metadata, &p->hdr_metadata))
+        return;
+
+    // Remember the metadata so we can re-apply it after swapchain recreation
+    p->hdr_metadata = *metadata;
+
+    // Ignore HDR metadata requests for SDR swapchains
+    if (!pl_color_transfer_is_hdr(p->color_space.transfer))
+        return;
+
+    if (!p->swapchain)
+        return;
+
+    vk->SetHdrMetadataEXT(vk->dev, 1, &p->swapchain, &(VkHdrMetadataEXT) {
+        .sType = VK_STRUCTURE_TYPE_HDR_METADATA_EXT,
+        .displayPrimaryRed   = { metadata->prim.red.x,   metadata->prim.red.y },
+        .displayPrimaryGreen = { metadata->prim.green.x, metadata->prim.green.y },
+        .displayPrimaryBlue  = { metadata->prim.blue.x,  metadata->prim.blue.y },
+        .whitePoint = { metadata->prim.white.x, metadata->prim.white.y },
+        .maxLuminance = metadata->max_luma,
+        .minLuminance = metadata->min_luma,
+        .maxContentLightLevel = metadata->max_cll,
+        .maxFrameAverageLightLevel = metadata->max_fall,
+    });
+}
+
+pl_swapchain pl_vulkan_create_swapchain(pl_vulkan plvk,
                               const struct pl_vulkan_swapchain_params *params)
 {
     struct vk_ctx *vk = PL_PRIV(plvk);
-    const struct pl_gpu *gpu = plvk->gpu;
+    pl_gpu gpu = plvk->gpu;
 
     if (!vk->CreateSwapchainKHR) {
         PL_ERR(gpu, VK_KHR_SWAPCHAIN_EXTENSION_NAME " not enabled!");
         return NULL;
     }
 
-    VkSurfaceFormatKHR sfmt = params->surface_format;
-    struct pl_color_space csp;
-    if (!pick_surf_format(gpu, vk, params->surface, params->prefer_hdr, &sfmt, &csp))
-        return NULL;
-
-    struct pl_swapchain *sw = pl_zalloc_priv(NULL, struct pl_swapchain, struct priv);
+    struct pl_swapchain *sw = pl_zalloc_obj(NULL, sw, struct priv);
     sw->impl = &vulkan_swapchain;
-    sw->ctx = vk->ctx;
+    sw->log = vk->log;
+    sw->ctx = sw->log;
     sw->gpu = gpu;
 
     struct priv *p = PL_PRIV(sw);
+    pl_mutex_init(&p->lock);
     p->params = *params;
     p->vk = vk;
     p->surf = params->surface;
     p->swapchain_depth = PL_DEF(params->swapchain_depth, 3);
     pl_assert(p->swapchain_depth > 0);
+    atomic_init(&p->frames_in_flight, 0);
+    p->last_imgidx = -1;
     p->protoInfo = (VkSwapchainCreateInfoKHR) {
         .sType = VK_STRUCTURE_TYPE_SWAPCHAIN_CREATE_INFO_KHR,
         .surface = p->surf,
-        .imageFormat = sfmt.format,
-        .imageColorSpace = sfmt.colorSpace,
         .imageArrayLayers = 1, // non-stereoscopic
         .imageSharingMode = VK_SHARING_MODE_EXCLUSIVE,
         .minImageCount = p->swapchain_depth + 1, // +1 for the FB
@@ -317,16 +335,17 @@ const struct pl_swapchain *pl_vulkan_create_swapchain(const struct pl_vulkan *pl
         .clipped = true,
     };
 
-    p->color_space = csp;
+    // These fields will be updated by `vk_sw_recreate`
+    p->color_space = pl_color_space_unknown;
     p->color_repr = (struct pl_color_repr) {
         .sys    = PL_COLOR_SYSTEM_RGB,
         .levels = PL_COLOR_LEVELS_FULL,
-        .alpha  = PL_ALPHA_UNKNOWN, // will be updated by vk_sw_recreate
+        .alpha  = PL_ALPHA_UNKNOWN,
     };
 
     // Make sure the swapchain present mode is supported
     VkPresentModeKHR *modes = NULL;
-    int num_modes = 0;
+    uint32_t num_modes = 0;
     VK(vk->GetPhysicalDeviceSurfacePresentModesKHR(vk->physd, p->surf, &num_modes, NULL));
     modes = pl_calloc_ptr(NULL, num_modes, modes);
     VK(vk->GetPhysicalDeviceSurfacePresentModesKHR(vk->physd, p->surf, &num_modes, modes));
@@ -342,6 +361,31 @@ const struct pl_swapchain *pl_vulkan_create_swapchain(const struct pl_vulkan *pl
         p->protoInfo.presentMode = VK_PRESENT_MODE_FIFO_KHR;
     }
 
+    // Enumerate the supported surface color spaces
+    uint32_t num_formats = 0;
+    VK(vk->GetPhysicalDeviceSurfaceFormatsKHR(vk->physd, p->surf, &num_formats, NULL));
+    PL_ARRAY_RESIZE(sw, p->formats, num_formats);
+    VK(vk->GetPhysicalDeviceSurfaceFormatsKHR(vk->physd, p->surf, &num_formats, p->formats.elem));
+    p->formats.num = num_formats;
+
+    PL_INFO(gpu, "Available surface configurations:");
+    for (int i = 0; i < p->formats.num; i++) {
+        PL_INFO(gpu, "    %d: %-40s %s", i,
+                vk_fmt_name(p->formats.elem[i].format),
+                vk_csp_name(p->formats.elem[i].colorSpace));
+    }
+
+    struct pl_swapchain_colors hint = {0};
+    if (params->prefer_hdr) {
+        hint.primaries = PL_COLOR_PRIM_BT_2020;
+        hint.transfer = PL_COLOR_TRC_PQ;
+        hint.hdr = pl_hdr_metadata_hdr10;
+    }
+
+    // Ensure there exists at least some valid renderable surface format
+    if (!pick_surf_format(sw, &hint))
+        goto error;
+
     return sw;
 
 error:
@@ -350,9 +394,9 @@ error:
     return NULL;
 }
 
-static void vk_sw_destroy(const struct pl_swapchain *sw)
+static void vk_sw_destroy(pl_swapchain sw)
 {
-    const struct pl_gpu *gpu = sw->gpu;
+    pl_gpu gpu = sw->gpu;
     struct priv *p = PL_PRIV(sw);
     struct vk_ctx *vk = p->vk;
 
@@ -366,10 +410,11 @@ static void vk_sw_destroy(const struct pl_swapchain *sw)
     }
 
     vk->DestroySwapchainKHR(vk->dev, p->swapchain, PL_VK_ALLOC);
+    pl_mutex_destroy(&p->lock);
     pl_free((void *) sw);
 }
 
-static int vk_sw_latency(const struct pl_swapchain *sw)
+static int vk_sw_latency(pl_swapchain sw)
 {
     struct priv *p = PL_PRIV(sw);
     return p->swapchain_depth;
@@ -394,10 +439,10 @@ static bool update_swapchain_info(struct priv *p, VkSwapchainCreateInfoKHR *info
     static const struct { VkCompositeAlphaFlagsKHR vk_mode;
                           enum pl_alpha_mode pl_mode;
                         } alphaModes[] = {
-        {VK_COMPOSITE_ALPHA_PRE_MULTIPLIED_BIT_KHR,  PL_ALPHA_PREMULTIPLIED},
         {VK_COMPOSITE_ALPHA_POST_MULTIPLIED_BIT_KHR, PL_ALPHA_INDEPENDENT},
-        {VK_COMPOSITE_ALPHA_INHERIT_BIT_KHR,         PL_ALPHA_UNKNOWN},
+        {VK_COMPOSITE_ALPHA_PRE_MULTIPLIED_BIT_KHR,  PL_ALPHA_PREMULTIPLIED},
         {VK_COMPOSITE_ALPHA_OPAQUE_BIT_KHR,          PL_ALPHA_UNKNOWN},
+        {VK_COMPOSITE_ALPHA_INHERIT_BIT_KHR,         PL_ALPHA_UNKNOWN},
     };
 
     for (int i = 0; i < PL_ARRAY_SIZE(alphaModes); i++) {
@@ -510,22 +555,20 @@ static void destroy_swapchain(struct vk_ctx *vk, struct priv *p)
     p->old_swapchain = VK_NULL_HANDLE;
 }
 
-static bool vk_sw_recreate(const struct pl_swapchain *sw, int w, int h)
+static bool vk_sw_recreate(pl_swapchain sw, int w, int h)
 {
-    const struct pl_gpu *gpu = sw->gpu;
+    pl_gpu gpu = sw->gpu;
     struct priv *p = PL_PRIV(sw);
     struct vk_ctx *vk = p->vk;
 
     VkImage *vkimages = NULL;
-    int num_images = 0;
+    uint32_t num_images = 0;
 
     // It's invalid to trigger another swapchain recreation while there's more
     // than one swapchain already active, so we need to flush any pending
     // asynchronous swapchain release operations that may be ongoing
-    while (p->old_swapchain) {
-        vk_flush_commands(vk);
+    while (p->old_swapchain)
         vk_poll_commands(vk, UINT64_MAX);
-    }
 
     VkSwapchainCreateInfoKHR sinfo = p->protoInfo;
     sinfo.oldSwapchain = p->swapchain;
@@ -540,6 +583,7 @@ static bool vk_sw_recreate(const struct pl_swapchain *sw, int w, int h)
     VK(vk->CreateSwapchainKHR(vk->dev, &sinfo, PL_VK_ALLOC, &p->swapchain));
 
     p->suboptimal = false;
+    p->needs_recreate = false;
     p->cur_width = sinfo.imageExtent.width;
     p->cur_height = sinfo.imageExtent.height;
 
@@ -582,13 +626,13 @@ static bool vk_sw_recreate(const struct pl_swapchain *sw, int w, int h)
 
     for (int i = 0; i < num_images; i++) {
         const VkExtent2D *ext = &sinfo.imageExtent;
-        const struct pl_tex *tex = pl_vulkan_wrap(gpu, &(struct pl_vulkan_wrap_params) {
+        pl_tex tex = pl_vulkan_wrap(gpu, pl_vulkan_wrap_params(
             .image = vkimages[i],
             .width = ext->width,
             .height = ext->height,
             .format = sinfo.imageFormat,
             .usage = sinfo.imageUsage,
-        });
+        ));
         if (!tex)
             goto error;
         PL_ARRAY_APPEND(sw, p->images, tex);
@@ -601,15 +645,20 @@ static bool vk_sw_recreate(const struct pl_swapchain *sw, int w, int h)
     // the actual color information (consider e.g. a2bgr10). Slight downside
     // in that it results in rounding r/b for e.g. rgb565, but we don't pick
     // surfaces with fewer than 8 bits anyway, so let's not care for now.
-    const struct pl_fmt *fmt = p->images.elem[0]->params.format;
+    pl_fmt fmt = p->images.elem[0]->params.format;
     for (int i = 0; i < fmt->num_components; i++)
         bits = PL_MAX(bits, fmt->component_depth[i]);
 
     p->color_repr.bits.sample_depth = bits;
     p->color_repr.bits.color_depth = bits;
 
-    // Re-set the HDR metadata
-    pl_swapchain_hdr_metadata(sw, &p->hdr_metadata);
+    // FIXME: infer `p->color_space.sig_peak` etc. from HDR metadata?
+    map_color_space(sinfo.imageColorSpace, &p->color_space);
+
+    // Forcibly re-apply HDR metadata, bypassing the no-op check
+    struct pl_hdr_metadata metadata = p->hdr_metadata;
+    p->hdr_metadata = pl_hdr_metadata_empty;
+    set_hdr_metadata(p, &metadata);
 
     pl_free(vkimages);
     return true;
@@ -622,22 +671,25 @@ error:
         p->swapchain = VK_NULL_HANDLE;
         p->cur_width = p->cur_height = 0;
         p->suboptimal = false;
+        p->needs_recreate = false;
     }
     return false;
 }
 
-static bool vk_sw_start_frame(const struct pl_swapchain *sw,
+static bool vk_sw_start_frame(pl_swapchain sw,
                               struct pl_swapchain_frame *out_frame)
 {
     struct priv *p = PL_PRIV(sw);
     struct vk_ctx *vk = p->vk;
-    if (!p->swapchain && !vk_sw_recreate(sw, 0, 0))
-        return false;
+    pl_mutex_lock(&p->lock);
 
-    if (p->suboptimal && !p->params.allow_suboptimal) {
-        PL_TRACE(vk, "Swapchain is suboptimal, recreating...");
-        if (!vk_sw_recreate(sw, 0, 0))
-            return false;
+    bool recreate = !p->swapchain || p->needs_recreate;
+    if (p->suboptimal && !p->params.allow_suboptimal)
+        recreate = true;
+
+    if (recreate && !vk_sw_recreate(sw, 0, 0)) {
+        pl_mutex_unlock(&p->lock);
+        return false;
     }
 
     VkSemaphore sem_in = p->sems.elem[p->idx_sems].in;
@@ -655,72 +707,83 @@ static bool vk_sw_start_frame(const struct pl_swapchain *sw,
         case VK_SUCCESS:
             p->last_imgidx = imgidx;
             pl_vulkan_release(sw->gpu, p->images.elem[imgidx],
-                              VK_IMAGE_LAYOUT_UNDEFINED, 0, sem_in);
+                              VK_IMAGE_LAYOUT_UNDEFINED,
+                              (pl_vulkan_sem){ sem_in });
             *out_frame = (struct pl_swapchain_frame) {
                 .fbo = p->images.elem[imgidx],
                 .flipped = false,
                 .color_repr = p->color_repr,
                 .color_space = p->color_space,
             };
+            // keep lock held
             return true;
 
         case VK_ERROR_OUT_OF_DATE_KHR: {
             // In these cases try recreating the swapchain
-            if (!vk_sw_recreate(sw, 0, 0))
+            if (!vk_sw_recreate(sw, 0, 0)) {
+                pl_mutex_unlock(&p->lock);
                 return false;
+            }
             continue;
         }
 
         default:
             PL_ERR(vk, "Failed acquiring swapchain image: %s", vk_res_str(res));
+            pl_mutex_unlock(&p->lock);
             return false;
         }
     }
 
     // If we've exhausted the number of attempts to recreate the swapchain,
     // just give up silently and let the user retry some time later.
+    pl_mutex_unlock(&p->lock);
     return false;
 }
 
 static void present_cb(struct priv *p, void *arg)
 {
-    p->frames_in_flight--;
+    (void) pl_rc_deref(&p->frames_in_flight);
 }
 
-static bool vk_sw_submit_frame(const struct pl_swapchain *sw)
+static bool vk_sw_submit_frame(pl_swapchain sw)
 {
-    const struct pl_gpu *gpu = sw->gpu;
+    pl_gpu gpu = sw->gpu;
     struct priv *p = PL_PRIV(sw);
     struct vk_ctx *vk = p->vk;
-    if (!p->swapchain)
-        return false;
-
+    pl_assert(p->last_imgidx >= 0);
+    pl_assert(p->swapchain);
+    uint32_t idx = p->last_imgidx;
     VkSemaphore sem_out = p->sems.elem[p->idx_sems++].out;
     p->idx_sems %= p->sems.num;
+    p->last_imgidx = -1;
 
-    bool held = pl_vulkan_hold(gpu, p->images.elem[p->last_imgidx],
+    bool held = pl_vulkan_hold(gpu, p->images.elem[idx],
                                VK_IMAGE_LAYOUT_PRESENT_SRC_KHR,
-                               VK_ACCESS_MEMORY_READ_BIT, sem_out);
+                               (pl_vulkan_sem){ sem_out });
     if (!held) {
         PL_ERR(gpu, "Failed holding swapchain image for presentation");
+        pl_mutex_unlock(&p->lock);
         return false;
     }
 
     struct vk_cmd *cmd = pl_vk_steal_cmd(gpu);
-    if (!cmd)
+    if (!cmd) {
+        pl_mutex_unlock(&p->lock);
         return false;
+    }
 
-    p->frames_in_flight++;
+    pl_rc_ref(&p->frames_in_flight);
     vk_cmd_callback(cmd, (vk_cb) present_cb, p, NULL);
-
-    vk_cmd_queue(vk, &cmd);
-    if (!vk_flush_commands(vk))
+    if (!vk_cmd_submit(vk, &cmd)) {
+        pl_mutex_unlock(&p->lock);
         return false;
+    }
 
     struct vk_cmdpool *pool = vk->pool_graphics;
     VkQueue queue = pool->queues[pool->idx_queues];
 
     vk_rotate_queues(p->vk);
+    vk_malloc_garbage_collect(vk->ma);
 
     VkPresentInfoKHR pinfo = {
         .sType = VK_STRUCTURE_TYPE_PRESENT_INFO_KHR,
@@ -728,11 +791,13 @@ static bool vk_sw_submit_frame(const struct pl_swapchain *sw)
         .pWaitSemaphores = &sem_out,
         .swapchainCount = 1,
         .pSwapchains = &p->swapchain,
-        .pImageIndices = &p->last_imgidx,
+        .pImageIndices = &idx,
     };
 
     PL_TRACE(vk, "vkQueuePresentKHR waits on %p", (void *) sem_out);
     VkResult res = vk->QueuePresentKHR(queue, &pinfo);
+    pl_mutex_unlock(&p->lock);
+
     switch (res) {
     case VK_SUBOPTIMAL_KHR:
         p->suboptimal = true;
@@ -752,77 +817,64 @@ static bool vk_sw_submit_frame(const struct pl_swapchain *sw)
     }
 }
 
-static void vk_sw_swap_buffers(const struct pl_swapchain *sw)
+static void vk_sw_swap_buffers(pl_swapchain sw)
 {
     struct priv *p = PL_PRIV(sw);
 
-    while (p->frames_in_flight >= p->swapchain_depth)
+    pl_mutex_lock(&p->lock);
+    while (pl_rc_count(&p->frames_in_flight) >= p->swapchain_depth) {
+        pl_mutex_unlock(&p->lock); // don't hold mutex while blocking
         vk_poll_commands(p->vk, UINT64_MAX);
+        pl_mutex_lock(&p->lock);
+    }
+    pl_mutex_unlock(&p->lock);
 }
 
-static bool vk_sw_resize(const struct pl_swapchain *sw, int *width, int *height)
+static bool vk_sw_resize(pl_swapchain sw, int *width, int *height)
 {
     struct priv *p = PL_PRIV(sw);
     bool ok = true;
 
+    pl_mutex_lock(&p->lock);
+
     bool width_changed = *width && *width != p->cur_width,
          height_changed = *height && *height != p->cur_height;
 
-    if (p->suboptimal || width_changed || height_changed)
+    if (p->suboptimal || p->needs_recreate || width_changed || height_changed)
         ok = vk_sw_recreate(sw, *width, *height);
 
     *width = p->cur_width;
     *height = p->cur_height;
+
+    pl_mutex_unlock(&p->lock);
     return ok;
 }
 
-static bool vk_sw_hdr_metadata(const struct pl_swapchain *sw,
-                               const struct pl_hdr_metadata *metadata)
+static void vk_sw_colorspace_hint(pl_swapchain sw, const struct pl_color_space *csp)
 {
     struct priv *p = PL_PRIV(sw);
-    struct vk_ctx *vk = p->vk;
-    if (!vk->SetHdrMetadataEXT)
-        return false;
+    pl_mutex_lock(&p->lock);
 
-    if (!pl_color_transfer_is_hdr(p->color_space.transfer))
-        return false;
+    // This should never fail if the swapchain already exists
+    bool ok = pick_surf_format(sw, csp);
+    set_hdr_metadata(p, &csp->hdr);
+    pl_assert(ok);
 
-    if (!p->swapchain && !vk_sw_recreate(sw, 0, 0))
-        return false;
-
-    if (!metadata)
-        return true;
-
-    // Remember the metadata so we can re-apply it after swapchain recreation
-    p->hdr_metadata = *metadata;
-
-    vk->SetHdrMetadataEXT(vk->dev, 1, &p->swapchain, &(VkHdrMetadataEXT) {
-        .sType = VK_STRUCTURE_TYPE_HDR_METADATA_EXT,
-        .displayPrimaryRed   = { metadata->prim.red.x,   metadata->prim.red.y },
-        .displayPrimaryGreen = { metadata->prim.green.x, metadata->prim.green.y },
-        .displayPrimaryBlue  = { metadata->prim.blue.x,  metadata->prim.blue.y },
-        .whitePoint = { metadata->prim.white.x, metadata->prim.white.y },
-        .maxLuminance = metadata->max_luma,
-        .minLuminance = metadata->min_luma,
-        .maxContentLightLevel = metadata->max_cll,
-        .maxFrameAverageLightLevel = metadata->max_fall,
-    });
-
-    return true;
+    pl_mutex_unlock(&p->lock);
 }
 
-bool pl_vulkan_swapchain_suboptimal(const struct pl_swapchain *sw)
+bool pl_vulkan_swapchain_suboptimal(pl_swapchain sw)
 {
     struct priv *p = PL_PRIV(sw);
     return p->suboptimal;
 }
 
 static struct pl_sw_fns vulkan_swapchain = {
-    .destroy      = vk_sw_destroy,
-    .latency      = vk_sw_latency,
-    .resize       = vk_sw_resize,
-    .hdr_metadata = vk_sw_hdr_metadata,
-    .start_frame  = vk_sw_start_frame,
-    .submit_frame = vk_sw_submit_frame,
-    .swap_buffers = vk_sw_swap_buffers,
+    .destroy            = vk_sw_destroy,
+    .latency            = vk_sw_latency,
+    .resize             = vk_sw_resize,
+    .colorspace_hint    = vk_sw_colorspace_hint,
+    .start_frame        = vk_sw_start_frame,
+    .submit_frame       = vk_sw_submit_frame,
+    .swap_buffers       = vk_sw_swap_buffers,
 };
