@@ -63,6 +63,17 @@ const char *vk_csp_name(VkColorSpaceKHR csp)
     }
 }
 
+const char *vk_handle_name(VkExternalMemoryHandleTypeFlagBitsKHR handle)
+{
+    switch (handle) {
+%for handle in vkhandles:
+    case ${handle}: return "${handle}";
+%endfor
+
+    default: return "unknown handle type";
+    }
+}
+
 const char *vk_alpha_mode(VkCompositeAlphaFlagsKHR alpha)
 {
     switch (alpha) {
@@ -107,6 +118,9 @@ size_t vk_struct_size(VkStructureType stype)
     default: return 0;
     }
 }
+
+const VkAccessFlags vk_access_read = ${hex(vkaccess.read)}LLU;
+const VkAccessFlags vk_access_write = ${hex(vkaccess.write)}LLU;
 """)
 
 class Obj(object):
@@ -120,7 +134,13 @@ def findall_enum(registry, name):
     for e in registry.iterfind('.//enum[@extends="{0}"]'.format(name)):
         # ext 289 is a non-existing extension that defines some names for
         # proprietary downstream consumers, causes problems unless excluded
-        if not 'alias' in e.attrib and e.attrib.get('extnumber', '0') != '289':
+        if e.attrib.get('extnumber', '0') == '289':
+            continue
+        # some other extensions contain reserved identifiers that generally
+        # translate to compile failures
+        if 'RESERVED' in e.attrib['name']:
+            continue
+        if not 'alias' in e.attrib:
             yield e
 
 def get_vkenum(registry, enum):
@@ -128,17 +148,17 @@ def get_vkenum(registry, enum):
         yield e.attrib['name']
 
 def get_vkobjects(registry):
-    for e in findall_enum(registry, 'VkObjectType'):
-        if 'comment' in e.attrib:
-            yield Obj(enum = e.attrib['name'],
-                      name = e.attrib['comment'])
+    for t in registry.iterfind('types/type[@category="handle"]'):
+        if 'objtypeenum' in t.attrib:
+            yield Obj(enum = t.attrib['objtypeenum'],
+                      name = t.find('name').text)
 
 def get_vkstructs(registry):
     for e in registry.iterfind('types/type[@category="struct"]'):
         # Strings for platform-specific crap we want to blacklist as they will
         # most likely cause build failures
         blacklist_strs = [
-            'ANDROID', 'Surface', 'Win32', 'D3D12', 'GGP'
+            'ANDROID', 'Surface', 'Win32', 'D3D12', 'GGP', 'FUCHSIA',
         ]
 
         if any([ str in e.attrib['name'] for str in blacklist_strs ]):
@@ -154,11 +174,21 @@ def get_vkstructs(registry):
             yield Obj(stype = stype.attrib['values'],
                       name = e.attrib['name'])
 
+def get_vkaccess(registry):
+    access = Obj(read = 0, write = 0)
+    for e in findall_enum(registry, 'VkAccessFlagBits'):
+        if '_READ_' in e.attrib['name']:
+            access.read |= 1 << int(e.attrib['bitpos'])
+        if '_WRITE_' in e.attrib['name']:
+            access.write |= 1 << int(e.attrib['bitpos'])
+    return access
+
 def find_registry_xml():
     registry_paths = [
-        '/usr/share/vulkan/registry/vk.xml',
         '%VULKAN_SDK%/share/vulkan/registry/vk.xml',
+        '$VULKAN_SDK/share/vulkan/registry/vk.xml',
         '$MINGW_PREFIX/share/vulkan/registry/vk.xml',
+        '/usr/share/vulkan/registry/vk.xml',
     ]
 
     for p in registry_paths:
@@ -186,8 +216,10 @@ if __name__ == '__main__':
             vkresults = get_vkenum(registry, 'VkResult'),
             vkformats = get_vkenum(registry, 'VkFormat'),
             vkspaces  = get_vkenum(registry, 'VkColorSpaceKHR'),
+            vkhandles = get_vkenum(registry, 'VkExternalMemoryHandleTypeFlagBits'),
             vkalphas  = get_vkenum(registry, 'VkCompositeAlphaFlagBitsKHR'),
             vktransforms = get_vkenum(registry, 'VkSurfaceTransformFlagBitsKHR'),
             vkobjects = get_vkobjects(registry),
             vkstructs = get_vkstructs(registry),
+            vkaccess = get_vkaccess(registry),
         ))
